// requires metric of 6 - Pixels
/*
    DateTimeSpinbox.cc

    A custom class that is completely native dBASE code,
    designed to try to allow a user to treat each element
    of a datetime object with a pair of pushbuttons on the
    right that act as a spinbox -- allowing the user to
    modify each element of the 'spinbox' separately.
    Based on a request for something of this nature in the
    dBASE Newsgroups.

    Note that this was a lot more difficult to create than
    it appeared it would be. The difficulty is all the options
    in dBASE for handling date, time and hour displays. 
    This control attempts to take the various options and
    work with them all. 

    Programmer: Ken Mayer
    Date......: June, 2006
    Thanks....: Ivar B. Jessen for beating it up a bit and helping
                with ideas and testing. While Ivar and I don't
                always see eye to eye on techniques and such, 
                he is a big help with international issues 
                that I am not always aware of.

    Usage:
       set procedure to :dUFLP:DateTimeSpinbox.cc
          drag an instance of this to your form
          in the form's overridden OPEN or READMODAL methods
          (all of the startup code shown should be set there,
          so that it is seen before the form opens, otherwise
          the various properties will not be created until
          after the form has opened, which is too late):

          form.DateTimeSpinbox1.DT = datetime()

       The value can be gotten by querying the control's "DT" property:
       ? form.DateTimeSpinbox1.DT
           or
       ? form.DateTimeSpinbox1.DT.toLocaleString()

       If using this with a table:
       This control does not understand dataLinks, so you
       will need to handle this manually. Suggest that
       in the rowset's onNavigate something like the
       following be used:

          form.DateTimeSpinbox1.DT = form.rowset.fields["MyDateTime"].value
          form.DateTimeSpinbox1.refresh()
          // or whatever you need to set this to -- it will accept
          // a date (value), a datetime (value) or a date object

          *** There may be a bug with null timestamp fields, but
          I may have worked around it, or there may have been
          simply a misunderstanding on my part, and everything is
          now working as it should be. Either way, I believe
          I have resolved the issue with null fields in a table.

          // suggest you use something like the rowset's canSave or
          // onSave event to ensure that the value is saved back
          // to the field:
          form.rowset.fields["MyDateTime"].value = form.DateTimeSpinbox1.DT
             **** although you may be better served to break it apart:
          form.rowset.fields["MyDateTime"].value.setYear( form.DateTimeSpinbox1.DT.year )
          form.rowset.fields["MyDateTime"].value.setMonth( form.DateTimeSpinbox1.DT.month )
          form.rowset.fields["MyDateTime"].value.setDate( form.DateTimeSpinbox1.DT.date )
          form.rowset.fields["MyDateTime"].value.setHours( form.DateTimeSpinbox1.DT.hour )
          form.rowset.fields["MyDateTime"].value.setMinutes( form.DateTimeSpinbox1.DT.minute )
          form.rowset.fields["MyDateTime"].value.setSeconds( form.DateTimeSpinbox1.DT.seconds )

       **** When assigning a new value to this control, you should
       call the Refresh() method:
          form.DateTimeSpinbox1.refresh()

       Optional Properties:
          It should be noted that these options modify 
          the display in the DateTimeSpinbox control, 
          but do NOT modify your system defaults. Hence,
          if you set the date format to German, and 
          use the default MARK that is generated when
          that occurs ("."), but your setup for dBASE
          or your application defaults to a date format 
          of, say, American or MDY, then the MARK will
          not be the German "." when you query the value
          as shown below ... it will be whatever your system
          is set for.

          form.DateTimeSpinbox1.DateFormat = "GERMAN"
             this uses the standard values for SET DATE 
             as noted in the help for dBASE. If you leave
             this off, it will use the default for your computer.
          form.DateTimeSpinbox1.Mark = "-"
             If not set, uses the system default 'Mark'
             character from SET MARK ... if you wish to
             use a MARK that is not the same as the default
             for a specific language, you may want to set
             this. For example, if you want to use GERMAN
             as date format, but do not wish to use "." 
             as the default MARK, you would set the Mark here.
             *** It should be noted that if you use the
             DateFormat option, the MARK will be
             changed, so if you have your own custom
             separator you prefer to use, then it is
             a good idea to use the MARK property
             here to be sure that is what is used.
          form.DateTimeSpinbox1.Hours = 24 (or 12)
             default is machine setting (SET HOURS)
          form.DateTimeSpinbox1.AMPM = "am,pm"
             default is "AM,PM", this string is used for
             12 hour time -- if you set this and 
             HOURS is set to 24, or you don't set HOURS
             manually and it defaults to 24, this will
             be ignored. There can only be two options,
             one for before noon (AM) and one for 
             after noon (PM). These options must be
             set as a string (in quotes) and separated
             by a comma (,). 
             *** NOTE that this assumes that your options 
             will be two characters for each (i.e.,
             "AM,PM" or "am,pm"), or four characters
             for each (i.e., "A.M.,P.M." or "a.m.,p.m.")
             and that if four, you are using a period/dot
             after the letter(s). 
          form.DateTimeSpinbox1.DefaultDate = {01/01/2006} 
             The default date to use if the value
             passed to the object is a null value.
             If no value is given, today's date will
             be used with a time of midnight.
             // of course you should use the date format
             // that works for your settings -- thie
             // value {01/01/2006} shown above is American
             // date format for January 1, 2006 ...

             Note that I can only trap for so many errors,
             and while {00/00/0000 00:00:00} is an 
             "empty" datetime value, and I can trap for
             that, I cannot trap for an invalid datetime
             for the most part. Please use your own caution ...

        Display Colors:
          // color normal affects container, text and entryfields
          form.DateTimeSpinbox1.colNormal = "windowtext/window"
          // color highlight affects entryfields as they get focus
          form.DateTimeSpinbox1.colHigh = "WindowText/0x80ffff"
             // note that if you use the default Windows
             // highlight colors, it will be less obvious 
             // when you've selected the values in the entryfields.
             // (this color string is "highlighttext/highlight")

          That's it for now. I have considered perhaps an
          option to allow for setting Speedtips for the
          entryfields to different languages ... anyone?
          (At the moment there are no speedtips defined)

    Wishlist:
        Better images for up/down buttons ... 

    Dependencies:
       :dUFLP:DateEx.cc
       :dUFLP:Time.cc
       :dUFLP:Up.bmp
       :dUFLP:Down.bmp
*/
class DateTimeSpinbox( oParent ) of Container( oParent ) custom
   protect Init, AMPMChange, AMPMEF_onChange, DateEF_oNChange,;
           DayChange, DownBtn_onClick, HourChange, HourEF_onChange,;
           MinuteChange, MinuteEF_onChange, MonthChange, SecondChange,;
           SecondEF_onChange, UpBtn_onClick, ValidDay, ValidMonth,;
           ValidYear, YearChange, RefreshDate, RefreshTime
   protect this.AM, this.PM

   with (this)
      onOpen = class::Init
      metric = 6 // Pixels
      width = 198.0
      height = 22.0
      colorNormal = "Window"
      borderStyle = 4	// Single
   endwith

   this.DATEPT1EF = new ENTRYFIELD(this)
   with (this.DATEPT1EF)
      onChange = class::DATEEF_onChange
      selectAll = true
      height = 14.0
      left = 1.0
      top = 2.0
      width = 31.0
      picture = "9999"
      function = "JL"
      maxLength = 4
      pageno = 0
      borderStyle = 3	// None
   endwith

   this.DATESEP1 = new TEXTLABEL(this)
   with (this.DATESEP1)
      height = 19.0
      left = 32.0
      top = 2.0
      width = 6.0
      text = "/"
      colorNormal = "WindowText/Window"
      pageno = 0
   endwith

   this.DATEPT2EF = new ENTRYFIELD(this)
   with (this.DATEPT2EF)
      onChange = class::DATEEF_onChange
      height = 14.0
      left = 36.0
      top = 2.0
      width = 16.0
      picture = "99"
      function = "JL"
      maxLength = 2
      pageno = 0
      borderStyle = 3	// None
   endwith

   this.DATESEP2 = new TEXTLABEL(this)
   with (this.DATESEP2)
      height = 19.0
      left = 54.0
      top = 2.0
      width = 6.0
      text = "/"
      colorNormal = "WindowText/Window"
      pageno = 0
   endwith

   this.DATEPT3EF = new ENTRYFIELD(this)
   with (this.DATEPT3EF)
      onChange = class::DATEEF_onChange
      height = 14.0
      left = 58.0
      top = 2.0
      width = 31.0
      picture = "9999"
      function = "JL"
      maxLength = 4
      pageno = 0
      borderStyle = 3	// None
   endwith

   this.HOUREF = new ENTRYFIELD(this)
   with (this.HOUREF)
      onChange = class::HOUREF_ONCHANGE
      height = 14.0
      left = 93.0
      top = 2.0
      width = 16.0
      picture = "99"
      function = "L99"
      value = "99"
      maxLength = 2
      pageno = 0
      borderStyle = 3	// None
   endwith

   this.TIMESEP1 = new TEXTLABEL(this)
   with (this.TIMESEP1)
      height = 19.0
      left = 110.0
      top = 1.0
      width = 4.0
      text = ":"
      colorNormal = "WindowText/Window"
      pageno = 0
   endwith

   this.MINUTEEF = new ENTRYFIELD(this)
   with (this.MINUTEEF)
      onChange = class::MINUTEEF_ONCHANGE
      height = 14.0
      left = 113.0
      top = 2.0
      width = 16.0
      picture = "99"
      function = "L99"
      value = "99"
      maxLength = 2
      pageno = 0
      borderStyle = 3	// None
   endwith

   this.TIMESEP2 = new TEXTLABEL(this)
   with (this.TIMESEP2)
      height = 19.0
      left = 130.0
      top = 1.0
      width = 4.0
      text = ":"
      colorNormal = "WindowText/Window"
      pageno = 0
   endwith

   this.SECONDEF = new ENTRYFIELD(this)
   with (this.SECONDEF)
      onChange = class::SECONDEF_ONCHANGE
      height = 14.0
      left = 133.0
      top = 2.0
      width = 16.0
      picture = "99"
      function = "L99"
      value = "99"
      maxLength = 2
      pageno = 0
      borderStyle = 3	// None
   endwith

   this.AMPMEF = new ENTRYFIELD(this)
   with (this.AMPMEF)
      onChange = class::AMPMEF_ONCHANGE
      height = 16.0
      left = 152.0
      top = 2.0
      width = 22.0
      value = "PM"
      maxLength = 2
      pageno = 0
      borderStyle = 3	// None
   endwith

   this.UPBTN = new PUSHBUTTON(this)
   with (this.UPBTN)
      onClick = class::UPBTN_ONCLICK
      height = 10.0
      left = 176.0
      top = 0
      width = 21.0
      text = ""
      speedBar = true
      upBitmap = "filename :dUFLP:up.bmp"
      pageno = 0
   endwith

   this.DOWNBTN = new PUSHBUTTON(this)
   with (this.DOWNBTN)
      onClick = class::DOWNBTN_ONCLICK
      height = 10.0
      left = 176.0
      top = 10.0
      width = 21.0
      text = ""
      speedBar = true
      upBitmap = "filename :dUFLP:down.bmp"
      pageno = 0
   endwith

   function Init
      // fontSize issues for users with LargeFonts versus
      // "normal" ...
      if SmallFonts()
? "Small Fonts"
         this.AMPMEF.fontSize    := 10
         this.DatePT1EF.fontSize := 10
         this.DatePT2EF.fontSize := 10
         this.DatePT3EF.fontSize := 10
         this.DateSep1.fontSize  := 10
         this.DateSep2.fontSize  := 10
         this.HourEF.fontSize    := 10
         this.MinuteEF.fontSize  := 10
         this.SecondEF.fontSize  := 10
         this.TimeSep1.fontSize  := 10
         this.TimeSep2.fontSize  := 10
      else
? "Large Fonts"
         this.AMPMEF.fontSize    := 8
         this.DatePT1EF.fontSize := 8
         this.DatePT2EF.fontSize := 8
         this.DatePT3EF.fontSize := 8
         this.DateSep1.fontSize  := 8
         this.DateSep2.fontSize  := 8
         this.HourEF.fontSize    := 8
         this.MinuteEF.fontSize  := 8
         this.SecondEF.fontSize  := 8
         this.TimeSep1.fontSize  := 8
         this.TimeSep2.fontSize  := 8
      endif

      // deal with settings for which EF is Month, Day, Year,
      // and for separators ...
      set procedure to :dUFLP:DateEx.cc additive
      this.DateEX = new DateEx()
      set procedure to :dUFLP:Time.cc additive
      this.Time = new Time()
      //
      // Save settings:
      this.OldDateFormat = set("date")
      this.OldDateMark   = set("mark")
      this.OldHours      = set("hours")

      // if colNormal and colHigh are not defined
      // use the default Windows settings ...
      if type( "this.colNormal" ) == "U"
         this.colNormal = "windowtext/window"
      endif
      if type( "this.colHigh" ) == "U"
         this.colHigh = "WindowText/0x80ffff"
      endif
      // set colors:
      this.colorNormal = right( this.colNormal, ;
                        len( this.colNormal ) - at( "/", this.colNormal ) )
      // text controls:
      this.DateSep1.colorNormal := this.colNormal
      this.DateSep2.colorNormal := this.colNormal
      this.TimeSep1.colorNormal := this.colNormal
      this.TimeSep2.colorNormal := this.colNormal

      // and entryfields get both colNormal and colHighlight:
      this.DatePT1EF.colorNormal    := this.colNormal
      this.DatePT1EF.colorHighlight := this.colHigh
      this.DatePT2EF.colorNormal    := this.colNormal
      this.DatePT2EF.colorHighlight := this.colHigh
      this.DatePT3EF.colorNormal    := this.colNormal
      this.DatePT3EF.colorHighlight := this.colHigh
      this.HourEF.colorNormal       := this.colNormal
      this.HourEF.colorHighlight    := this.colHigh
      this.MinuteEF.colorNormal     := this.colNormal
      this.MinuteEF.colorHighlight  := this.colHigh
      this.SecondEF.colorNormal     := this.colNormal
      this.SecondEF.colorHighlight  := this.colHigh
      this.AMPMEF.colorNormal       := this.colNormal
      this.AMPMEF.colorHighlight    := this.colHigh
      // end of color processing
      if type( "this.DT" ) == "U" and;
         type( "this.DT" ) # "DT" and;
         type( "this.DT" ) # "D"  and;
         type( "this.DT" ) # "O" // date object
         this.DT = datetime()
      endif

      // if the value we have been given is empty, 
      // then we need to do something about it,
      // setting the value properly, getting the 
      // correct elements of the date, and so on ...
      if empty( this.DT )
         // Default Date is a "D" type or "Date"
         if type( "this.DefaultDate" ) == "D" 
            this.DT = new Date( year( this.DefaultDate ), ;
                                month( this.DefaultDate )-1, ;
                                day( this.DefaultDate ) )
         // DateTime
         elseif type( "this.DefaultDate" ) == "DT"
            this.DT = new date( this.DefaultDate.year,;
                                this.DefaultDate.month,;
                                this.DefaultDate.date,;
                                this.DefaultDate.hour,;
                                this.DefaultDate.minute,;
                                this.DefaultDate.second )
         // No Default Date:
         else // this.DefaultDate == "U"
            this.DT = new date( year( date() ),;
                                month( date()-1 ),;
                                day( date() ) )
         endif
      else // We have a value -- just to be sure
           // we get everything:
           this.DT = new date( this.DT.year,;
                               this.DT.month,;
                               this.DT.date,;
                               this.DT.hour,;
                               this.DT.minute,;
                               this.DT.second )
      endif // empty( this.DT )

      // DateFormat should be the standards set up
      // in SET DATE in the online help
      if type( "this.DateFormat" ) # "C"
         this.DateFormat = set("DATE")
      endif
      this.DateFormat = this.DateFormat.toUpperCase()
      if not (this.DateFormat $ "MDY, AMERICAN, USA, ANSI, BRITISH, FRENCH, DMY, GERMAN, ITALIAN, JAPAN, YMD" )
         this.DateFormat = set("DATE")
      endif
      cCmd = [set date ]+this.DateFormat
      &cCmd.

      // get separator character between elements of the date:
      if type( "this.Mark" ) # "C"
         this.Mark = set("MARK" )
      endif
      cCmd = [set mark to "]+this.Mark+["]
      &cCmd.

      if type( "this.Hours" ) # "N" or;
         (this.Hours # 12 and this.Hours # 24)
         this.Hours = set("Hours")
      endif
      cCmd = [set hours to ]+this.Hours
      &cCmd.

      if this.Hours == 24 
         // don't display the AM/PM indicator,
         // bring the buttons left, and 
         // shorten the width of the container
         local nWidth
         nWidth = this.AMPMEF.width
         this.AMPMEF.visible  := false
         this.UpBtn.left      := this.UpBtn.left - nWidth
         this.DownBtn.left    := this.DownBtn.left - nWidth
         this.width           := this.width - nWidth
      endif

      // if we have 12 hour time, what
      // do we display? Ivar B. Jessen
      // suggested some code, but the
      // problem is that in some cases
      // the values are empty in the output
      // from his code, which doesn't 
      // necessarily mean that we are 
      // doing 12 hour time or 24 hour. 
      // Option: allow the developer
      // to handle this, and use default
      // of AM/PM if nothing else.
      if this.Hours == 12
         if type( "this.AMPM" ) == "C" and ;
            "," $ this.AMPM
            // do nothing, I guess.
         else
            this.AMPM = "AM,PM"
         endif
         // now we have the string ...:
         this.AM = lTrim(rTrim( left( this.AMPM, at( ",", this.AMPM ) - 1 ) ) )
         this.PM = lTrim(rTrim( right( this.AMPM, rat(",", this.AMPM ) - 1 ) ) )

         // set width of AMPMEF entryfield object,
         // and if necessary move buttons right and
         // widen container ...
         if len( this.AM ) == 4 // we need to modify the width
            local nWidthChg, nNewWidth
            nNewWidth = 28
            nWidthChg = nNewWidth-this.AMPMEF.width
            this.AMPMEF.width := nNewWidth
            this.UpBtn.left   += nWidthChg
            this.DownBtn.left += nWidthChg
            this.width        += nWidthChg
         endif
      endif

      // set what each of the date entryfields are for:
      do case
         case this.DateFormat == "MDY" or;
              this.DateFormat == "AMERICAN" or;
              this.DateFormat == "USA"
              // properties of the Entryfields, "What" is used
              // in code for up/down buttons ...:
              this.DatePT1EF.What = "M"
              this.DatePT2EF.What = "D"
              this.DatePT3EF.What = "Y"

         case this.DateFormat == "ANSI" or;
              this.DateFormat == "JAPAN" or;
              this.DateFormat == "YMD"
              this.DatePT1EF.What = "Y"
              this.DatePT2EF.What = "M"
              this.DatePT3EF.What = "D"

         case this.DateFormat == "BRITISH" or;
              this.DateFormat == "FRENCH" or;
              this.DateFormat == "GERMAN" or;
              this.DateFormat == "ITALIAN" or;
              this.DateFormat == "DMY"

              this.DatePT1EF.What = "D"
              this.DatePT2EF.What = "M"
              this.DatePT3EF.What = "Y"
      endcase

      // set the separator character for the dates:
      this.DateSEP1.Text = this.Mark
      this.DateSEP2.Text = this.Mark

      // deal with setting widths of entryfields for
      // year display and all that. Only the first
      // and third entryfields are affected, as none
      // of the formats sets the Year in the middle:
      nCentury = set("CENTURY") // show all four digits in year?
      nYrWidth = iif( nCentury == "ON", 31, 16 )
      // default width of this entryfield is 31 pixels:
      if this.DatePT1EF.What == "Y"
         this.DatePT1EF.width     := nYrWidth
         this.DatePT1EF.picture   := "9999"
         this.DatePT1EF.maxLength := 4
      else
         this.DatePT1EF.width     := 16
         this.DatePT1EF.picture   := "99"
         this.DatePT1EF.maxLength := 2
      endif

      if this.DatePT1EF.width == 16
         // shift all controls left, shorten width:
         this.DateSep1.left  -= 15
         this.DatePT2EF.left -= 15
         this.DateSep2.left  -= 15
         this.DatePT3EF.left -= 15
         this.HOUREF.left    -= 15
         this.TIMESEP1.left  -= 15
         this.MINUTEEF.left  -= 15
         this.TIMESEP2.left  -= 15
         this.SECONDEF.left  -= 15
         if this.AMPMEF.visible
            this.AMPMEF.left -= 15
         endif
         this.UpBtn.left     -= 15
         this.DownBtn.left   -= 15
         // and the width of the container:
         this.width          -= 15
      endif

      if this.DatePT3EF.What == "Y"
         this.DatePT3EF.width     := nYrWidth
         this.DatePT3EF.picture   := "9999"
         this.DatePT3EF.maxLength := 4
      else
         this.DatePT3EF.width     := 16
         this.DatePT3EF.picture   := "99"
         this.DatePT3EF.maxLength := 2
      endif
      // and we do the same as above, if needed:
      if this.DatePT3EF.width == 16
         // shift all controls left, shorten width:
         this.HOUREF.left    -= 15
         this.TIMESEP1.left  -= 15
         this.MINUTEEF.left  -= 15
         this.TIMESEP2.left  -= 15
         this.SECONDEF.left  -= 15
         if this.AMPMEF.visible
            this.AMPMEF.left -= 15
         endif
         this.UpBtn.left     -= 15
         this.DownBtn.left   -= 15
         // and the width of the container:
         this.width          -= 15
      endif

      // set the date and time:
      class::Refresh()

      // reset:
      cCmd = [set date ]+this.OldDateFormat
      &cCmd.
      cCmd = [set mark to "]+this.OldDateMark+["]
      &cCmd.
      cCmd = [set hours to ]+this.OldHours
      &cCmd.
   return

   function UPBTN_onClick
      // check to see which control had focus:
      local oObj
      oObj = form.ActiveControl
      // don't do anything if the object that has focus
      // is not one of the ones belonging to this control:
      if oObj.parent # this.parent
         return
      endif

      if oObj.name == "DATEPT1EF" or ;
         oObj.name == "DATEPT2EF" or ;
         oObj.name == "DATEPT3EF"
         if oObj.What == "M"
            class::MonthChange(1)
         elseif oObj.What == "D"
            class::DayChange(1)
         else // oObj.What == "Y"
            class::YearChange(1)
         endif
      endif

      if oObj.name == "HOUREF"
         class::HourChange(1)
      endif

      if oObj.name == "MINUTEEF"
         class::MinuteChange(1)
      endif

      if oObj.name == "SECONDEF"
         class::SecondChange(1)
      endif

      if oObj.name == "AMPMEF"
         class::AMPMChange() // no parameter necessary
      endif
   return

   function MonthChange( nChg )
      // rather than trying to do all the data validation
      // in the universe, what we're going to do is
      // use the DateEX class in the dUFLP and let it
      // handle a lot of the details:
      this.parent.DT = this.parent.DateEx.addMonths( this.parent.DT, nChg, true )
      // now break the month, day and year back out:
      class::RefreshDate()
   return

   function DayChange( nChg )
      this.parent.DT = (this.parent.DT + nChg )
      // now break the month, day and year back out:
      class::RefreshDate()
   return

   function DOWNBTN_onClick
      // check to see which control had focus:
      local oObj
      oObj = form.ActiveControl
      // don't do anything if the object that has focus
      // is not one of the ones belonging to this control:
      if oObj.parent # this.parent
         return
      endif

      if oObj.name == "DATEPT1EF" or ;
         oObj.name == "DATEPT2EF" or ;
         oObj.name == "DATEPT3EF"
         if oObj.What == "M"
            class::MonthChange(-1)
         elseif oObj.What == "D"
            class::DayChange(-1)
         else // oObj.What == "Y"
            class::YearChange(-1)
         endif
      endif

      if oObj.name == "HOUREF"
         class::HourChange(-1)
      endif

      if oObj.name == "MINUTEEF"
         class::MinuteChange(-1)
      endif

      if oObj.name == "SECONDEF"
         class::SecondChange(-1)
      endif

      if oObj.name == "AMPMEF"
         class::AMPMChange() // no parameter necessary
      endif
   return

   function YearChange( nChg )
      // simply change the year, the rest should
      // be automatic, I believe:
      this.parent.DT.setYear( this.parent.DT.year + nChg )
      // now break the month, day and year back out:
      class::RefreshDate()
   return

   function HourChange( nChg )
      nHour = this.parent.DT.hour
      nHour += nChg
      if nHour > 23
         nHour := 0
      endif
      if nHour < 0
         nHour := 23
      endif
      this.parent.DT.setHours( nHour )
      // break the time back out:
      class::RefreshTime()
   return

   function HOUREF_onChange
      // validate hour
      local nMin, nMax
      if this.parent.Hours == 12
         nMin = 1
         nMax = 12
      else
         nMin = 0
         nMax = 23
      endif
      if this.value < nMin or this.value > nMax
         class::RefreshTime()
         return
      endif
      this.parent.DT.setHours( this.value )
      class::RefreshTime()
   return

   function MinuteChange( nChg )
      // rather than trying to deal with all this manually,
      // let's rely on existing code:
      local cTime, nTimeInSeconds
      // Get time values:
      cTime = this.parent.DT.Hour+":"+;
              this.parent.DT.Minute+":"+;
              this.parent.DT.Second
      // convert to seconds
      nTimeInSeconds = this.parent.Time.Time2Sec( cTime )
      // add/subtract:
      nTimeInSeconds += ( nChg * 60 ) // +/- 1 minute ...
      // convert back to time string
      cTime = this.parent.Time.Sec2Time( nTimeInSeconds )
      // pull the values out:
      this.parent.DT.setHours( val( left( cTime,2 ) ) )
      this.parent.DT.setMinutes( val( substr( cTime, 4,5 ) ) )
      this.parent.DT.setSeconds( val( right( cTime, 2 ) ) )

      // break the time back out:
      class::RefreshTime()
   return

   function MINUTEEF_onChange
      // validate Minutes
      if this.value < 0 or this.value > 59
         class::RefreshTime()
         return
      endif
      this.parent.DT.setMinutes( this.value )
      class::RefreshTime()
   return

   function SecondChange( nChg )
      local cTime, nTimeInSeconds
      // Get time values:
      cTime = this.parent.DT.Hour+":"+;
              this.parent.DT.Minute+":"+;
              this.parent.DT.Second
      // convert to seconds
      nTimeInSeconds = this.parent.Time.Time2Sec( cTime )
      // add/subtract:
      nTimeInSeconds += ( nChg ) // +/- 1 second ...
      // convert back to time string
      cTime = this.parent.Time.Sec2Time( nTimeInSeconds )
      // pull the values out:
      this.parent.DT.setHours( val( left( cTime,2 ) ) )
      this.parent.DT.setMinutes( val( substr( cTime, 4,5 ) ) )
      this.parent.DT.setSeconds( val( right( cTime, 2 ) ) )

      // break the time back out:
      class::RefreshTime()
   return

   function SECONDEF_onChange
      // validate Seconds
      if this.value < 0 or this.value > 59
         class::RefreshTime()
         return
      endif
      this.parent.DT.setSeconds( this.value )
      class::RefreshTime()
   return

   function AMPMChange
      // not as simple as it seems, but ...
      local cAMPM
      cAMPM = this.parent.AMPMEF.value.rightTrim()
      if cAMPM == this.parent.AM
         cAMPM := this.parent.PM
         this.parent.DT.setHours( this.parent.DT.Hour+12 )
      else
         cAMPM := this.parent.AM
         this.parent.DT.setHours( this.parent.DT.Hour-12 )
      endif

      // break the time back out:
      class::RefreshTime()
   return

   function AMPMEF_onChange
      // only allow set values:
      if upper(left(this.value,1)) == upper(left(this.parent.AM,1)) or;
         upper(this.value) == upper(this.parent.AM)
         this.value := this.parent.AM
      elseif upper(left(this.value,1)) == upper(left(this.parent.PM,1)) or;
             upper(this.value) == upper(this.parent.PM)
         this.value := this.parent.PM
      else 
         class::RefreshTime()
         return
      endif

      // reset the hour if needed:
      if this.parent.DT.hour > 12 and this.value == this.parent.AM
         this.parent.DT.setHours( this.parent.DT.hour - 12 )
      elseif this.parent.DT.Hour <= 12 and this.value == this.parent.PM
         this.parent.DT.setHours( this.parent.DT.hour + 12 )
      endif
      class::RefreshTime()
   return

   function DATEEF_onChange
      // validate month/day/year
      if this.What == "M"
         if not class::ValidMonth(this.value)
            return
         endif
         // should this handle invalid day of month?
         this.parent.DT.SetMonth( this.value-1 )
         class::RefreshDate()
      elseif this.What == "D"
         if not class::ValidDay(this.value)
            return
         endif
         // should this handle invalid day of month?
         this.parent.DT.SetDate( this.value )
         class::RefreshDate()
      else // "Y"
         if not class::ValidYear(this.value)
            class::RefreshDate()
         endif
      endif
   return

   function ValidMonth( nMonth )
      local bReturn
      bReturn = false

      // if it's invalid, just reset it to current
      if nMonth < 1 or nMonth > 12
         class::RefreshDate()
      else
         bReturn := true
      endif

   return bReturn

   function ValidDay( nDay )
      local bReturn
      bReturn = false
      // if less than 1 we have an invalid month,
      // just reset to current and have done:
      if nDay < 1 
         class::RefreshDate()
         return bReturn
      endif

      // validate the day of month based on the current month:
      nMonth = this.parent.DT.month+1 // month is zero based:

      // get the months with 31 days (January, March, May, July, 
      // August, October, December):
      if nMonth == 1 or;
         nMonth == 3 or;
         nMonth == 5 or;
         nMonth == 7 or;
         nMonth == 8 or;
         nMonth == 10 or;
         nMonth == 12

         if nDay > 31
            class::RefreshDate()
            return bReturn
         endif

             // 30 days (April, June, September, November):
      elseif nMonth == 4 or;
             nMonth == 6 or;
             nMonth == 9 or;
             nMonth == 11

         if nDay > 30
            class::RefreshDate()
            return bReturn
         endif

      else // February!
         // need to determine if the current year is a leap year:
         nMax = 28
         if this.DateEx.isLeap( this.DT.Year )
            nMax = 29
         endif

         if nDay > nMax
            class::RefreshDate()
            return bReturn
         endif

      endif

      // okay, we got here, the day appears to be
      // valid, let's set it:
      this.parent.DT.setDate( nDay )
      class::RefreshDate()
      bReturn = true

   return bReturn

   function ValidYear( nYear )
      local bReturn, nEpoch, nDay, nMonth, dTemp
      bReturn = false
      nDay = this.parent.DT.date
      nMonth = this.parent.DT.month

      dTemp = new date( nYear, nMonth, nDay )
      nYear = dTemp.year

      this.parent.DT.setYear( nYear )
      class::RefreshDate()
      bReturn = true
   return bReturn

   function Refresh
      /*
         This should be the only method the user can modify:
         when assigning/reassigning a value to the object,
         call this method and it will call the RefreshDate
         and RefreshTime methods ...
      */

      // If the value passed is null or empty:
      if empty( this.DT )
         // Default Date is a "D" type or "Date"
         if type( "this.DefaultDate" ) == "D" and ;
            not empty( this.DefaultDate )
            
            this.DT = new Date( year( this.DefaultDate ), ;
                                month( this.DefaultDate )-1, ;
                                day( this.DefaultDate ) )
         // DateTime
         elseif type( "this.DefaultDate" ) == "DT" and;
                not empty( this.DefaultDate )
            this.DT = new date( this.DefaultDate.year,;
                                this.DefaultDate.month,;
                                this.DefaultDate.date,;
                                this.DefaultDate.hour,;
                                this.DefaultDate.minute,;
                                this.DefaultDate.second )
         // No Default Date:
         else // this.DefaultDate == "U"
            this.DT = new date( year( date() ),;
                                month( date()-1 ),;
                                day( date() ) )
         endif
      else // We have a value -- just to be sure
           // we get everything:
           this.DT = new date( this.DT.year,;
                               this.DT.month,;
                               this.DT.date,;
                               this.DT.hour,;
                               this.DT.minute,;
                               this.DT.second )
      endif
      // force this thing to deal with allowing
      // entry by keyboard -- without this, 
      // it doesn't seem to want to work:
      this.DT.setMonth( this.DT.month+1 )
      this.DT.setMonth( this.DT.month-1 )
      // update:
      class::RefreshDate()
      class::RefreshTime()
   return

   function RefreshDate
      // break the date:
      local oThis, nYr, nMon, nDay
      if this.baseClassName == "CONTAINER"
         oThis = this
      else
         oThis = this.parent
      endif

      nYr   = oThis.DT.year
      nMon  = oThis.DT.month
      nDay  = oThis.DT.date

      // deal with setting these up:
      if oThis.DatePT1EF.What == "M"
         oThis.DatePT1EF.value := (nMon+1) // months are zero based with date objects
      elseif oThis.DatePT1EF.What == "D"
         oThis.DatePT1EF.value := nDay
      else
         oThis.DatePT1EF.value := nYr
      endif

      // None of the date formats puts the year in the 
      // center, so ...
      if oThis.DatePT2EF.What == "M"
         oThis.DatePT2EF.value := (nMon+1) // months are zero based with date objects
      elseif oThis.DatePT2EF.What == "D"
         oThis.DatePT2EF.value := nDay
      endif

      if oThis.DatePT3EF.What == "M"
         oThis.DatePT3EF.value := (nMon+1) // months are zero based with date objects
      elseif oThis.DatePT3EF.What == "D"
         oThis.DatePT3EF.value := nDay
      else
         oThis.DatePT3EF.value := nYr
      endif
   return

   function RefreshTime
      // break the time down:
      local oThis, nHour, nMin, nSec
      if this.baseClassName == "CONTAINER"
         oThis = this
      else
         oThis = this.parent
      endif
      nHour = oThis.DT.Hour
      nMin  = oThis.DT.Minute
      nSec  = oThis.DT.Second

      oThis.MinuteEF.value := nMin
      oThis.SecondEF.value := nSec
      // only do the following if this.Hours is 12
      if oThis.Hours == 12
         // 12 hour time
         if nHour > 12
            nHour -= 12 // subtract 12
            oThis.AMPMEF.value := oThis.PM
         else
            oThis.AMPMEF.value := oThis.AM
         endif
         if nHour == 0
            nHour = 12
         endif
      endif
      oThis.HourEF.value   := nHour
   return

endclass
// additional code
Function Smallfonts
/*
   Programmer   : Romain Strieff [dBVIPS]
   Last Modified: 04/20/1999
   Rev. History.: 04/20/1999 -- Original
   Notes........: Returns True if Smallfonts are used
   .............: in Windows menus
   Written for..: Visual dBASE (32-bit versions)
   Usage .......: ? SmallFonts()
*/
   local dc,result
   extern CINT        GetDeviceCaps(cHandle, CINT) gdi32
   extern CINT        ReleaseDC(cHandle, cHandle) user32
   extern cHandle     GetDC(cHandle) user32
   #define LOGPIXELSX    88
   dc=getdc(0)
   result=GetDeviceCaps(dc,LOGPIXELSX)
   ReleaseDc(0,dc)
return result=96
// eof: SmallFonts()

