/*
   ----------------------------------------------------------------
   Description:
 
    This file contains three classes
    Class GETWINCOLOR - this is subclassed from a listbox. 
    	It's more of a development tool than it is a production item.
      The API call GetSysColor() will allow you to retreive colors
      in use during your windows session. The list box when dropped on
      a form provides the different windows color types and once selected
      you can return the actual color usable within your program.
      You could also use this API, inserted in your program to keep 
      your forms or other items consistant with the default windows colors.
		See the Test_List.wfm that came with this zip file

    Class GetWindowColor - This is a class for use if you wished to use a
    	windows color in your program. See first example below.

    Class ConvertColors - is a class that just converts one color format
    	to another. See additional notes in each function within the class
      RGBColor is the format returned by the GetColor() function  ("255,255,255")
      HexColor is the format used within web pages such as ("FFFF80")
      WinColor is the format used by windows API calls such as (16776960)
      dBaseHex is the format returned by the GetColor function with the form 
      	designer such as ("0x80FFFF")
     
    Programmer: 
 
       Rich at www.autotraker.com
 
    History: 
 
       04/29/2002 -- first version
       04/30/2002 -- fixed change line # 276 to this.WinColor = f.WinColor
 
    Methods:
       More details on each example is contained in the 
       commentary for each method and class...
 
      Example for GetWinColor:
        set procedure to GetWinColor.cc additive
        local f
        f = new GetWindowColor()
		  f.GetWinColor(COLOR_DESKTOP)
        f.ConvertWinColor()
        form.ColorNormal = f.dBaseHex
        f.release()

      Example for ConvertColors:
        set procedure to GetWinColor.cc additive
        x = new ConvertColors() 
        x.HexColor = "80FFFF"
        x.ConvertFromHex()
        form.ColorNormal = x.dBaseHex
        x.release()
        x = NULL

        Or you could use the GetColor() function to set your
        programs colors during Runtime
         local oldColor,x
         oldColor = form.colorNormal
         x = getcolor()
         if at(",",x)>0 // color was chosen
            local f
            f = new ConvertColors()
            f.RGBColor = x
            f.ConvertFromRGB()
            form.ENTRYFIELD1.value = f.dBaseHex
            form.ENTRYFIELD2.value = f.RGBColor
            form.ENTRYFIELD3.value = f.HexColor
            form.ENTRYFIELD4.value = f.WinColor
            form.colorNormal = f.dBaseHex
            f.release()
            f = null
         endif

      Example for ConvertColors:
        set procedure to GetWinColor.cc additive
        x = new ConvertColors()
         
         ? "Covert from Windows color interger"
         ? "Win intergers are used in API calls"
         ? "Win2RGBColor      " + x.Win2RGBColor(nColor) 
         ? "Win2HexColor      " + x.Win2HexColor(nColor)
         ? "Win2dBaseHex      " + x.Win2dBaseHex(nColor)

         cRGB = Win2RGBColor(nColor)
         ? " "
         ? "Convert from RGB"
         ? "RGB is the color returned by GetColor()"
         ? "RGB2HexColor      " + x.RGB2HexColor(cRGB)
         ? "RGB2dBaseHex      " + x.RGB2dBaseHex(cRGB)
         ? "RGB2WinColor      " + x.RGB2WinColor(cRGB)

         cdBase = Win2dBaseHex(nColor)
         ? " "
         ? "Convert from dBaseHex"
         ? "dBaseHex is used defining colors used in forms."
         ? "dBaseHex2Hex      " + x.dBaseHex2Hex(cdBase)
         ? "dBaseHex2RGB      " + x.dBaseHex2RGB(cdBase)
         ? "dBaseHex2WinColor " + x.dBaseHex2WinColor(cdBase)

         cHex = Win2HexColor(nColor)
         ? " "
         ? "Convert from Hex"
         ? "Hex colors are the colors used in web pages"
         ? "Hex2RGBColor      " + x.Hex2RGBColor(cHex)
         ? "Hex2dBaseHex      " + x.Hex2dBaseHex(cHex)
         ? "Hex2WinColor      " + x.Hex2WinColor(cHex)
      x.release()
      x = NULL


   ----------------------------------------------------------------
   Copyright (C) 2002 AutoTraker Inc. www.autotraker.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    rich@autotraker.com
   ----------------------------------------------------------------

*/


#define COLOR_SCROLLBAR 0						// The color of a scroll bar's gray area. 
#define COLOR_BACKGROUND 1						// The desktop color. 
#define COLOR_ACTIVECAPTION 2					// The active window title bar color. Windows 98, 2000: The color of the left side of the active window title bar gradient, if the gradient effect is used. 
#define COLOR_INACTIVECAPTION 3				// The color of an inactive window's caption. Windows 98, 2000: The color of the right side of an inactive window's title bar gradient, if the gradient effect is used. 
#define COLOR_MENU 4								// The background color of menus. 
#define COLOR_WINDOW 5							// The background color of a window. 
#define COLOR_WINDOWFRAME 6					// The color of a window frame. 
#define COLOR_MENUTEXT 7						// The color of menu text. 
#define COLOR_WINDOWTEXT 8  					// The color of text in a window. 
#define COLOR_CAPTIONTEXT 9					// The color of window caption text, size boxes, and scroll bar arrow boxes. 
#define COLOR_ACTIVEBORDER 10					// The active window border color. 
#define COLOR_INACTIVEBORDER 11				// The color of an inactive window's border. 
#define COLOR_APPWORKSPACE 12					// The background color of multiple document interface (MDI) windows. 
#define COLOR_HIGHLIGHT 13						// The color used to highlight selected items. 
#define COLOR_HIGHLIGHTTEXT 14				// The color used for the text of highlighted items. 
#define COLOR_BTNFACE 15                  // The color of a button's face.
#define COLOR_BTNSHADOW 16						// The color of a button's shadow area
#define COLOR_GRAYTEXT 17						// The color for disabled (grayed-out) text. 
#define COLOR_BTNTEXT 18                  // The text color for pushbuttons. 
#define COLOR_INACTIVECAPTIONTEXT 19		// The color of an inactive window's caption text. 
#define COLOR_BTNHIGHLIGHT 20             // The color of a button's highlight area

#define COLOR_3DDKSHADOW 21					// The Darkest shadow color for 3-D display elements
#define COLOR_3DLIGHT 22						// The light (opposite of shadow) color for 3D objects. 
#define COLOR_INFOTEXT 23						// The text color for tooltip controls. 
#define COLOR_INFOBK 24							// The background color for tooltop controls. 
#define COLOR_HOTLIGHT 26						// Windows 98, 2000: The color of a hot-tracked item, which is executed with a single click. 
#define COLOR_GRADIENTACTIVECAPTION 27		// Windows 98, 2000: The color of the right side of the active window title bar gradient, if the gradient effect is used. 
#define COLOR_GRADIENTINACTIVECAPTION 28	// Windows 98, 2000: The color of the right side of an inactive window's title bar gradient, if the gradient effect is used. 

#define COLOR_DESKTOP COLOR_BACKGROUND    // The Desktop color
#define COLOR_3DFACE COLOR_BTNFACE			// The face color for 3D objects. 
#define COLOR_3DSHADOW COLOR_BTNSHADOW		// The shadow color for 3D objects. 
#define COLOR_3DHIGHLIGHT COLOR_BTNHIGHLIGHT // The Highlight color for 3-D display elements
#define COLOR_3DHILIGHT COLOR_BTNHIGHLIGHT// The highlight (opposite of shadow) color for 3D objects. 

#define COLOR_BTNHILIGHT COLOR_BTNHIGHLIGHT // The color of a button's highlight area

   class GETWINCOLORS(parentObj, name) of LISTBOX(parentObj, name) custom
   // custom listbox to add in selecting windows colors
      with (this)
         onSelChange = class::GETWINCOLOR_ONSELCHANGE
         sorted = true
         height = 5.30
         aChooseColor = new array()
         GetChoices()
         dataSource = "ARRAY this.aChooseColor"
      endwith
      this.RGBColor = "0,0,0"
      this.dBaseHex = "0x000000"
      this.HexColor = "000000"
      this.WinColor = 0

	function GetWinColor_onSelChange
   	this.SelectChoice(this.value)
      return

	function GetChoices
      this.aChooseColor = new array()
      this.aChooseColor.add("3DDKSHADOW")
      this.aChooseColor.add("3DFACE")
      this.aChooseColor.add("3DHIGHLIGHT")
      this.aChooseColor.add("3DHILIGHT")
      this.aChooseColor.add("3DLIGHT")
      this.aChooseColor.add("3DSHADOW")
      this.aChooseColor.add("ACTIVEBORDER")
      this.aChooseColor.add("ACTIVECAPTION")
      this.aChooseColor.add("APPWORKSPACE")
      this.aChooseColor.add("BACKGROUND")
      this.aChooseColor.add("BTNFACE")
      this.aChooseColor.add("BTNHIGHLIGHT")
      this.aChooseColor.add("BTNHILIGHT")
      this.aChooseColor.add("BTNSHADOW")
      this.aChooseColor.add("BTNTEXT")
      this.aChooseColor.add("CAPTIONTEXT")
      this.aChooseColor.add("DESKTOP")
      this.aChooseColor.add("GRADIENTACTIVECAPTION")
      this.aChooseColor.add("GRADIENTINACTIVECAPTION")
      this.aChooseColor.add("GRAYTEXT")
      this.aChooseColor.add("HIGHLIGHT")
      this.aChooseColor.add("HIGHLIGHTTEXT")
      this.aChooseColor.add("HOTLIGHT")
      this.aChooseColor.add("INACTIVEBORDER")
      this.aChooseColor.add("INACTIVECAPTION")
      this.aChooseColor.add("INACTIVECAPTIONTEXT")
      this.aChooseColor.add("INFOBK")
      this.aChooseColor.add("INFOTEXT")
      this.aChooseColor.add("MENU")
      this.aChooseColor.add("MENUTEXT")
      this.aChooseColor.add("SCROLLBAR")
      this.aChooseColor.add("WINDOW")
      this.aChooseColor.add("WINDOWFRAME")
      this.aChooseColor.add("WINDOWTEXT")
		return this.aChooseColor		

	function SelectChoice(cChoice)
      do case
      	case cChoice = "3DDKSHADOW" ; ChooseColor = COLOR_3DDKSHADOW
      	case cChoice = "3DFACE" ; ChooseColor = COLOR_3DFACE
      	case cChoice = "3DHIGHLIGHT" ; ChooseColor = COLOR_3DHIGHLIGHT
      	case cChoice = "3DHILIGHT" ; ChooseColor = COLOR_3DHILIGHT
      	case cChoice = "3DLIGHT" ; ChooseColor = COLOR_3DLIGHT
      	case cChoice = "3DSHADOW" ; ChooseColor = COLOR_3DSHADOW
      	case cChoice = "ACTIVEBORDER" ; ChooseColor = COLOR_ACTIVEBORDER
      	case cChoice = "ACTIVECAPTION" ; ChooseColor = COLOR_ACTIVECAPTION
      	case cChoice = "APPWORKSPACE" ; ChooseColor = COLOR_APPWORKSPACE
      	case cChoice = "BACKGROUND" ; ChooseColor = COLOR_BACKGROUND
      	case cChoice = "BTNFACE" ; ChooseColor = COLOR_BTNFACE
      	case cChoice = "BTNHIGHLIGHT" ; ChooseColor = COLOR_BTNHIGHLIGHT
      	case cChoice = "BTNHILIGHT" ; ChooseColor = COLOR_BTNHILIGHT
      	case cChoice = "BTNSHADOW" ; ChooseColor = COLOR_BTNSHADOW
      	case cChoice = "BTNTEXT" ; ChooseColor = COLOR_BTNTEXT
      	case cChoice = "CAPTIONTEXT" ; ChooseColor = COLOR_CAPTIONTEXT
      	case cChoice = "DESKTOP" ; ChooseColor = COLOR_DESKTOP
      	case cChoice = "GRADIENTACTIVECAPTION" ; ChooseColor = COLOR_GRADIENTACTIVECAPTION
      	case cChoice = "GRADIENTINACTIVECAPTION" ; ChooseColor = COLOR_GRADIENTINACTIVECAPTION
      	case cChoice = "GRAYTEXT" ; ChooseColor = COLOR_GRAYTEXT
      	case cChoice = "HIGHLIGHT" ; ChooseColor = COLOR_HIGHLIGHT
      	case cChoice = "HIGHLIGHTTEXT" ; ChooseColor = COLOR_HIGHLIGHTTEXT
      	case cChoice = "HOTLIGHT" ; ChooseColor = COLOR_HOTLIGHT
      	case cChoice = "INACTIVEBORDER" ; ChooseColor = COLOR_INACTIVEBORDER
      	case cChoice = "INACTIVECAPTION" ; ChooseColor = COLOR_INACTIVECAPTION
      	case cChoice = "INACTIVECAPTIONTEXT" ; ChooseColor = COLOR_INACTIVECAPTIONTEXT
      	case cChoice = "INFOBK" ; ChooseColor = COLOR_INFOBK
      	case cChoice = "INFOTEXT" ; ChooseColor = COLOR_INFOTEXT
      	case cChoice = "MENU" ; ChooseColor = COLOR_MENU
      	case cChoice = "MENUTEXT" ; ChooseColor = COLOR_MENUTEXT
      	case cChoice = "SCROLLBAR" ; ChooseColor = COLOR_SCROLLBAR
      	case cChoice = "WINDOW" ; ChooseColor = COLOR_WINDOW
      	case cChoice = "WINDOWFRAME" ; ChooseColor = COLOR_WINDOWFRAME
      	case cChoice = "WINDOWTEXT" ; ChooseColor = COLOR_WINDOWTEXT
         endcase
         local f
         f = new GetWindowColor()
			f.GetWinColor(ChooseColor)
         f.ConvertWinColor()
         this.RGBColor = f.RGBColor
         this.dBaseHex = f.dBaseHex
         this.HexColor = f.HexColor
         this.WinColor = f.WinColor
         f.release()
         f = NULL
		return 

		function release;release object this;return true

endclass

class GetWindowColor
	// Class to get the color of a Windows element
   this.RGBColor = "0,0,0"
   this.dBaseHex = "0x000000"
   this.HexColor = "000000"
   this.WinColor = 0
	this.Init()

   function Init
      if type("GetSysColor") # "FP"
         extern culong GetSysColor(cint) user32
      endif
		return

	function GetWinColor(nColor)
   	// get the WinColor via the API call
      this.WinColor = int(GetSysColor(nColor))
      return

	function ConvertWinColor
		// convert the WinColor to some other usable formats
      local x
      x = new ConvertColors() 
      this.RGBColor = x.Win2RGBColor(this.WinColor)
      this.dBaseHex = x.Win2dBaseHex(this.WinColor)
      this.HexColor = x.Win2HexColor(this.WinColor)
      x.release()
      x = NULL
      return

	function release;release object this;return true

endclass

class ConvertColors
   // set one of the below colors and then use the below fuctions
   // to fill the the remainder of the colors
   this.RGBColor = "0,0,0"
   this.dBaseHex = "0x000000"
   this.HexColor = "000000"
   this.WinColor = 0

   function ConvertFromdBaseHex
   	// convert a dBaseHex color to other usable formats
   	this.RGBColor = this.dBaseHex2RGB(this.dBaseHex)
      this.HexColor = this.dBaseHex2Hex(this.dBaseHex)
      this.WinColor = this.dBaseHex2WinColor(this.dBaseHex)
		return true

   function ConvertFromRGB
   	// convert a RGB color to other usable formats
      this.dBaseHex = this.RGB2dBaseHex(this.RGBColor)
      this.HexColor = this.RGB2HexColor(this.RGBColor)
      this.WinColor = this.RGB2WinColor(this.RGBColor)
		return true

   function ConvertFromWinColor
   	// convert a WinColor to other usable formats
   	this.RGBColor = this.Win2RGBColor(this.WinColor) 
      this.dBaseHex = this.Win2dBaseHex(this.WinColor)
      this.HexColor = this.Win2HexColor(this.WinColor)
		return true

   function ConvertFromHex
   	// convert a hex color to other usable formats
   	this.RGBColor = this.Hex2RGBColor(this.HexColor)
      this.dBaseHex = this.Hex2dBaseHex(this.HexColor)
      this.WinColor = this.Hex2WinColor(this.HexColor)
		return true

   // the below functions are for use in the above functions or
   // could be called directly - see the third example in the header
   // of this file

   // Get RGB from Windows color interger. Used in functions below
   function REDVAL(x);return int(bitand(x,0x0000FF))
   function GREENVAL(x);return int(bitrshift(bitand(x,0x00FF00), 8))
   function BLUEVAL(x);return int(bitrshift(bitand(x,0xFF0000), 16))

   // Covert from Windows color interger. Used in API calls
   function Win2RGBColor(x);return ""+this.REDVAL(x) +","+ this.GREENVAL(x) +","+ this.BLUEVAL(x)
   function Win2HexColor(x);return itoh(this.REDVAL(x),2) + itoh(this.GREENVAL(x),2) + itoh(this.BLUEVAL(x),2)
   function Win2dBaseHex(x);return "0x"+itoh(this.BLUEVAL(x),2) + itoh(this.GREENVAL(x),2) + itoh(this.REDVAL(x),2)

   // Convert from RGB  // same as getColor()
   function RGB2HexColor(x);return itoh(val(x),2) + itoh(val(substr(x,at(',',x)+1)),2) + itoh(val(substr(x,rat(",",x)+1)),2)
   function RGB2dBaseHex(x);return "0x"+itoh(val(substr(x,rat(",",x)+1)),2) + itoh(val(substr(x,at(',',x)+1)),2) + itoh(val(x),2)
   function RGB2WinColor(x);return int(val(x) + bitlshift(val(substr(x,at(',',x)+1)),8) + bitlshift(val(substr(x,rat(",",x)+1)),16))

   // Convert from dBaseHex // getColor() in form designer
   function dBaseHex2Hex(x);return substr(x,7,2)+substr(x,5,2)+substr(x,3,2)
   function dBaseHex2RGB(x);return ""+htoi(substr(x,7,2))+","+htoi(substr(x,5,2))+","+htoi(substr(x,3,2))
   function dBaseHex2WinColor(x);return this.RGB2WinColor(this.dBaseHex2RGB(x))

   // Convert from Hex // same as internet hex colors
   function Hex2RGBColor(x);return ""+htoi(substr(x,1,2))+","+htoi(substr(x,3,2))+","+htoi(substr(x,5,2))
   function Hex2dBaseHex(x);return "0x"+substr(x,5,2)+substr(x,3,2)+substr(x,1,2) 
   function Hex2WinColor(x);return this.RGB2WinColor(this.Hex2RGBColor(x))

   function release;release object this;return true
endclass
