//Klasse zum Kopieren von dBASE 7 Tabellen in Level 5 bzw. 4 Format
//
//Programmierung: Lutz Conrad
//                Bermannstrae 19
//                D-15345 Eggersdorf
//                Telefon : 03341 422852
//                Fax     : 03341 422853
//                Funk    : 0172/3249022
//                EMail   : lutz.conrad@t-online.de
//                Internet: http://www.signsoft.com/dbase
//
//Verwendung:     Erstellung von Level 4 und 5 Tabellenstrukturen ohne Zugriff auf die BDE.
//
//                Anlegen leerer Tabellenstrukturen
//                <oRef>=new LC_DBF5()
//                <oRef>.init()
//                <oRef>.aStructure.add({<cFeldname>,<cFeldtyp>,<nFeldlnge>,<nDezimalstellen>})
//                ....
//                <oRef>.BuildBDF5(<cZieldatei>,[<nSprachtreiber>])
//
//                Kopieren von dBASE7 nach dBASE 4/5
//                <oRef>=new LC_DBF5()
//                [aFields=new ARRAY()]
//                [aFields.add(<Feldname DBF7>)]
//                [aFields.add({<Feldname DBF7>,[<Feldname DBF5>]})
//                ....
//                <oRef>.Copy2DBF5(cSource,cDestination,[lCopy,[aFields]])
//                
//                Neben den Namen der Quell- und Zieldatei, die nicht identisch sein drfen,
//                kann hier ein Array der zu kopierenden Felder angegeben werden.
//                Die Elemente enthalten Informationen fr jedes zu bernehmende Feld,
//                ist das Element vom Typ Zeichen, wird es als zu bernehmender Feldname interpretiert,
//                ist das Element vom Typ Array, wird das erste Element des Subarrays als Quellfeld,
//                das zweite Element als Zielfeld interpretiert.
//                Ist lCopy=true, werden alle Datenstze der Ausgangstabelle in die Zieltabelle geschrieben,
//                ist lCopy nicht angegeben oder false, wird nur die Zielstruktur angelegt.
//                Wird kein Array bergeben, so werden alle Felder der Quelldatei in die Zieldatei bernommen.
//                Quellfeld und Zielfeld sind stets vom gleichen Feldtyp. Fr die Feldumwandlung gelten
//                folgende Ausnahmen:
//                Zhler      -> Numerisch
//                Long        -> Numerisch
//                Double      -> Numerisch
//                Zeitstempel -> Datum (ohne Zeit)
//                OLE-Felder werden nur bernommen, wenn Sie keine Verknpfungen enthalten.
CLASS LC_DBF5
   FUNCTION Init
   //Erstellt das Strukturarray fr die Zieltabelle
   this.aStructure=new ARRAY()
   RETURN

   FUNCTION BuildDBF5(cFilename,nDriver)
   //Erstellt die Tabellenstruktur
   //cFilename -> Name der Zieldatei; ist diese nicht angegeben wird 'DateiDBF.DBF' vorgegeben
   //nDriver   -> ID des Sprachtreibers; ist diese nicht angegeben, wird der Sprachtreiber der BDE verwendet
   LOCAL nFields
   LOCAL nHandle
   LOCAL nRecordSize
   LOCAL lMemo
   LOCAL nHeader
   LOCAL cDriver
   LOCAL x,y
   LOCAL cMemoFile
   LOCAL nCheck
   LOCAL lOk

   lOk=true
   DO CASE
      CASE EMPTY(cFilename)
         cFilename="DateiDBF.DBF"
      CASE UPPER(RIGHT(cFilename,4))#".DBF"
         cFilename+=".DBF"
   ENDCASE
   IF EMPTY(nDriver)
      nDriver=CLASS::GetLangDriver()
   ENDIF
   IF TYPE("this.aStructure")=="A" .AND. this.aStructure.size>0
      nFields=this.aStructure.size
      //Arrays definieren
      aFileHeader =new ARRAY(32)
      aFieldHeader=new ARRAY(nFields,32)
      //Ausgangsbedingungen herstellen
      aFileHeader.fill(0)
      aFieldHeader.fill(0)
      nRecordSize=1 //Lschflag immer bercksichtigen
      lMemo=false
      nHeader=0
      //Info fr Felder zusammenstellen
      FOR X=1 TO nFields
         aFieldHeader[X,1]=ASC(SUBSTR(this.aStructure[X][1],1,1))
         aFieldHeader[X,2]=ASC(SUBSTR(this.aStructure[X][1],2,1))
         aFieldHeader[X,3]=ASC(SUBSTR(this.aStructure[X][1],3,1))
         aFieldHeader[X,4]=ASC(SUBSTR(this.aStructure[X][1],4,1))
         aFieldHeader[X,5]=ASC(SUBSTR(this.aStructure[X][1],5,1))
         aFieldHeader[X,6]=ASC(SUBSTR(this.aStructure[X][1],6,1))
         aFieldHeader[X,7]=ASC(SUBSTR(this.aStructure[X][1],7,1))
         aFieldHeader[X,8]=ASC(SUBSTR(this.aStructure[X][1],8,1))
         aFieldHeader[X,9]=ASC(SUBSTR(this.aStructure[X][1],9,1))
         aFieldHeader[X,10]=ASC(SUBSTR(this.aStructure[X][1],10,1))
         //11. Stelle ist immer 0
         aFieldHeader[X,12]=ASC(this.aStructure[X][2]) //Feldtyp
         //13. Stelle ist immer 0
         //14. Stelle ist immer 0
         //15. Stelle ist immer 0
         //16. Stelle ist immer 0
         aFieldHeader[X,17]=this.aStructure[X][3] //Feldlnge
         aFieldHeader[X,18]=this.aStructure[X][4] //Dezimalstellen
         //19. -32. Stelle Stelle ist immer 0
         nRecordSize+=aFieldHeader[X,17]
         lMemo=lMemo .OR. CHR(aFieldHeader[X,12])$"MGB"
      NEXT
      //Info fr Dateikopf zusammenstellen
      aFileHeader[1]=IIF(lMemo,HTOI("8B"),HTOI("03")) //Versionsnummer
      aFileHeader[2]=YEAR(DATE())-1900                //Jahr
      aFileHeader[3]=MONTH(DATE())                    //Monat
      aFileHeader[4]=DAY(DATE())                      //Tag
      //5.-8.Stelle Anzahl der Datenstze; hier immer 0
      nHeader=(nFields+1)*32+1                        //Anzahl der Bytes im Dateikopf
      aFileHeader[9]=MOD(nHeader,256)
      aFileHeader[10]=INT(nHeader/256)
      aFileHeader[11]=MOD(nRecordSize,256)
      aFileHeader[12]=INT(nRecordSize/256)
      //13.-14.Stelle immer 0
      //15. Stelle Transaktionsflag; hier immer 0
      //16. Stelle Verschlsselungsflag; hier immer 0
      //17. - 28.Stelle Netzwerkumgebung; hier immer 0
      //29. Stelle Indexflag fr MDX; hier immer 0
      aFileHeader[30]=nDriver
      nCheck=nFields*56+33
      aFileHeader[31]=MOD(nCheck,256)
      aFileHeader[32]=INT(nCheck/256)
      //Tabelle erstellen
      TRY
         nHandle=FCREATE(cFilename)
         FOR X=1 TO 32
            FWRITE(nHandle,CHR(aFileHeader[X]))
         NEXT
         FOR X=1 TO nFields
            FOR Y=1 TO 32
               FWRITE(nHandle,CHR(aFieldHeader[X,Y]))
            NEXT
         NEXT
         FWRITE(nHandle,CHR(13))
         FWRITE(nHandle,CHR(26))
         FCLOSE(nHandle)
         IF lMemo //Memodatei wird bentigt
            TRY
               //Dateiheader fr Memo-Datei erzeugen
               cMemoFile=LEFT(cFilename,LEN(cFilename)-4)
               nHandle=FCREATE(cMemoFile+".DBT")
               FWRITE(nHandle,CHR(1))
               FWRITE(nHandle,CHR(0))
               FWRITE(nHandle,CHR(0))
               FWRITE(nHandle,CHR(0))
               FWRITE(nHandle,CHR(0))
               FWRITE(nHandle,CHR(0))
               FWRITE(nHandle,CHR(0))
               FWRITE(nHandle,CHR(0))
               FOR X=1 TO 8
                  FWRITE(nHandle,CHR(ASC(SUBSTR(cMemofile,X,1))))
               NEXT
               FWRITE(nHandle,CHR(0))
               FWRITE(nHandle,CHR(0))
               FWRITE(nHandle,CHR(2))
               FWRITE(nHandle,CHR(1))
               FWRITE(nHandle,CHR(0))
               FWRITE(nHandle,CHR(2))
               FWRITE(nHandle,CHR(0))
               FWRITE(nHandle,CHR(0))
               FCLOSE(nHandle)
            CATCH (exception e)
               lOk=false
               MSGBOX("Memo-Datei konnte nicht erstellt werden.","Achtung",16)
            ENDTRY
         ENDIF
      CATCH (exception e)
         lOk=false
         MSGBOX("Zieldatei konnte nicht erstellt werden.","Achtung",16)
      ENDTRY
   ELSE
      lOk=false
      MSGBOX("Fehlerhafte Strukturangaben","Achtung",16)
   ENDIF
   RETURN lOk

   FUNCTION GetLangDriver
   //Ermitteln der Sprachtreiber-ID aus den aktuellen BDE-Einstellungen
   LOCAL cDriver
   LOCAL nDriver

   cDriver=UPPER(LDRIVER())
   DO CASE
      CASE cDriver=="ASCII"
         nDriver=1
      CASE cDriver=="INTL"
         nDriver=2
      CASE cDriver=="JAPAN"
         nDriver=3
      CASE cDriver=="NORDAN"
         nDriver=4
      CASE cDriver=="NORDAN40"
         nDriver=5
      CASE cDriver=="SWEDFIN"
         nDriver=6
      CASE cDriver=="DB865DA0"
         nDriver=8
      CASE cDriver=="DB437NL0"
         nDriver=9
      CASE cDriver=="DB850NL0"
         nDriver=10
      CASE cDriver=="DB437FI0"
         nDriver=11
      CASE cDriver=="DB437FR0"
         nDriver=13
      CASE cDriver=="DB850FR0"
         nDriver=14
      CASE cDriver=="DB437DE0"
         nDriver=15
      CASE cDriver=="DB850DE0"
         nDriver=16
      CASE cDriver=="DB437IT0"
         nDriver=17
      CASE cDriver=="DB850IT1"
         nDriver=18
      CASE cDriver=="DB932JP0"
         nDriver=19
      CASE cDriver=="DB850ES0"
         nDriver=20
      CASE cDriver=="DB437SV0"
         nDriver=21
      CASE cDriver=="DB850SV1"
         nDriver=22
      CASE cDriver=="DB865NO0"
         nDriver=23
      CASE cDriver=="DB437ES1"
         nDriver=24
      CASE cDriver=="DB437UK0"
         nDriver=25
      CASE cDriver=="DB850UK0"
         nDriver=26
      CASE cDriver=="DB437US0"
         nDriver=27
      CASE cDriver=="DB863CF1"
         nDriver=28
      CASE cDriver=="DB850CF0"
         nDriver=29
      CASE cDriver=="DB852CZ0"
         nDriver=31
      CASE cDriver=="DB867CZ0"
         nDriver=32
      CASE cDriver=="DB852HDC"
         nDriver=34
      CASE cDriver=="DB852PO0"
         nDriver=35
      CASE cDriver=="DB860PT0"
         nDriver=36
      CASE cDriver=="DB850PT0"
         nDriver=37
      CASE cDriver=="DB866RU0"
         nDriver=38
      CASE cDriver=="BLLT1DA0"
         nDriver=39
      CASE cDriver=="BLLT1NL0"
         nDriver=40
      CASE cDriver=="BLLT1FI0"
         nDriver=41
      CASE cDriver=="BLLT1FR0"
         nDriver=42
      CASE cDriver=="BLLT1CA0"
         nDriver=43
      CASE cDriver=="BLLT1DE0"
         nDriver=44
      CASE cDriver=="BLLT1IS0"
         nDriver=45
      CASE cDriver=="BLLT1IT0"
         nDriver=46
      CASE cDriver=="BLLT1NO0"
         nDriver=48
      CASE cDriver=="BLLT1ES0"
         nDriver=49
      CASE cDriver=="BLLT1SV0"
         nDriver=51
      CASE cDriver=="BLLT1UK0"
         nDriver=52
      CASE cDriver=="BLLT1US0"
         nDriver=53
      CASE cDriver=="BLLT1PT0"
         nDriver=54
      CASE cDriver=="DB850US0"
         nDriver=55
      CASE cDriver=="INTL850"
         nDriver=57
      CASE cDriver=="SPANISH"
         nDriver=58
      CASE cDriver=="ICELAND"
         nDriver=59
      CASE cDriver=="ANSIINTL"
         nDriver=60
      CASE cDriver=="ANSII850"
         nDriver=61
      CASE cDriver=="ANSISPAN"
         nDriver=62
      CASE cDriver=="ANSISWFN"
         nDriver=63
      CASE cDriver=="ANSINOR4"
         nDriver=64
      CASE cDriver=="CHINA"
         nDriver=69
      CASE cDriver=="KOREA"
         nDriver=70
      CASE cDriver=="TAIWAN"
         nDriver=71
      CASE cDriver=="THAI"
         nDriver=72
      CASE cDriver=="DB936CN0"
         nDriver=77
      CASE cDriver=="DB949KO0"
         nDriver=78
      CASE cDriver=="DB950TW0"
         nDriver=79
      CASE cDriver=="DB874TH0"
         nDriver=80
      CASE cDriver=="DB932JP1"
         nDriver=86
      CASE cDriver=="DBWINUS0"
         nDriver=87
      CASE cDriver=="DBWINWE0"
         nDriver=88
      CASE cDriver=="DBWINES0"
         nDriver=89
      CASE cDriver=="ACCGEN"
         nDriver=90
      CASE cDriver=="ACCNRDAN"
         nDriver=91
      CASE cDriver=="ACCSWFIN"
         nDriver=92
      CASE cDriver=="ACCJAPAN"
         nDriver=93
      CASE cDriver=="FOXDE437"
         nDriver=94
      CASE cDriver=="FOXNO437"
         nDriver=95
      CASE cDriver=="FOXNO850"
         nDriver=96
      CASE cDriver=="FOXDEWIN"
         nDriver=97
      CASE cDriver=="FOXNOWIN"
         nDriver=98
      CASE cDriver=="CZECH"
         nDriver=100
      CASE cDriver=="IL2CZW"
         nDriver=101
      CASE cDriver=="POLISH"
         nDriver=102
      CASE cDriver=="CYRR"
         nDriver=103
      CASE cDriver=="HUN852DC"
         nDriver=104
      CASE cDriver=="GRCP437"
         nDriver=105
      CASE cDriver=="HEBREW"
         nDriver=107
      CASE cDriver=="SLOVENE"
         nDriver=109
      CASE cDriver=="TURK"
         nDriver=110
      CASE cDriver=="BULGARIA"
         nDriver=113
      CASE cDriver=="BLWINAR0"
         nDriver=116
      CASE cDriver=="ANHEBREW"
         nDriver=117
      CASE cDriver=="ANCZECH"
         nDriver=120
      CASE cDriver=="CZECHW"
         nDriver=121
      CASE cDriver=="CSKAMENW"
         nDriver=122
      CASE cDriver=="CSKAMEN"
         nDriver=123
      CASE cDriver=="ANCZECHW"
         nDriver=124
      CASE cDriver=="NPOLISH"
         nDriver=125
      CASE cDriver=="ANCYRR"
         nDriver=126
      CASE cDriver=="ANHUNDC"
         nDriver=127
      CASE cDriver=="ANGREEK1"
         nDriver=128
      CASE cDriver=="ANSISLOV"
         nDriver=130
      CASE cDriver=="ANTURK"
         nDriver=131
      CASE cDriver=="DBHEBREW"
         nDriver=133
      CASE cDriver=="DB437GR0"
         nDriver=134
      CASE cDriver=="DB852SL0"
         nDriver=135
      CASE cDriver=="DB857TR0"
         nDriver=136
      CASE cDriver=="BGDB868"
         nDriver=142
      CASE cDriver=="ACCGREEK"
         nDriver=145
      CASE cDriver=="ANIL2CZW"
         nDriver=146
      CASE cDriver=="BGPD1251"
         nDriver=148
      CASE cDriver=="FOXCZWIN"
         nDriver=155
      CASE cDriver=="FOXCZ895"
         nDriver=156
      CASE cDriver=="BLROM800"
         nDriver=201
      CASE cDriver=="ORAWE850"
         nDriver=202
      CASE cDriver=="SYDC850"
         nDriver=203
      CASE cDriver=="SYDC437"
         nDriver=204
      CASE cDriver=="DB2ANDEU"
         nDriver=205
      CASE cDriver=="MSSGRWIN"
         nDriver=209
      OTHERWISE
         nDriver=15  //"DB437DE0"
   ENDCASE
   RETURN nDriver

   FUNCTION Copy2DBF5(cSource,cDestination,lCopy,aFields)
   //cSource      Quelldatei
   //cDestination Zieldatei
   //lCopy        Sollen Datenstze kopiert werden?
   //aFields      zu bernehmende Felder
   LOCAL cError
   LOCAL q
   LOCAL nFields
   LOCAL cName
   LOCAL cType
   LOCAL nLen
   LOCAL nDec
   LOCAL X
   LOCAL Y

   cError=""
   DO CASE
      CASE EMPTY(cSource) .OR. EMPTY(cDestination)
         cError="Quell- und Zieldatei mssen angegeben werden."
      CASE UPPER(cSource)==UPPER(cDestination)
         cError="Quell- und Zieldatei drfen nicht identisch sein"
      OTHERWISE
         IF .NOT. TYPE("this.aStructure")=="A"
            CLASS::Init()
         ENDIF
         q=new TABLEDEF()
         q.tablename=cSource
         q.load()
         aSource=new ARRAY()
         aDest  =new ARRAY()
         IF TYPE("aFields")=="A" .AND. aFields.size>0
            //Feldliste wurde bergeben
            nFields=aFields.size
            aSource=new ARRAY()
            aDest  =new ARRAY()
            FOR X=1 TO nFields
               IF TYPE("aFields[X]")=="A"
                  IF aFields[X].size=2
                     aSource.add(aFields[X][1])
                     aDest.add(LEFT(aFields[X][2],10))
                  ELSE
                     aSource.add(aFields[X][1])
                     aDest.add(LEFT(aFields[X][1],10))
                  ENDIF
               ELSE
                  aSource.add(aFields[X])
                  aDest.add(LEFT(aFields[X],10))
               ENDIF
            NEXT
         ELSE
            //Keine Feldliste vorhanden, also alles rein
            nFields=q.fields.size
            aSource=new ARRAY()
            aDest  =new ARRAY()
            FOR X=1 TO nFields
               IF .NOT. q.fields[X].fieldName=="_DBASELOCK"
                  aSource.add(q.fields[X].fieldName)
                  aDest.add(LEFT(q.fields[X].fieldname,10))
               ENDIF
            NEXT
            TRY
               IF q.fields["_DBASELOCK"]="O"
                  nFields--
               ENDIF
            CATCH (exception e)
            ENDTRY
         ENDIF
         //Feldliste auf Namen mit Leerzeichen prfen
         FOR X=1 TO nFields
            DO WHILE " "$aDest[X]
               aDest[X]=STUFF(aDest[X],AT(" ",aDest[X]),1,"")
            ENDDO
         NEXT
         //Feldliste auf doppelte Namen prfen
         lDouble=false
         DO
            nThis=1
            nNext=2
            FOR X=nThis TO nFields-1
               FOR Y=X+1 TO nFields
                  IF UPPER(aDest[X])==UPPER(aDest[Y])
                     aDest[Y]=LEFT(aDest[Y],6)+IIF(VAL(RIGHT(aDest[Y],2))=0,"01",STR(VAL(RIGHT(aDest[Y],2))+1,2,0,"0"))
                     Y=nFields+1
                     X=nFields
                     lDouble=true
                  ENDIF
               NEXT
            NEXT
         UNTIL .NOT. lDouble
         //Feldliste zum Anlegen der Zieldatei aufbauen
         FOR X=1 TO nFields
            cName=aDest[X]
            cType=q.fields[aSource[X]].type
            nLen =q.fields[aSource[X]].length
            nDec =q.fields[aSource[X]].decimalLength
            DO CASE
               CASE cType="OLE"
                  cType="G"
               CASE cType="AUTOINC"
                  cType="N"
                  nLen =20
                  nDec = 0
               CASE cType="DATETIME"
                  cType="D"
                  nLen = 8
                  nDec = 0
               CASE cType="LONG"
                  cType="N"
                  nLen =20
                  nDec = 0
               CASE cType="DOUBLE"
                  cType="N"
                  nLen =20
                  nDec = 0
               OTHERWISE
                  cType=LEFT(cType,1)
            ENDCASE
            this.aStructure.add({cName,cType,nLen,nDec}) 
         NEXT
         IF CLASS::BuildDBF5(cDestination)
            IF .NOT. EMPTY(lCopy)
               CLASS::CopyData(cSource,cDestination)
            ENDIF
         ELSE
            cError="Fehler beim Erstellen der Zieldatei"
         ENDIF
   ENDCASE
   RETURN cError

   FUNCTION CopyData(cSource,cDestination)
   //Datenbernahme aus dBASE 7 in Level 4/5 Tabelle
   LOCAL qSource
   LOCAL qDest
   LOCAL nFields
   LOCAL X
   LOCAL nActual
   LOCAL nRecords
   LOCAL lSafety

   lSafety=SET("SAFETY")="ON"
   SET SAFETY OFF
   qSource=new QUERY()
   qSource.sql="SELECT * FROM '"+cSource+"'"
   qSource.requestLive=false
   qSource.active=true
   qDest=new QUERY()
   qDest.sql="SELECT * FROM "+cDestination
   qDest.active=true
   nRecords=qSource.rowset.count()
   nActual =0
   nFields =aDest.size
   f=new Form()
   f.top=10
   f.Left=20
   f.height=1.5
   f.width = 87
   f.text = "Bearbeitungsfortschritt"
   f.autoCenter=true
   f.Progress1 = new PROGRESS(f)
   f.Progress1.height = 0.5
   f.Progress1.left = 1
   f.Progress1.top = 0.5
   f.Progress1.width = 85
   f.Progress1.value = 0
   f.open()
   DO WHILE .NOT. qSource.rowset.endOfSet
      qDest.rowset.beginAppend()
      FOR X=1 TO nFields
         DO CASE
            CASE qDest.rowset.fields[aDest[X]].type="BINARY"
               //Hier ist ein Umweg ber eine Zwischendatei erforderlich
               qSource.rowset.fields[aSource[X]].copyToFile("XX____XX.XXX")
               qDest.rowset.fields[aDest[X]].replaceFromFile("XX____XX.XXX")
            CASE qDest.rowset.fields[aDest[X]].type="OLE"
               TRY //TRY-Block zum Abfangen von Verknfungen 
                  qSource.rowset.fields[aSource[X]].copyToFile("XX____XX.XXX")
                  qDest.rowset.fields[aDest[X]].replaceFromFile("XX____XX.XXX")
               CATCH(exception e)
               ENDTRY
            OTHERWISE
               qDest.rowset.fields[aDest[X]].value=qSource.rowset.fields[aSource[X]].value
         ENDCASE
      NEXT
      IF FILE("XX____XX.XXX")
         DELETE FILE("XX____XX.XXX")
      ENDIF
      qDest.rowset.save()
      qSource.rowset.next()
      nActual++
      f.Progress1.value=nActual*100/nRecords
   ENDDO
   IF lSafety
      SET SAFETY ON
   ENDIF
   f.close()
   qDest.active=false
   aSource.active=false
   RETURN

ENDCLASS