// Filename:    RegistryEx.CC
// Written by:  Jim Sare - 07/26/2003
// Purpose:     Extend the Registry class (:samples:registry.prg) to read
//              values of type REG_MULTI_SZ.
// Usage:       See :samples:registry.prg for usage.
//              The only differences are:
//              - Instead of using:
//                Set Procedure To :samples:Registry.prg additive
//                Use:
//                Set Procedure To :dUFLP:RegistryEx.CC additive
//              - Instead of using:
//                  reg = new Registry(HKEY_LOCAL_MACHINE, ;
//                                     BDE_REG_KEY + ;
//                                     "\Settings\DRIVERS\DBASE\TABLE CREATE" )
//                Use:
//                  reg = new RegistryEx(HKEY_LOCAL_MACHINE, ;
//                                       BDE_REG_KEY + ;
//                                       "\Settings\DRIVERS\DBASE\TABLE CREATE" )
//                When calling reg.queryValue, if the value is of any type
//                currently supported by registry.prg, then that value will be
//                returned. If the value is of type REG_MULTI_SZ, then an array
//                of those values will be returned.

#include <windef.h>
#include <winreg.h>

CLASS RegistryEx(openKey, subKey, bCreate);
      of Registry(openKey, subKey, bCreate);
      from :samples:registry.prg

   if type("RegQueryValueExP5") # "FP"
      extern CLONG RegQueryValueExP5(HKEY, LPTSTR, DWORD, LPDWORD, CPTR, LPDWORD);
        ADVAPI32 from "RegQueryValueExA"
   endIf

   FUNCTION queryValue(keyName)
      LOCAL nType, cData, nLen, nResult, c, i

      this.open()
      if not this.isOpen
         return false
      endIf
      nType = 0
      cData = chr(0)
      nLen = 1
      nResult = RegQueryValueEx(this.key, keyName, 0, nType, cData, nLen)
      if (nType == REG_MULTI_SZ)
         cData = replicate(chr(0), ceiling(nLen / 2))
         nResult := RegQueryValueExP5(this.key, keyName, 0, nType, cData, nlen)
         if (nResult == ERROR_SUCCESS)
            keyValue = new array()
            c = ""
            for i = 0 to nLen - 1
               if cData.getByte(i) # 0
                  c += chr(cData.getByte(i))
               elseIf not empty(c)
                  keyValue.add(c)
                  c := ""
               endIf
            endFor
         endIf
      endIf
      this.close()
      RETURN iIf(nType == REG_MULTI_SZ, keyValue, SUPER::queryValue(keyName))

ENDCLASS  // RegistryEx of Registry
