// SUPERQRY.CC

/*

   Authors:  Romain Strieff
             Ken Mayer
   Date...:  February 25, 1999

   This is a query that includes many methods and such
   created by Romain Strieff and others. The idea is to
   encapsulate a lot of cool stuff into one query object.

   NOTE: This file includes a form class after the query
   class -- this form is used by the selIndex() method.

   Methods:
       Execute()  
         When the query opens set some calls to the BDE API.

       RecNo()
         Returns the current row number (like the xBASE DML
         function RecNo() )

       RecCount()
         Returns the number of rows in the table currently.

       RecOfTotalString()
         Returns a character string showing current row OF
         Reccount ...

       GotoRecord()
         Use the record number returned by the RecNo() method,
         or if a number is known, use that ...

       GetTableName() (this is a method of the 
                       query, not the rowset!)
         Returns the name of the table used in the select
         statement of the sql property ...

       SelIndex()
         SelIndex allows the user to select an index from
         a list of index tags, or the developer to call a
         specific index tag (uses a tableDef object)

       Count()
         Override the rowset's count method to return a useful 
         result

  Usage Notes:
     Use as a normal query, with the exception that you place
     the name of the custom query in place of "query()":

     set procedure to superqry.cc additive
     q = new SuperQry()
     q.sql = "select * from mytable"
     q.active = true


     The extra methods can be used as you might expect:

     ? q.rowset.RecNo()
     ? q.rowset.RecCount()
     ? q.rowset.RecOfTotalString()
     q.rowset.GotoRecord( 13 )
     ? q.rowset.RecNo()
     ? q.getTableName()
     q.rowset.selIndex()
     q.rowset.selIndex( "myindexname" )
     ? q.rowset.Count()
 
       NOTE: The code used for selIndex() always sets the 
       query at the top of the rowset (this.rowset.first() ) 
       -- there are two places it does this -- if you would 
       rather it didn't, simply comment them out ...

*/

class SuperQry of QUERY custom

   Procedure execute
     // override the execute method

     // we really do need to call the super-class's method:
     super::execute()

     // ----------------------------------------------
     // Necessary for GetRecNo, GotoRec, GetRecCount
     // methods
     If Type("GetRecNo") # "FP"
        extern cint GetRecNo  (cHandle,cInt,CPTR,CPTR);
        IDAPI32 from "DbiGetRecord"
     endif
     If Type("GotoRec") # "FP"
        extern cint GotoRec  (cHandle,cInt);
        IDAPI32 from "DbiSetToRecordNo"
     endif
     If Type("GetRecCount") # "FP"
        extern cint GetRecCount (cHandle,cptr);
        IDAPI32 from "DbiGetRecordCount"
     endif
    
     this.Rowset.Recno            = class::Recno
     this.Rowset.GotoRecord       = class::GotoRecord
     this.Rowset.Reccount         = class::Reccount
     this.Rowset.RecOfTotalString = class::RecOfTotalString
     this.Rowset.Count            = class::Count

     // ----------------------------------------------
     // -- Necessary for the SelIndex() method:
     // get tableName:
     cName = this.GetTableName()
     if empty( cName )
        msgbox( "Error -- invalid table name" )
        this.active := false
        return
     endif

     // setup a tableDef:
     this.t = new TableDef()
     this.t.tableName   := cName
     this.t.database    := this.database
     this.t.load()
     this.rowset.indexFlag = iif( this.t.indexes.size > 0, true, false )

     // Assign method to rowset:
     this.rowset.selIndex = class::SelIndex


   // ===============================================
   // Romain's RecNo, etc., methods:
   Function RecNo
     local c,nRet
     nRet=0
     c = space(8) 
     if (GetRecNo(this.rowset.handle,0,"",c) == 0)
        nRet= c.getByte( 4 ) + bitlshift( c.getByte( 4+1 ), 4 )
     endif
   return int(nRet)


   Function GoToRecord
      Parameters nRec
      lRet=.f.
      if pcount()=1 and type("nRec")="N"
         if (GoToRec(this.rowset.handle,nRec) == 0)
            lRet=.t.
         endif
      endif
   return lRet


   Function RecCount
      local c,nRet
      nRet=0
      c=space(4)
      if (GetRecCount(this.rowset.handle,c) == 0)
         nRet=c.getByte( 0 ) + bitlshift( c.getByte( 0+1 ), 0 )
      endif
   return int(nRet)


   Procedure RecOfTotalString
   return str(this.recno()).lefttrim()+"/"+;
          str(this.reccount()).lefttrim()

   // ===================================================
   // Romain's code to get the tableName from the query
   // -- this uses the BDE API:  
   function GetTableName
      local cRet
      cRet = ""
      if type( "GetTableNameFromQueryHandle" ) # "FP"
         extern cInt GetTableNameFromQueryHandle (cHandle, cString);
               IDAPI32 from "DbiGetCursorProps"
      endif
      nH = this.rowset.handle
      c=space(2048)
      if (GetTableNameFromQueryHandle( nH, c) == 0)
         cRet = c.left(255).lefttrim()
      endif

   return cRet

   // ===================================================
   // SelIndex:
   Procedure SelIndex( cIndexName )

     if this.rowset.indexFlag
        // if no indexName passed, we must
        // assume the user wants to _select_ an
        // index tag from a list:
        if empty( cIndexName )
           fQry = new IndexQryForm()
           fQry.Indexes = new Array()
           // Load array used in form for listbox:
           fQry.Indexes.add( "<No Index>" )
           for i = 1 to this.t.indexes.size
               fQry.Indexes.Add( this.t.indexes[i].indexName )
           next
           fQry.indexes.sort()
           fQry.indexListBox.dataSource := "array fQry.indexes"
           fQry.indexListBox.curSel     := 1
           // Open it!
           fQry.ReadModal()

           // set it!
           if fQry.indexListBox.value == "<No Index>"
              this.rowset.indexName := ""
           else
              this.rowset.indexName := fQry.indexListBox.value                     
           endif
           this.rowset.first()

        else
           // let's try to find it ...
           this.rowset.indexName := "" // empty it out
           for i = 1 to this.parent.t.indexes.size
               if upper(cIndexName) == upper( this.t.indexes[i].indexName )
                  this.rowset.indexName := cIndexName
                  this.rowset.first()
                  exit
               endif
           next

           // if the indexName is still empty:
           if empty( this.rowset.indexName )
              msgbox( "Could not find indexName: "+cIndexName+"." )
           endif
             
        endif // empty( cIndexName )
     else
        // Let user know somewhere there's a mistake ... <G>
        msgbox( "There are no indexes for this table" )
     endif

   // Romain's code to get the tableName from the query
   // -- this uses the BDE API:  
   function GetTableName
      local cRet
      cRet = ""
      if type( "GetTableNameFromQueryHandle" ) # "FP"
         extern cInt GetTableNameFromQueryHandle (cHandle, cString);
               IDAPI32 from "DbiGetCursorProps"
      endif
      nH = this.rowset.handle
      c=space(2048)
      if (GetTableNameFromQueryHandle( nH, c) == 0)
         cRet = c.left(255).lefttrim()
      endif

   return cRet

   // ============================
   // Either Bowen or Peter Rorlick's routine to override
   // the default count() method with a brute-force
   // method that is more accurate ...
   function Count
      // the overridden method of the rowset --
      // during this code, "this" refers to the
      // rowset, not the query, as the rowset is
      // what will be calling this code:
      local b, nCount, lSaveNotify

      // don't notify any active controls that are 
      // datalinked of what we are doing:
      lSaveNotify         = this.notifyControls
      this.notifyControls := false

      // save current location in the table:
      b      = this.bookMark()
      // set the counter:
      nCount = 0

      // loop through the table:
      this.first()
      do while not this.endOfSet
         nCount++
         this.next()
      enddo

      // back to where we were:
      this.goto( b )

      // make it possible to notify the controls again:
      this.notifyControls := lSaveNotify

   return nCount


endclass
// end of class: SUPERQRY()

// form used above in SELINDEX method:
class indexqryForm of FORM
   with (this)
      scaleFontBold = false
      height = 11
      left = 31.2857
      top = 4.9545
      width = 40
      text = "Select index for data"
      clientEdge = true
      topMost = true
      autoCenter = true
      mdi = false
      smallTitle = true
      sizeable = false
      escExit = false
      maximize = false
      minimize = false
   endwith


   this.INDEXLISTBOX = new LISTBOX(this)
   with (this.INDEXLISTBOX)
      height = 5.7727
      left = 1.8571
      top = 0.8182
      width = 35.5714
      id = 101
      colorHighLight = "white/blue"
   endwith


   this.TEXTEXPL = new TEXT(this)
   with (this.TEXTEXPL)
      height = 2.1818
      left = 2.5714
      top = 6.9545
      width = 34.1429
      border = true
      fontSize = 8
      fontItalic = true
      text = "Select the index you wish to use. When you click the button below, whatever index you have chosen will take effect on your table."
      borderStyle = 10	// Etched Out
   endwith


   this.DONEPUSHBUTTON = new PUSHBUTTON(this)
   with (this.DONEPUSHBUTTON)
      height = 1.0909
      left = 17.2857
      top = 9.4091
      width = 4.7143
      text = ""
      default = true
      upBitmap = "RESOURCE #1005"
      downBitmap = "RESOURCE #1006"
      speedTip = "Done"
      value = false
      onClick = {; form.close() }
   endwith

endclass

/*
   end of file: superqry.cc
*/

