/*
      
   Version 1.4 
 
   UsvOut.prg
   Posted in dBASE, Inc. newsgroups, January 17, 2002
  

   This program will output a text-formatted file with
   User-Defined Separated Value from a table.

   ****

   Original Name: CsvOut.prg 
   Posted in dBASE, Inc. newsgroups, May 11, 1999
   by Gary White
 
   January 17, 2002
   E Sanders
   Version 1.0
   Added: PARAMETER cSeparator
   Because user can choose any separator,
   the program's Name is changed from CsvOut.prg to UsvOut.prg 

   October 24, 2007
   Version 1.1																 
   Ivar B. Jessen                                                     
   Added: PARAMETER bMemo                                             
   Allows user to include or not include data from memo fields        
   in the text-formatted file                                         

   February 26, 2008
   Version 1.2
   Roland Wingerter
   Changes:
   - Use CRLF instead of CR only
   - Replace double quotes with double double quotes in character data,
     because they are used as delimiters and otherwise cause problems 
     when importing. 
   - Added set proc to :dUFLP:stringEx.cc to be able to use strTran()

   January 18, 2016
   Version 1.3
   Mervyn Bick
   Added: If fieldnames are inclused in the .csv file and the first field 
      name is id or ID Excel sees the .csv file as a .slk file.  
      This can be prevented by prefixing the fieldname with a space.  
   Added: Version numbers  

   September 22, 2017
   Version 1.4
   Mervyn Bick
   Changes: A spurious seperator was added to the end of a record if the last 
        field was _dBASELOCK or if the last field was a memofield where memofields 
        were excluded.  This was caused by writing the contents of fields one by 
        one and then writing the separator after each field except for the last
        field in the fields array.  Testing to avoid placing the spurious separtor
        was complicated.  Changed to building each record in memory with a 
        separator after every field.  Before writing the record to the output file 
        the final separator is removed.  Much simpler!              

   Added: The ability to change delimiters for character fields to bring
       the program into line with dBASE's  COPY TO ... DELIMITED WITH ...
       As the delimiter can now be changed it is no longer necessary to force
       the replacement of a double quote in text with two double quotes if a
       different delimiter is used.  It is, however, necessary to replace TAB in
       text with 5 spaces where TAB is used as a separator without a delimiter i.e
       to create what Microsoft calls a 'TAB delimited' file.  Strictly spaking,
       these files are 'TAB separated, non-delimited' files as there is no leading 
       TAB before the first field nor is there a TAB after the last field. 
       
   Added: Dates included in a .csv file by the dBASE command  
        COPY TO ... DELIMITED WITH ... are in the form yyyymmdd i.e no separators
        and no delimiters, no matter what the date format on the computer is set 
        to and this is the format dBASE requires when reading a date from a .csv 
        file into a table. 
        
        Not all 3rd-party programs are happy with this and the option is provided 
        for changing the date format in the output file.  Previous vesions of 
        usvout.prg output dates in the date format set for the computer albeit 
        without delimiters e.g 2017-09-26.  Where a delimited option is selected for 
        the date the delimiter specified in the parameters is used for the date 
        and time.  
        
   Added: The dBASE  command COPY TO ... DELIMITED WITH ... outputs the contents
        of a logical field as T|F which may not be acceptable to 3rd-party
        programs.  Previous versions of usvout.prg output the contents of 
        logical fields as true|false.  As an alternative the user now has the 
        option of selecting T|F, .T.|.F., Y|N or true|false        
       
   Usage:
   
   set procedure to :duflp:usvout.prg
   UsvOut( <rowset>, <filename>, <fieldnames>, <cSeparator>, <bMemo>, <cDelimiter>, cDateFormat, cLogicFormat ) //Changed by MB Sept 2017

   <rowset>     = the rowset object you want to export.
   <filename>   = character expression - the name of the file to create
   <fieldnames> = logical - true to include fieldnames, false,
   					if omitted, fieldnames are excluded.					
   <cSeparator> = character expression - the character used as the separator
                  if omitted, the comma is used as separator.
	<bMemo>		 = logical - true/false to include/exclude memo fields,
   					if omitted, memo fields are included  
   <cDelimiter> = character expression. Single chracter to be used to delimit charcter fields.   
   <cDateFormat>  =  character expression in the range '1' to '10' 
                     1  YYYYMMDD       This is the dBASE default for .csv files. 
                     2  MM/DD/YYYY     Americsn not delimited 
                     3  DD/MM/YYYY     British not delimited
                     4  DD.MM.YYYY     German not delimited
                     5  YYYY-MM-DD     YMD not delimited
                     6  "YYYYMMDD"     Delimited
                     7  "MM/DD/YYYY"   American delimited
                     8  "DD/MM/YYYY"   British delimited
                     9  "DD.MM.YYYY"   German delimited
                    10  "YYYY-MM-DD"   YMD delimited
    <cTimeFormat> =  character expression in the range '1' to '2'
                     1  date 00:00:00 AM|PM     12 hr time added to undelimited date   Default
                     2  date 00;00;00           24 hr time added to undelimited date
    <cLogicFormat> = character expression in the range '1' to '4'
                     1   T|F         Default
                     2   .T.|.F.
                     3   Y|N
                     4   true|false                     

   Examples:
     set procedure to :duflp:usvout.prg
     q = new query()
     q.sql              := "select * from MyTable"
     q.active           := true
     q.rowset.indexName := "SomeIndex"
     // if you want to filter the data:
     // q.rowset.filter = ...
     q.rowset.first()
     
     UsvOut( q.rowset, "MyCSVFile.CSV", false , ",",false )  // Equvalent to COPY ... TO ... DELIMITED
   or  
     UsvOut( q.rowset, "MyCSVFile.CSV", true|false ) 
   or
     UsvOut( q.rowset, "MyASVFile.ASV", true|false , "'" )
   or
     UsvOut( q.rowset, "MyTSVFile.TSV", true|false , chr(9) )
   or                            
     UsvOut( q.rowset, "MyTSVFile.TSV", true|false , chr(9), true|false )  //tab separator, default double qoute delimiter
   or
     UsvOut( q.rowset, "MyCSVFile.CSV", true|false , ',', true|false,"'" )  //comma separator, single quote delimiter
   or
     UsvOut( q.rowset, "MyTXTFile.TXT", true|false , chr(9), true|false, "" ) // Excel "Tab delimited" file 
   or
     if not UsvOut( q.rowset, "MyCSVFile.TSV", false , chr(9) )
        msgbox("Problem with converting!","error",16)
     endif
  

*/

   parameters oRowset, cFileName, bFieldNames, cSeparator, bMemo, cDelimiter,;
              cDateFormat,cTimeFormat, cLogicFormat //Changed MB Sept 2017

   ** added R.W.
   set procedure to :dUFLP:stringEx.cc addi
   **

   private xValue, cSepa, cMem, cDelim,cDateForm,cTimeForm, cLogicForm	//Changed MB Sept 2017
   
   cSepa = ','
   cMem = ''                     
   cDelim = '"'
   cDateform = '1'
   cTimeForm= '1'
   cLogicForm = '1'
   
   do case   // Case construct re-written MB Sept 2017.                
      case argcount() < 3 
         bFieldnames = false    
      case argcount() = 3 
         //bFieldnames set by parameter   
      case argcount() = 4
         cSepa = cSeparator
      case argcount() = 5 
         cSepa = cSeparator
         if bMemo = false           
            cMem = "MEMO"
         else
            cMem = ''      
         endif 
      case argcount() = 6 
         cSepa = cSeparator
         if bMemo = false           
            cMem = "MEMO"
         else
            cMem = ''      
         endif 
         cDelim = cDelimiter
      case argcount() = 7 
         cSepa = cSeparator
         if bMemo = false           
            cMem = "MEMO"
         else
            cMem = ''      
         endif 
         cDelim = cDelimiter
         cDateForm = cDateFormat
      case argcount() = 8 
         cSepa = cSeparator
         if bMemo = false           
            cMem = "MEMO"
         else
            cMem = ''      
         endif 
         cDelim = cDelimiter
         cDateForm = cDateFormat
         cTimeForm = cTimeFormat 
      case argcount() = 9
         cSepa = cSeparator
         if bMemo = false           
            cMem = "MEMO"
         else
            cMem = ''      
         endif 
         cDelim = cDelimiter
         cDateForm = cDateFormat
         cTimeForm = cTimeFormat 
         cLogicForm = cLogicFormat  
      endcase                       

   local bRetVal, i, f, cType
   bRetVal = true
   try
      if type('oRowset.filterOptions') == 'U'
         throw new csvError('Type mismatch - expecting rowset')
      endif
      if type('cFileName') # 'C' or empty(cFileName)
         throw new csvError('Invalid output file name')
      endif
      fOutFile = new file()
      fOutFile.create( cFileName, 'W' )
      oRowset.first()
      f = oRowset.fields
      if bFieldNames
         cLine = ''
         for i = 1 to f.size
            if f[i].fieldName # '_DBASELOCK' and ;
               not f[i].type $ 'OLE|BINARY|&cMem'
               *********** Added by Mervyn Bick  Jan 2016
               if i = 1 and upper(f[i].fieldname) = "ID"
                  cLine = cLine + cDelim+' '+f[i].fieldname+cDelim+cSepa
               else
                  cLine = cLine + cDelim+f[i].fieldname+cDelim+cSepa
               endif
               **********
            endif
         endfor
         cLine = substr(cLine,1,rat(cSepa,cLine)-1) //remove final separator
         fOutFile.puts(cLine)
      endif

      do while not oRowset.endOfSet
         cLine = ''
         for i = 1 to f.size
            if f[i].fieldName # '_DBASELOCK' and ;
               not f[i].type $ 'OLE|BINARY|&cMem'
               xValue = f[i].value
               // The following section reworked by MB Sept 2017 to cater for new options.
               if f[i].type = 'CHAR' or f[i].type = 'MEMO'
                  if empty(xValue)
                     xValue = ' '
                  endif   
                  if cSepa = chr(9) and cDelim = '' // Added MB Sept 2017
                     xValue := new stringEx().strTran(xValue, chr(9), [   ])
                  else
                     xValue := trim(xValue)
                  endif
                  cLine = cLine + cDelim + trim(xValue) + cDelim + cSepa
               elseif type('xValue') == 'L'
                  cLine = cLine+ formatLogic(xValue,cLogicForm) +cSepa
               elseif f[i].type = 'DATETIME'
                  cLine = cLine + formatDate(xValue,cDateForm,cDelim,cTimeForm) +cSepa
               elseif f[i].type = 'DATE'
                  cLine = cLine + formatDate(xValue,cDateForm,cDelim) +cSepa
               else   // type = AUTOINC, NUMERIC, DOUBLE or LONG
                  cLine = cLine+xValue+cSepa
               endif
            endif
         next
         cLine = substr(cLine,1,rat(cSepa,cLine)-1) //remove final separator
         fOutFile.puts(cLine)     
         oRowset.next()
      enddo
   catch( exception e )
      msgbox( e.message, 'Error', 16 )
      bRetVal := false
   finally
      if type('fOutFile') == 'O' and ;
         fOutFile.handle > 0
         fOutFile.close()
         release object fOutFile
         release fOutFile
      endif
   endtry
   return bRetVal
   
   function formatLogic(xValue,cLogicForm)
      local cRet
      if cLogicForm = '1'
         cRet = iif(xValue=true,'T','F') 
      elseif cLogicform = '2'
         cRet = iif(xValue=true,'.T.','.F.')
      elseif cLogicForm = '3'
         cRet = iif(xValue=true,'Y','N')
      else
         cRet = iif(xValue=true,'true','false')
      endif
      return cRet
 
   function formatDate(xValue,cDateForm,cDelim,cTimeForm)
       local cRet,nDateForm
       cRet = ''
       nDateForm = val(cDateForm)
       cDate = dtos(xValue)  //DTOS() formats date both for Date and DateTime values      
       if nDateForm = 1 or nDateForm = 6  // Default DTOS() format
          cRet = cDate
       elseif nDateForm = 2 or nDateForm = 7
          cRet = substr(cDate,5,2)+'/'+substr(cDate,7,2)+'/'+substr(cDate,1,4)   //American
       elseif nDateForm = 3 or nDateForm = 8
          cRet = substr(cDate,7,2)+'/'+substr(cDate,5,2)+'/'+substr(cDate,1,4)   //British
       elseif nDateForm = 4 or nDateForm = 9
          cRet = substr(cDate,7,2)+'.'+substr(cDate,5,2)+'.'+substr(cDate,1,4)   //German
       elseif nDateForm = 5 or nDateForm = 10
          cRet = substr(cDate,1,4)+'-'+substr(cDate,5,2)+'-'+substr(cDate,7,2)   //YMD
       endif   

       if argCount() = 4 // DateTime value passed to fuction
          cDateTime = dttoc(xValue)
          cTime = substr(cDateTime,at(' ',cDateTime)+1)
          cAMPM = ''
          If 'M'$cTime
              if 'P'$cTime 
                 cHr = str(val(substr(cTime,1,2))+12,2,0,'0')
                 cAMPM = 'PM'
              else   
                 cHr = str(val(substr(cTime,1,2)),2,0,'0')
                 cAMPM = 'AM'
              endif
          else
             cHr = str(val(substr(cTime,1,2)),2,0,'0') 
          endif 
          cMinSec = substr(cTime,4,5)

          if cTimeForm = '1' and 'M'$cTime
             cRet = cRet + ' '+cTime
          elseif cTimeform ='1'
             cRet = cRet + ' '+cHr+':'+cMinSec+' '+cAMPM
          elseif cTimeForm = '2'
             cRet = cRet + ' '+cHr+':'+cMinSec
          endif
       endif
       
       if nDateForm >= 7  //Delimited options
          cRet = cDelim+ cRet + cDelim
       endif
      return cRet

class csvError( cMsg ) of exception
   this.message = cMsg
endclass
