CLASS DateEx of Date
/*
  -----------------------------------------------------------------
    Description:
 
       This class contains many useful date-handling methods. For 
       'Holidays' and 'Hebrew dates', check the files: "HOLIDAY.CC" 
       and "HEBREW.CC"
 
    Programmers: 
 
       Ken Mayer
       Jay Parsons
       Martin Leon
       Kenneth Chan 
       David Love
       Miriam Liskin
       Angus Scott-Fleming
       Todd Barry
       Romain Strieff 
       Raymond S. Leventhal
       Bowen Moursund 
       David Kanter
       Jose Luis Perez 
       Ivar B. Jessen
       Marc van der Bergen
       Roland Wingerter

 
    History: 
 
       12/23/1996 -- original version of these routines as a class.
       01/26/1998 -- Jose fixed date routines to deal with the
                     default SET EPOCH settings and such.
       10/08/1998 -- Attempts to deal with internationalization
                     issues -- in some cases this simply means
                     using the date object ...
       10/20/1998 -- Made this a subclass of the date object,
                     which with a bit of work bypasses all of
                     the international date issues!
                     The only difficulty lies in the fact that
                     the month property in the date object
                     starts at zero (0 = January, 11 = December)
                     and the day property starts at zero as well
                     (0 = Sunday, 6 = Saturday) which can throw
                     calculations off ... (day relates to DOW() )
                     (KM)
       March, 2000 -- Added Ivar's EWeekNo and FDoEW routines
       January, 2003 --
                      Added AddMonths2, for use in calendar
                      routines, so that navigation from month
                      to month is correct ...
       March, 2004 -- Massive changes, using Ivar B. Jessen's
                      code to deal with International Date
                      Month and Day of Week names ... This
                      uses Windows calls to determine internal
                      identifiers and returns appropriate
                      strings based on the language assigned.
                      See below for more. To see a list of
                      all languages supported, see the startup
                      code for this class.

                      Added more "European Week" methods.
       March, 2009 -- Roland W. fixed getDifference bug ...
       April, 2020 -- Updates due to issues found by Vasile Buza.
                      The following methods were update:
                        addMonths
                        addMonths2
                        AddYears
                        Num2Dat
                        StoD
                        Text2Date 

    Instantiation: 
 
       set procedure to dateex.cc additive
       dMyDate = new DateEx()
       // for International Purposes, if you want to use the
       // IntlCDOW and/or IntlCMON methods of this class,
       // you can force a language by using:
       dMyDate.language = "MyLanguage"
          i.e.,
          // IMPORTANT NOTE: This must be entered 
          // as shown, and it is case sensitive ...:
          dMyDate.language = "Dutch (Netherlands)"
       ? dMyDate.IntlCDOW( date() )

       // ----- To know more about whether or not your
       //       computer can display information for
       //       any of the Language options listed below,
       //       see 'CountryInfo.wfm' in the dUFLP ...
 
    Methods:
       More details on each of these is contained in the 
       commentary for each method ...
 
       Release()
         Releases this object from memory
 
       AddMonths()
         Add months to a date
 
       AddYears()
         Add years to a date
  
       Age()
         Gives 'years' between date passed as parm and date()
  
       Age2()
         'Years' between two dates 
 
       Annivrsry()
         Is a date within a range of dates?

       BeginEWeekDate()
         Date of first day of the EWeek containing the date.

       BeginEWeekNoYear()
         Date of first day of week having EWeekNo and belonging
         to the year.

       ChangeLanguage()
         This is a UI bit, it allows the user to bring up
         a listing of all dates available, and change
         the language used by the International date methods:
         IntlCDOW, IntlCDOWShort, IntlCMON, IntlCMONShort
 
       Dat2Jul()
         Convert date to number of days since January 1, 4713 B.C.
 
       DateText()
         Returns date as January 1, 1997
 
       DateText2()
         Returns date as Wednesday, January 1, 1997
 
       DateText3()
         Returns date as January, 1997
  
       DoY()
         Number of days from beginning of year
  
       Epoch  -- REMOVED -- see SET EPOCH in online help

       EWeekNo()
         Calculates European Week number based on first day
         of week being Monday

       EWeeksInYear()
         Returns number of European Weeks in the year.
 
       FDoD()
         First Date of Day of month

       FDoEW()
           ** Note that this method only really serves one
              purpose, and that is it's use by the EWeekNo()
 
       FDoM()
         First Day of Month
 
       FDoY()
         First Day of Year
 
       FDoNextMonth()
         First Day of Next Month
 
       Fiscal()
         Returns fiscal year for a given date (based on US Gov't
         Fiscal year starting in October).

       getDifference()
         Returns an object with nDays, nMonths and nYears difference 
         between two dates.

      getDifferenceInMonths()
         Uses getDifference above to return the difference between two
         dates in months.

      getDifferenceInYears()
         Uses getDifference above to return the difference between two
         dates in years.
 
       IntlCDOW()
         International Character Day of Week

       IntlCDOWShort()
         International Character Day of Week - Short Name
 
       IntlCMON()
         International Character Month

       IntlCMONShort()
         International Character Month -- Short Name

       IFirstWeekOfYear()
         Returns a numeric value corresponding to what the
         locale is, based on the first week of the year.
         This is useful to determine if the user's language
         is for a country, for example, that uses
         the European Week Number that there are lots of
         routines for in this class.

       IntlLongFormat()
         Returns character string based on "correct" long
         format of date for user, based on language. The
         information is gotten from Windows ...

       IntlShortFormat()
         Returns character string based on "correct" short
         format of date for user, based on language. The
         information is gotten from Windows ... (Short
         format is, for example, 3/22/2004)

       IntlSetDate()
         Issues dBL command:   set date X where 'X' is
            MDY, DMY, or YMD, based on the language property
            of the dateex object.

       IntlSetMark()
         Issues dBL command:   set mark to 'x' 
            where the mark is the separator character for short
            date formats, based on the language property of
            the dateex object.
 
       IsLeap()
         Is the year a leap year?
 
       Jul2Dat()
         Converts 'Julian' date to dBASE date
 
       LDoD()
         Last day of week in month (i.e., last monday of the month)
 
       LDoM()
         Last day of month

       LDoM2()
         Last day of month -- this uses a date object as the parameter rather than
         a date.
 
       LDoY()
         Last day of year

       NDoW()
         Numeric Day of Week
 
       Num2Dat()
         Numeric date value to dBASE Date
 
       NumDoD()
         x day of week in a month (i.e., the third monday ...)
 
       Quarter()
         Quarter of the year for specified date
 
       StoD()
         String to date (reverse of DtoS())
 
       Text2Date()
         Converts text date in format: dd-mmm-yy  to dBASE date
 
       WeekNo()
         Week a date falls in from first of year.

       YearOfEweek()
         Year to which an EWeek belongs.
 
    Properties:
 
       Language -- determined in the startup code, or at least
                   an attempt is made to do so ... 
                   default: System Default Language
                   (this is the language of the version
                    of the Operating System)
   ----------------------------------------------------------------
*/

   // -----------------------------------------------------
   // -------------------- STARTUP CODE -------------------
   // -----------------------------------------------------
   // This Startup Code is based on work Ivar B. Jessen 
   // has done, digging into the Windows API and finding
   // useful information for the international community.
   //
   // Added March, 2004, to try to help
   // with language issues. If all else fails, the four
   // methods that rely on the language default to 
   // the computer's system default language:
   this.Language = "System Default Language"

   // These should be in the dBASE folders, installed 
   // with dBASE ...
   #include <windef.h>
   #include <winnt.h>

   // define API calls:
   If type('GetSystemDefaultLCID') # 'FP'
      extern LCID GetSystemDefaultLCID(VOID) kernel32
   endif
   If type('GetLocaleInfoA') # 'FP'
      extern CINT GetLocaleInfoA( CLONG, CINT, CSTRING, CINT) kernel32
   endif

   // AssocArray of language names and codes:
   this.asLang = new AssocArray()
   fsL =this.asLang

   // Language Identifiers -- this list is taken from
   // the Microsoft Windows SDK website, Table of Language
   // Indentifiers (International Features) ...
   // This info is from:
   //   http://msdn.microsoft.com/library/default.asp?url=/library/en-us/intl/nls_238z.asp
   // If you put that in your browser and don't see a list of
   // the following information, you should see a list on the
   // left of the screen, click on: "Table of Language Identifiers"
   //
   // Note that some of the languages in the list may
   // return 'garbage' characters on your machine unless
   // it is installed with the appropriate settings (i.e.,
   // languages such as Arabic, Japanese, Korean, Chinese,
   // Russian, and others require very special character sets,
   // some require Unicode, etc....).
   //
   // The comments above some of the individual items were
   // taken from the SDK table as noted above and included
   // here.
   // ---------------------------------------------------------------
   //   Locale	                               Identifier
   fsL["Language Neutral"			         ] = 0x0000
   fsL["Process Default Language"         ] = 0x0400
   fsL["System Default Language"          ] = 0x0800
   fsL["Afrikaans"                        ] = 0x0436
   fsL["Albanian"                         ] = 0x041c
   fsL["Arabic (Saudi Arabia)"            ] = 0x0401
   fsL["Arabic (Iraq)"                    ] = 0x0801
   fsL["Arabic (Egypt)"                   ] = 0x0c01
   fsL["Arabic (Libya)"                   ] = 0x1001
   fsL["Arabic (Algeria)"                 ] = 0x1401
   fsL["Arabic (Morocco)"                 ] = 0x1801
   fsL["Arabic (Tunisia)"                 ] = 0x1c01
   fsL["Arabic (Oman)"                    ] = 0x2001
   fsL["Arabic (Yemen)"                   ] = 0x2401
   fsL["Arabic (Syria)"                   ] = 0x2801
   fsL["Arabic (Jordan)"                  ] = 0x2c01
   fsL["Arabic (Lebanon)"                 ] = 0x3001
   fsL["Arabic (Kuwait)"                  ] = 0x3401
   fsL["Arabic (U.A.E.)"                  ] = 0x3801
   fsL["Arabic (Bahrain)"                 ] = 0x3c01
   fsL["Arabic (Qatar)"                   ] = 0x4001
   // Windows 2000/XP only, and Unicode only:
   fsL["Armenian"                         ] = 0x042b
   fsL["Azeri (Latin)"                    ] = 0x042c
   fsL["Azeri (Cyrillic)"                 ] = 0x082c
   fsL["Basque"                           ] = 0x042d
   fsL["Belarusian"                       ] = 0x0423
   fsL["Bulgarian"                        ] = 0x0402
   fsL["Burmese"                          ] = 0x0455
   fsL["Catalan"                          ] = 0x0403
   // Chinese -- Unicode only:
   fsL["Chinese (Taiwan)"                 ] = 0x0404
   fsL["Chinese (PRC)"                    ] = 0x0804
   fsL["Chinese (Hong Kong SAR, PRC)"     ] = 0x0c04
   fsL["Chinese (Singapore)"              ] = 0x1004
   // Windows 98/ME, Windows 2000/XP:
   fsL["Chinese (Macau SAR)"              ] = 0x1404
   fsL["Croatian"                         ] = 0x041a
   fsL["Czech"                            ] = 0x0405
   fsL["Danish"                           ] = 0x0406
   // Windows XP, Unicode only:
   fsL["Divehi"                           ] = 0x0465
   fsL["Dutch (Netherlands)"              ] = 0x0413
   fsL["Dutch (Belgium)"                  ] = 0x0813
   fsL["English (United States)"          ] = 0x0409
   fsL["English (United Kingdom)"         ] = 0x0809
   fsL["English (Australian)"             ] = 0x0c09
   fsL["English (Canadian)"               ] = 0x1009
   fsL["English (New Zealand)"            ] = 0x1409
   fsL["English (Ireland)"                ] = 0x1809
   fsL["English (South Africa)"           ] = 0x1c09
   fsL["English (Jamaica)"                ] = 0x2009
   fsL["English (Caribbean)"              ] = 0x2409
   fsL["English (Belize)"                 ] = 0x2809
   fsL["English (Trinidad)"               ] = 0x2c09
   // Windows 98/ME, Windows 2000/XP:
   fsL["English (Zimbabwe)"               ] = 0x3009
   // Windows 98/ME, Windows 2000/XP:
   fsL["English (Philippines)"            ] = 0x3409
   fsL["Estonian"                         ] = 0x0425
   fsL["Faeroese"                         ] = 0x0438
   fsL["Farsi"                            ] = 0x0429
   fsL["Finnish"                          ] = 0x040b
   fsL["French (Standard)"                ] = 0x040c
   fsL["French (Belgian)"                 ] = 0x080c
   fsL["French (Canadian)"                ] = 0x0c0c
   fsL["French (Swiss)"                   ] = 0x100c
   fsL["French (Luxembourg)"              ] = 0x140c
   // Windows 98/ME, Windows 2000/XP:
   fsL["French (Monaco)"                  ] = 0x180c
   // Windows XP:
   fsL["Galacian"                         ] = 0x0456
   // Windows 2000/XP, Unicode Only:
   fsL["Georgian"                         ] = 0x0437
   fsL["German (Standard)"                ] = 0x0407
   fsL["German (Swiss)"                   ] = 0x0807
   fsL["German (Austrian)"                ] = 0x0c07
   fsL["German (Luxembourg)"              ] = 0x1007
   fsL["German (Liechtenstein)"           ] = 0x1407
   fsL["Greek"                            ] = 0x0408
   // Windows XP, Unicode Only:
   fsL["Gujarti"                          ] = 0x0447
   fsL["Hebrew"                           ] = 0x040d
   fsL["Hungarian"                        ] = 0x040e
   fsL["Icelandic"                        ] = 0x040f
   fsL["Indonesian"                       ] = 0x0421
   fsL["Italian (Standard)"               ] = 0x0410
   fsL["Italian (Swiss)"                  ] = 0x0810
   fsL["Japanese"                         ] = 0x0411
   // Windows XP, Unicode:
   fsL["Kannada"                          ] = 0x044b
   // Windows 2000/XP, Unicode:
   fsL["Konkani"                          ] = 0x0457
   fsL["Korean"                           ] = 0x0412
   // Windows 95, Windows NT 4.0 only:
   fsL["Korean (JoHab)"                   ] = 0x0812
   // Windows XP:
   fsL["Kyrgyz"                           ] = 0x0440
   fsL["Latvian"                          ] = 0x0426
   fsL["Lithuanian"                       ] = 0x0427
   // Windows 98 only:
   fsL["Lithuanian (Classic)"             ] = 0x0827
   fsL["FYRO Macedonian"                  ] = 0x042f
   fsL["Malay (Malaysian)"                ] = 0x043e
   fsL["Malay (Brunei Darussalam)"        ] = 0x083e
   // Windows 2000/XP, Unicode:
   fsL["Marathi"                          ] = 0x044e
   // Windows XP:
   fsL["Mongolian"                        ] = 0x0450
   fsL["Norwegian (Bokmal)"               ] = 0x0414
   fsL["Norwegian (Nynorsk)"              ] = 0x0814
   fsL["Polish"                           ] = 0x0415
   fsL["Portuguese (Brazilian)"           ] = 0x0416
   fsL["Portuguese (Standard)"            ] = 0x0816
   // Windows XP, Unicode:
   fsL["Punjabi"                          ] = 0x0446
   fsL["Romanian"                         ] = 0x0418
   fsL["Russian"                          ] = 0x0419
   // Windows 2000/XP, Unicode:
   fsL["Sanskrit"                         ] = 0x044f
   fsL["Serbian (Cyrillic)"               ] = 0x0c1a
   fsL["Serbian (Latin)"                  ] = 0x081a
   fsL["Slovak"                           ] = 0x041b
   fsL["Slovenian"                        ] = 0x0424
   fsL["Spanish (Spain, Traditional Sort)"] = 0x040a
   fsL["Spanish (Mexican)"                ] = 0x080a
   fsL["Spanish (Spain, Modern Sort)"     ] = 0x0c0a
   fsL["Spanish (Guatemala)"              ] = 0x100a
   fsL["Spanish (Costa Rica)"             ] = 0x140a
   fsL["Spanish (Panama)"                 ] = 0x180a
   fsL["Spanish (Dominican Republic)"     ] = 0x1c0a
   fsL["Spanish (Venezuela)"              ] = 0x200a
   fsL["Spanish (Colombia)"               ] = 0x240a
   fsL["Spanish (Peru)"                   ] = 0x280a
   fsL["Spanish (Argentina)"              ] = 0x2c0a
   fsL["Spanish (Ecuador)"                ] = 0x300a
   fsL["Spanish (Chile)"                  ] = 0x340a
   fsL["Spanish (Uruguay)"                ] = 0x380a
   fsL["Spanish (Paraguay)"               ] = 0x3c0a
   fsL["Spanish (Bolivia)"                ] = 0x400a
   fsL["Spanish (El Salvador)"            ] = 0x440a
   fsL["Spanish (Honduras)"               ] = 0x480a
   fsL["Spanish (Nicaragua)"              ] = 0x4c0a
   fsL["Spanish (Puerto Rico)"            ] = 0x500a
   fsL["Sutu"                             ] = 0x0430
   fsL["Swahili (Kenya)"                  ] = 0x0441
   fsL["Swedish"                          ] = 0x041d
   fsL["Swedish (Finland)"                ] = 0x081d
   // Windows XP, Unicode:
   fsL["Syriac"                           ] = 0x045a
   // Windows 2000/XP, Unicode:
   fsL["Tamil"                            ] = 0x0449
   fsL["Tatar (Tatarstan)"                ] = 0x0444
   // Windows XP, Unicode:
   fsL["Telugu"                           ] = 0x044a
   fsL["Thai"                             ] = 0x041e
   fsL["Turkish"                          ] = 0x041f
   fsL["Ukrainian"                        ] = 0x0422
   // Windows 98/ME, Windows 2000/XP:
   fsL["Urdu (Pakistan)"                  ] = 0x0420
   fsL["Urdu (India)"                     ] = 0x0820
   fsL["Urzbek (Latin)"                   ] = 0x0443
   fsL["Urzbek (Cyrillic)"                ] = 0x0843
   // Windows 98/ME, Windows NT 4.0 and later:
   fsL["Vietnamese"                       ] = 0x042a

   // -----------------------------------------------------
   // ------------- END OF STARTUP CODE -------------------
   // -----------------------------------------------------

   FUNCTION Release
   /*
       -------------------------------------------------------------
       Programmer..: Ken Mayer
       Date........: 11/24/1996
       Notes.......: Used to release this object from memory.
       Written for.: Visual dBASE 5.5
       Rev. History: 11/24/1996 -- Original Code
       Calls.......: none
       Usage.......: dateex.Release()
       Example.....: dMyDate.Release()
                     dMyDate = NULL   // not required, but a good 
                                      // idea
       Returns.....: Null
       Parameters..: None
       -------------------------------------------------------------
   */

      release object this

   RETURN NULL
   // - EoM: Release()

   FUNCTION addMonths( dOld, nMonths, bRetainTime )
   /*
      -------------------------------------------------------------
       Programmer..: Ken Chan
       Date........: January, 2002
       Notes.......: Finds same day as given date N months ahead.
                     This function will return the first day of 
                     the following month if there is no equal 
                     date in the month otherwise returned and 
                     nMonths is positive, or the last day of 
                     the month if nMonths is negative.  That 
                     is, a call with {01/31/91} (January 31, 
                     1991) and 1 would yield March 1, there 
                     being no February 31. Do not use this 
                     successively to find first the 
                     date one month ahead, then the date one 
                     month beyond that. Instead, to find the 
                     date two months ahead from the original 
                     date, call this function with the original 
                     date and nMonths = 2.  Otherwise, in the
                     example, you'll get April 1 the second 
                     time rather than the correct March 31.
       Written for.: dB2K
       Rev. History: January 28, 2002 -- This is a complete
                     revision of the original code originally
                     written by Jay Parsons. The code now
                     here was written by Ken Chan, and is
                     designed to take into account negative
                     values and positive values for nMonths,
                     and to deal with a datetime or timestamp,
                     allowing you to retain the time ...
       Calls.......: None
       Usage.......: DateEx.AddMonths(<dDate>,<nMonths> [,<bRetain>])
       Example.....: ? dMyDate.AddMonths({01/01/91},1)
       Returns.....: Date
       Parameters..: dDate   = Date being tested for ...
                     nMonths = Number of months "ahead"
                     bRetain = logical - retain time?
      -------------------------------------------------------------
   */
      local dRet, nNewMonth

		// Initialiaze return date to 01.01.1970 (Vasile Buza, April 2020)
		dRet = new Date(0)
		
      // Add months to current month, which is zero-based
      // (Use round() on arg just in case it's non-int)
      nNewMonth = dOld.month + round( nMonths, 0 )

      // Apply year offset, if any
      dRet.year = dOld.year + floor( nNewMonth / 12 )

      // Convert month number to valid 0-11 range
      nNewMonth %= 12
      // MOD() and % operator use INT(), so adjust when
      // month is negative
      if nNewMonth < 0
         nNewMonth += 12
      endif

      // Apply new month to Date object
      dRet.month := nNewMonth
      // Restore day (note property is named "date")
      dRet.date  := dOld.date

      // If new month gets nudged, day was out of range
      if dRet.month # nNewMonth
         if nMonths < 0
           // When negative, return last day of month. Month is
           // always nudged forward, so put it back
           dRet -= dRet.date
         else
           // Otherwise, return first day of month
           dRet.date := 1
         endif
      endif
   // Convert to date type if required
   RETURN iif( bRetainTime, dRet, dttod( dRet ) )
   // - EoM: AddMonths()

   FUNCTION addMonths2( dOld, nMonths, bRetainTime )
   /*
      -------------------------------------------------------------
       Programmer..: Ken Chan
       Date........: January, 2002
       Notes.......: Finds same day as given date N months ahead.
                     This function will return the first day of 
                     the following month if there is no equal 
                     date in the month otherwise returned and 
                     nMonths is positive, or the last day of 
                     the month if nMonths is negative.  That 
                     is, a call with {01/31/91} (January 31, 
                     1991) and 1 would yield March 1, there 
                     being no February 31. Do not use this 
                     successively to find first the 
                     date one month ahead, then the date one 
                     month beyond that. Instead, to find the 
                     date two months ahead from the original 
                     date, call this function with the original 
                     date and nMonths = 2.  Otherwise, in the
                     example, you'll get April 1 the second 
                     time rather than the correct March 31.
       Written for.: dB2K
       Rev. History: January 28, 2002 -- This is a complete
                     revision of the original code originally
                     written by Jay Parsons. The code now
                     here was written by Ken Chan, and is
                     designed to take into account negative
                     values and positive values for nMonths,
                     and to deal with a datetime or timestamp,
                     allowing you to retain the time ...

                     January 29, 2003 -- Modification to return
                     last day of month if date is not found in
                     that month (i.e., as shown above, if the
                     date is {01/31/91}, going to the next month
                     would take it to {03/01/91}, which 
                     can cause some consternation for some 
                     users ...). If the date is {01/31/91},
                     and you select to add a month, since
                     Feb. 31 does not exist, this routine should
                     return {02/28/91} (last day of the 
                     month).
       Calls.......: None
       Usage.......: DateEx.AddMonths(<dDate>,<nMonths> [,<bRetain>])
       Example.....: ? dMyDate.AddMonths({01/01/91},1)
       Returns.....: Date
       Parameters..: dDate   = Date being tested for ...
                     nMonths = Number of months "ahead" or "behind"
                     bRetain = logical - retain time?
      -------------------------------------------------------------
   */
      local dRet, nNewMonth

		// Initialiaze return date to 01.01.1970 (Vasile Buza, April 2020)
		dRet = new Date(0)

      // Add months to current month, which is zero-based
      // (Use round() on arg just in case it's non-int)
      nNewMonth = dOld.month + round( nMonths, 0 )

      // Apply year offset, if any
      dRet.year = dOld.year + floor( nNewMonth / 12 )

      // Convert month number to valid 0-11 range
      nNewMonth %= 12
      // MOD() and % operator use INT(), so adjust when
      // month is negative
      if nNewMonth < 0
         nNewMonth += 12
      endif

      // Apply new month to Date object
      dRet.month := nNewMonth
      // Restore day (note property is named "date")
      dRet.date  := dOld.date

      // If new month gets nudged, day was out of range
      if dRet.month # nNewMonth
         // When no match, return last day of month. Month is
         // always nudged forward, so put it back
         dRet -= dRet.date
      endif

   // Convert to date type if required
   RETURN iif( bRetainTime, dRet, dttod( dRet ) )
   // - EoM: AddMonths2()

   FUNCTION AddYears( dDate, nYears )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 04/22/1994
       Notes.......: Finds same day as given date N years ahead. 
                     Using this function dBASE IV will take care 
                     of converting February 29 to March 1 if 
                     moving from a leap to a non-leap year.  
                     However, it may not be used backwards 
                     (negative value of nYears) since 
                     the date a year before February 29, 1992 
                     will be returned as March 1, 1991, not 
                     February 28, 1991. If you must move back, 
                     either check explicitly for February 29 as 
                     the original date or add code as in
                     the AddMonths() function to test for 
                     the date returned being of a different 
                     month than the original and, if it is, 
                     to subtract its day().
       Written for.: dBASE IV, 1.1
       Rev. History: 11/10/1991 - original function.
                     11/14/1991 - Ken Mayer - expanded out 
                        to make it easier to read, and see 
                        what's happening.
                     04/22/1994 - Jay Parsons - made 
                        international
                     11/02/1996 -- Tinkered with minimally 
                        to make part of custom class. Ken Mayer
       Calls.......: None
       Usage.......: DateEx.AddYears(<dDate>,<nYears>)
       Example.....: ? dMyDate.AddYears({01/01/91},1)
       Returns.....: Date
       Parameters..: dDate  = Date being tested for ...
                     dYears = Number of Years "ahead"
      -------------------------------------------------------------
   */
   
      local dReturn
		
		this.setTime(0)  // Initialiaze to 01.01.1970 (Vasile Buza, April 2020)

      this.year  := year( dDate ) + nYears
      this.month := month( dDate ) - 1 // starts at zero
      this.date  := day( dDate )
      // deal with century setting (either two digit or
      // four digit year)
      dReturn = CtoD( left( this.toLocaleString(), ;
                      iif( set("CENTURY") == "ON", 10, 8 ) ) )
      
   RETURN dReturn
   // -- EoM: AddYears()

   FUNCTION Age( dBirth )
   /*
      -------------------------------------------------------------
       Programmer..: Martin Leon
       Date........: 10/23/1991
       Notes.......: Returns age of thing as of date() given its 
                     birthdate.
       Written for.: dBASE IV, 1.1
       Rev. History: 10/23/1991 -- Original
                     08/10/1993 -- Ken Mayer, "stole" algorithm 
                      from the routine AGE2. It's more efficient.
                     Fall, 1994 -- revised for dBASE 5.0 for 
                      Windows. Jay Parsons, 72662,1302
                     11/02/1996 -- Tinkered with minimally to make 
                      part of custom class. Ken Mayer
       Calls.......: None
       Usage.......: DateEx.Age(<dBirth>)
       Example.....: ? "Joe is "+ltrim(str(dMyDate.Age(dBirth)))+;
                       " today ..."
       Returns.....: Numeric value in years
       Parameters..: dBirth = birthdate of thing to return age of.
      -------------------------------------------------------------
   */
   RETURN ( floor( ( val( dtos( date() ) ) ;
                   - val( dtos( dBirth ) ) ) / 10000 ) ) 
   // -- EoM: Age()

   FUNCTION Age2( d1, d2 )
   /*
      -------------------------------------------------------------
       Programmer..: Kenneth Chan 
       Date........: 09/25/1994
       Notes.......: Returns number of full years between two 
                     dates, which is the age of a person born on 
                     the first date as of the second.
                     Returns 0 if second date is less than a year 
                     after first date, -1 if second date is before 
                     first date by one year or less, and so forth.  
                     Does not check for blank dates.
       Written for.: dBASE IV, 1.1 and above
       Rev. History: 10/23/1991 -- Martin Leon ( HMAN ) - original 
                                   function
                     04/22/1992 -- Jay Parsons - description 
                                   modified and parameters changed
                     06/20/1993 -- Ken Chan - replaced, better 
                                   algorithm
                     07/01/1993 -- Jay Parsons - changed to use 
                                   floor() instead of int() for 
                                   coordination with WorkDays()
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/02/1996 -- Tinkered with minimally to 
                                   make part of custom class. 
                                   Ken Mayer
       Calls.......: None
       Usage.......: DateEx.Age2(<d1>,<d2>)
       Example.....: ? "A person born "+dtoc( d1 )+" was " ;
                       + dtoc( dMyDate.Age2( d1, d2 ) ) +;
                       " years old on " + dtoc( d2 )+"."
       Returns.....: Numeric value in years
       Parameters..: d1 = first date, such as date of birth
                     d2 = second date, when age is wanted
      -------------------------------------------------------------
   */
   RETURN ( floor( ( val( dtos(d2) ) ;
                  - val( dtos(d1) ) ) / 10000 ) )
   *-- EoM: Age2()

   FUNCTION Annivrsry( dTest, dBegin, dEnd )
   /*
      -------------------------------------------------------------
       Programmer..: David Love and Jay Parsons
       Date........: 11/10/1991
       Notes.......: Checks to see if an anniversary date falls 
                     within a range of dates (handy for mailings 
                     for organizations, checking to see if 
                     someone's birthday falls within certain 
                     dates, etc.)
       Written for.: dBASE IV, 1.1
       Rev. History: 11/10/1991 - Original Release
                     06/30/1992 - Jay Parsons - revised to work 
                       with AGE2
                     11/02/1996 -- Tinkered with minimally to 
                       make part of custom class. Ken Mayer
       Calls.......: dateex.Age2()
       Usage.......: dateex.Annivrsry(<dTest>,<dBegin>,<dEnd>)
       Example.....: if dMyDate.Annivrsry(dBDay,{03/01/91},{03/31/91})
                        *-- do something
                     endif
       Returns.....: true if a date (dTest), or its anniversary, 
                     falls within the period beginning at dBegin 
                     or ending at dEnd, inclusive.  Returns false 
                     for any other result, including invalid 
                     ranges or blank dates.
       Parameters..: dTest  = Date being tested for ...
                     dBegin = Beginning of range
                     dEnd   = End of range
      -------------------------------------------------------------
   */

      local nYears

      nYears = 0
      // This "if" test will fail if a date is blank
      if dBegin <= dEnd AND dTest <= dEnd
         nYears = CLASS::Age2( dTest, dEnd ) - ;
                  CLASS::Age2( dTest, dBegin - 1 )
      endif

   RETURN nYears # 0
   // -- EoM: Annivrsry()

   FUNCTION Dat2Jul( dDate )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 09/25/1994
       Notes.......: Converts dBASE date to Julian # of days (from 
                     January 1, 4713 B.C.)  This is the value of 
                     the float dBASE uses to store dates, and is 
                     used by astronomers.  Do not use for dBASE 
                     dates before 1582, or without considering 
                     time zones if using for astronomy.
       Rev. History: 03/01/1992 -- Original Release
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/02/1996 -- Tinkered with minimally to 
                                   make part of custom class. 
                                   Ken Mayer
                     01/26/1998 -- Jose update
       Written for.: dBASE IV
       Rev. History: None
       Calls.......: None
       Usage.......: dateex.Dat2Jul(<dDate>)
       Example.....: ? dMyDate.Dat2Jul(date())
       Returns.....: Numeric
       Parameters..: dDate = Date to convert to Julian ...
      -------------------------------------------------------------
   */
   RETURN ( 1721426 + dDate - ctod( "001/001/001" ) )
   // -- EoM: Dat2Jul

   FUNCTION DateText( dDate )
   /*
      -------------------------------------------------------------
       Programmer..: Miriam Liskin
       Date........: 09/25/1994
       Notes.......: Display date in format Month day, year (e.g.,
                     July 1, 1991)
       Written for.: dBASE IV, 1.1
       Rev. History: 05/23/1991 -- Original
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/02/1996 -- Tinkered with minimally to 
                                   make part of custom class. 
                                   Ken Mayer
       Calls.......: None          
       Usage.......: dateex.DateText(<dDate>)
       Example.....: ? dMyDate.DateText(date())
       Returns.....: July 1, 1991
       Parameters..: dDate = date to be converted
      -------------------------------------------------------------
   */

   RETURN ( cmonth(dDate) +" "+ltrim(str(day(dDate),2))+", "+;
                        str(year(dDate) ,4) )
   // -- EoM: DateText

   FUNCTION DateText2( dDate )
   /*
      -------------------------------------------------------------
       Programmer..: Miriam Liskin
       Date........: 09/25/1994
       Notes.......: Display date in format day-of-week, Month day, 
                     year
       Written for.: dBASE IV, 1.1
       Rev. History: 05/23/1991 -- Original
                     11/02/1996 -- Tinkered with minimally to make 
                                   part of custom class. Ken Mayer
       Calls.......: DateEx.DateText
       Usage.......: DateEx.DateText2(<dDate>)
       Example.....: ? dMyDate.DateText2(date())
       Returns.....: the date formatted as: Monday, July 1, 1991
       Parameters..: dDate = date to be converted
      -------------------------------------------------------------
   */

   RETURN ( cdow(dDate) +  ", " + CLASS::DateText(dDate) )
   // -- EoM: DateText2

   FUNCTION DateText3( dDate )
   /*
      -------------------------------------------------------------
       Programmer..: Miriam Liskin
       Date........: 09/25/1994
       Notes.......: Display date in format Month, year
       Written for.: dBASE IV, 1.1
       Rev. History: 05/21/1991 -- original function.
                     03/02/1992 -- This one is Douglas P. Saine's 
                      invention.  In his words: "I just removed 
                      the middle part looking for the day. For the 
                      things I do, I only need the month and year. 
                      (I work for a defense contractor, accuracy 
                      of dates has never been of great concern. 
                      <G>)"
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/02/1996 -- Tinkered with minimally to 
                                   make part of custom class. 
                                   Ken Mayer
       Calls.......: None
       Usage.......: DateEx.DateText3(<dDate>)
       Example.....: ? dMyDate.DateText3(date())
       Returns.....: July, 1991
       Parameters..: dDate = date to be converted
      -------------------------------------------------------------
   */

   RETURN ( cmonth(dDate)+", "+str(year(dDate),4) )
   // -- EoM: DateText3

   FUNCTION DoY( dDate )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 07/07/1993
       Notes.......: Returns the day of the year of a date (from 
                     beginning of the year).  Known as "Julian" 
                     date in government and military contexts.
       Written for.: dBASE IV, 1.1
       Rev. History: 11/10/1991 -- original function.
                     11/14/1991 -- Ken Mayer - expanded for 
                                   readability
                     07/07/1993 -- Jay Parsons - revised to use 
                                   Zak's FDoY
                     11/02/1996 -- Tinkered with minimally to 
                                   make part of custom class. 
                                   Ken Mayer
       Calls.......: DateEx.FDoY 
       Usage.......: DateEx.DoY(<dDate>)
       Example.....: ? dMyDate.DoY({01/01/91})
       Returns.....: Numeric value of day of year
       Parameters..: dDate  = Date being tested for ...
      -------------------------------------------------------------
   */

   RETURN dDate - CLASS::FDoY( dDate ) + 1
   // -- EoM: DoY()

   FUNCTION FDoD( dDate, cDay )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 08/02/1993
       Notes.......: First Date of Day of the week. This function 
                     gives the date of the first occurrence in 
                     the month of the given day of the week.
                     It returns a blank date if the day of week is 
                     invalid, but is not case sensitive. 
       Written for.: dBASE IV, 1.1
       Rev. History: 02/25/1992 - Original Release
                     08/02/1993 - Jay Parsons - revised to use 
                                  FDoM().
                     11/02/1996 - Tinkered with minimally to make 
                                  part of custom class. Ken Mayer
       Calls.......: DateEx.FDoM()
                     DateEx.NDoW()
       Usage.......: DateEx.FDoD(<dDate>,<cDay>)
       Example.....: ? dMyDate.FDoD(DATE(),"Tuesday")
       Returns.....: dBASE Date
       Parameters..: dDate  -- date to work from ...
                     cDay   -- Day of week to look for ...
      -------------------------------------------------------------
   */
      local dReturn, nDay
   
      nDay = CLASS::NDoW(cDay)
      dReturn = CLASS::FDoM( dDate )
   
   RETURN dReturn + mod( nDay + 7 - dow( dReturn ), 7 )
   // -- EoM: FDoD()

   FUNCTION FDoM( dArg )
   /*
      -------------------------------------------------------------
       Programmer..: Kenneth Chan 
       Date........: 01/05/1993
       Notes.......: First Day of Month 
       Written for.: dBASE IV, 1.5
       Rev. History: 01/05/1993 -- Original Release
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/02/1996 -- Tinkered with minimally to make 
                                   part of custom class. Ken Mayer
       Calls.......: None
       Usage.......: DateEx.FDoM(<dArg>)
       Example.....: ? dMyDate.FDoM(date())
       Returns.....: Date
       Parameters..: dArg = a Date argument -- function returns 
                            first day of the month of this date.
      -------------------------------------------------------------
   */

   RETURN ( ( dArg ) - day( dArg ) + 1 )
   // -- EoM: FDoM

   FUNCTION FDoY( dArg )
   /*
      -------------------------------------------------------------
       Programmer..: Kenneth Chan
       Date........: 01/05/1993
       Notes.......: Returns January 1 of the year of the date 
                     argument passed to it.
       Written for.: dBASE IV, 1.5
       Rev. History: 01/05/1993 -- Original Release
                     11/02/1996 -- Tinkered with minimally to make 
                                   part of custom class. Ken Mayer
       Calls.......: None
       Usage.......: DateEx.FDoY(<dArg>))
       Example.....: ? dMyDate.FDoY(date())
       Returns.....: January 1 of the year in dArg
       Parameters..: dArg = date data
      -------------------------------------------------------------
   */

      local dJan

      dJan = dArg - day( dArg ) + 1 - 28 * ( month( dArg ) - 1)

   RETURN dJan - day( dJan ) + 1
   // -- EoM: FDoY()

   FUNCTION FDoNextMonth( dDate )
   /*
      -------------------------------------------------------------
       Programmer..: Todd Barry  (Originally: FRSTNXTMTH())
       Date........: 11/02/1996
       Notes.......: Returns first day of next month
       Written for.: dBASE IV, 1.1
       Rev. History: 04/04/1992 - Original Release
                     02/25/1993 - Jay Parsons - shortened
                     08/02/1993 - Jay Parsons - conformed to use 
                                  FDoM()
                     Fall, 1994 - revised for dBASE 5.0 for Windows.
                                  Jay Parsons
                     11/02/1996 - Tinkered with minimally to make 
                                  part of custom class. Ken Mayer
       Calls.......: FDoM()
       Usage.......: DateEx.FDoNextMonth(<dDate>)
       Example.....: ? dMyDate.FDoNextMonth( date() )
       Returns.....: dBASE Date
       Parameters..: dDate  -- date to work from ...
      -------------------------------------------------------------
    */

   RETURN ( CLASS::FDoM( (dDate) - day(dDate) + 45 ) )
   // -- EoM: FDoNextMonth()

   FUNCTION FiscalYear
   /*
      -------------------------------------------------------------
       Programmer..: David Kanter
       Date........: 12/23/1996
       Notes.......: Returns the fiscal year for a given date.  
                     Defaults to current date and 10 as the 
                     beginning of the fiscal year
       Written for.: Visual dBase 5.5
       Rev. History: 12/23/1996 - original function.
       Calls.......: Nothing
       Usage.......: FiscalYear( dDate, nBeginMonth )
       Example.....: ? dMyDate.FiscalYear( {12/01/96}, 10 )   
                                                           && 1997
                     ? dMyDate.FiscalYear( {09/01/96}, 10 )   
                                                           && 1996
                     ? dMyDate.FiscalYear( {09/01/96} )  
                                                           && 1996
       Returns.....: Numeric fiscal year with century
       Parameters..: dDate     = Date being tested for (date())
                     nBegMonth = beginning month of the fiscal 
                                 year (10)
      -------------------------------------------------------------
   */
      parameters dDate, nBegMonth
      local nYear, nMonth
      
      // -- make parameters correct
      dDate = iif( pcount() = 0, date(), dDate )
      if type( "dDate" ) # "D"
         dDate = ctod( dDate )
      endif
      
      if pCount() < 2
         nBegMonth = 10
      elseif pcount() == 2 and type( "nBegMonth" ) # "N"
         nBegMonth = 10
      endif
      
      nYear = year( dDate )
      nMonth = month( dDate )
      
      if nMonth >= nBegMonth
         nYear = nYear + 1
      endif

   RETURN nYear
   // -- EoM: FiscalYear()

   FUNCTION IntlCDOW
   /*
      -------------------------------------------------------------
       Programmer..: Ken Mayer, Ivar B. Jessen
       Date........: March, 2004
       Notes.......: Function that returns the "long" name
                     of the day of the week based on a variety
                     of languages. Code is based heavily
                     on sample code written by Ivar B. Jessen,
                     and handles multiple language types.
                     It relies on startup code for the DateEx.cc
                     class. See also the IntlCDOWShort() method, 
                     which does the same thing, but returns 
                     the 'short' version of the name of 
                     the day of the week ...
       Written for.: dBASE Plus
       Rev. History: None
       Calls.......: None
       Usage.......: DateEx.IntlCDOW(dDate)
       Example.....: cMyDay = dMyDate.IntlCDOW(date())
       Returns.....: The character day of the week
       Parameters..: dDate = a date
      -------------------------------------------------------------
   */
      parameter dDate
      local bufaAdr, bufaLen, cStr, nDoW

      // error trap
      if type("dDate") # "D" and type( "dDate" ) # "O"
         return null
      endif

      // the code Ivar provided 
      // returns Sunday as day 7,
      // and Monday as day 1, so we 
      // have to modify the value returned
      // by the dow() function:
      nDow = dow( dDate ) - 1
      if nDow == 0
         nDow := 7
      endif

      this.localeID = this.asLang[this.Language]
      
      bufAdr = space(64)
      bufLen = len(bufAdr)

      cStr = GetLocaleInfoA(this.localeID, 41 + nDow, bufAdr, bufLen)  
      this.dayLong = substr(bufAdr, 1, cStr - 1)

   return this.dayLong
   // -- EoM: IntlCDoW()

   FUNCTION IntlCDOWShort
   /*
      -------------------------------------------------------------
       Programmer..: Ken Mayer, Ivar B. Jessen
       Date........: March, 2004
       Notes.......: Function that returns the "short" name
                     of the day of the week based on a variety
                     of languages. Code is based heavily
                     on sample code written by Ivar B. Jessen,
                     and handles multiple language types.
                     It relies on startup code for the DateEx.cc
                     class. See also the IntlCDOW() method, 
                     which does the same thing, but returns the 
                     'long' version of the name of the day of 
                     the week ...
       Written for.: dBASE Plus
       Rev. History: None
       Calls.......: None
       Usage.......: DateEx.IntlCDOWShort(dDate)
       Example.....: cMyDay = dMyDate.IntlCDOWShort(date())
       Returns.....: The character day of the week
       Parameters..: dDate = a date
      -------------------------------------------------------------
   */
      parameter dDate
      local bufaAdr, bufaLen, cStr, nDoW

      // error trap
      if type("dDate") # "D" and type( "dDate" ) # "O"
         return null
      endif

      // the code Ivar provided 
      // returns Sunday as day 7,
      // and Monday as day 1, so we 
      // have to modify the value returned
      // by the dow() function:
      nDow = dow( dDate ) - 1
      if nDow == 0
         nDow := 7
      endif

      this.localeID = this.asLang[this.Language]
     
      bufAdr = space(64)
      bufLen = len(bufAdr)

      cStr = GetLocaleInfoA(this.localeID, 48 + nDow, bufAdr, bufLen)  
      this.dayShort = substr(bufAdr, 1, cStr - 1)

   return this.dayShort
   // -- EoM: IntlCDoWShort()

   FUNCTION IntlCMON
   /*
      -------------------------------------------------------------
       Programmer..: Ken Mayer, Ivar B. Jessen
       Date........: March, 2004
       Notes.......: Function that returns the "long" name
                     of the month based on a variety
                     of languages. Code is based heavily
                     on sample code written by Ivar B. Jessen,
                     and handles multiple language types.
                     It relies on startup code for the DateEx.cc
                     class. See also the IntlCMonShort() method, 
                     which does the same thing, but returns the 
                     'short' version of the name of month ...
       Written for.: dBASE Plus
       Rev. History: None
       Calls.......: None
       Usage.......: DateEx.IntlCMON(dDate)
       Example.....: cMyMonth = dMyDate.IntlCMON(date())
       Returns.....: The character name of the month 
       Parameters..: dDate = a date
      -------------------------------------------------------------
   */
      parameter dDate
      local bufaAdr, bufaLen, cStr, nMonth

      // error trap
      if type("dDate") # "D" and type( "dDate" ) # "O"
         return null
      endif 

      // determine the month:
      nMonth = month( dDate ) 

      this.localeID = this.asLang[this.Language]
      
      bufAdr = space(64)
      bufLen = len(bufAdr)

      cStr = GetLocaleInfoA(this.localeID, 55 + nMonth, bufAdr, bufLen)  
      this.monthLong = substr(bufAdr, 1, cStr - 1)

   return this.monthLong
   // -- EoM: IntlCMonth() 

   FUNCTION IntlCMONShort
   /*
      -------------------------------------------------------------
       Programmer..: Ken Mayer, Ivar B. Jessen
       Date........: March, 2004
       Notes.......: Function that returns the "short" name
                     of the month based on a variety
                     of languages. Code is based heavily
                     on sample code written by Ivar B. Jessen,
                     and handles multiple language types.
                     It relies on startup code for the DateEx.cc
                     class. See also the IntlCMon() method, which 
                     does the same thing, but returns the 'long' 
                     version of the name of month ...
       Written for.: dBASE Plus
       Rev. History: None
       Calls.......: None
       Usage.......: DateEx.IntlCMONShort(dDate)
       Example.....: cMyMonth = dMyDate.IntlCMONShort(date())
       Returns.....: The character name of the month 
       Parameters..: dDate = a date
      -------------------------------------------------------------
   */
      parameter dDate
      local bufaAdr, bufaLen, cStr, nMonth

      // error trap
      if type("dDate") # "D" and type( "dDate" ) # "O"
         return null
      endif

      // determine the month:
      nMonth = month( dDate ) 

      this.localeID = this.asLang[this.Language]
      
      bufAdr = space(64)
      bufLen = len(bufAdr)

      cStr = GetLocaleInfoA(this.localeID, 67 + nMonth, bufAdr, bufLen)  
      this.monthShort = substr(bufAdr, 1, cStr - 1)

   return this.monthShort
   // -- EoM: IntlCMonthShort() 

   Function IntlLongFormat
      /*
         -------------------------------------------------------------
          Programmer..: Marc van der Bergen
          Date........: March, 2004
          Notes.......: Returns date in format based on values returned
                        from windows for date formats, and language ...
          Written for.: dBASE Plus
          Rev. History: None
          Calls.......: None
          Usage.......: DateEx.IntlLongFormat(dDate)
          Example.....: cMyDate = dMyDate.IntlLongFormat(date())
          Returns.....: The character version of date in 'long' format
                        based on what Windows thinks is appropriate
                        for any defined language
          Parameters..: dDate = a date
         -------------------------------------------------------------
      */
      parameter dDate
      local bufDate, bufLen, nLen, cStr, cResult

      if type( "dDate" ) # "D" and type( "dDate" ) # "O"
         return null
      endif

      if type('GetDateFormatA') # 'FP'
         extern cint GetDateFormatA(cint,cint,cptr,cptr,cstring,cint) kernel32
      endif

      cStr=replicate(chr(0),16)  // this is the wrapper for the structure

      // now fill it with the date values :  year
      cStr.setByte(0, bitand(dDate.year, 0xFF))
      cStr.setByte(1, bitrshift(bitand(dDate.Year, 0xFF00), 8))

      // month
      cStr.setByte(2, bitand(dDate.month+1, 0xFF))
      cStr.setByte(3, bitrshift(bitand(dDate.month+1, 0xFF00), 8))

      // date
      cStr.setByte(6, bitand(dDate.date, 0xFF))
      cStr.setByte(7, bitrshift(bitand(dDate.date, 0xFF00), 8))

      // Prepare the output buffer
      bufDate=space(64)
      bufLen=len(bufDate)

      // note, we are getting the localeID from
      // the array in the DateEx class, using the Language
      // specified by the user:
      this.localeID = this.asLang[this.Language]

      // Now do the API-Call to fill the output buffer with the long date
      nLen = GetDateFormatA(this.localeID,2,cstr,0,bufDate,bufLen)
      cResult = substr(bufDate, 1, nLen - 1)

   return cResult
   // EoM: IntlLongFormat()

   Function IntlShortFormat
      /*
         -------------------------------------------------------------
          Programmer..: Marc van der Bergen
          Date........: March, 2004
          Notes.......: Returns date in format based on values returned
                        from windows for date formats, and language ...
                        This is the short format (MM/DD/YYYY, for 
                        example)
          Written for.: dBASE Plus
          Rev. History: None
          Calls.......: None
          Usage.......: DateEx.IntlShortFormat(dDate)
          Example.....: cMyDate = dMyDate.IntlShortFormat(date())
          Returns.....: The character version of date in 'short' format
                        based on what Windows thinks is appropriate
                        for any defined language
          Parameters..: dDate = a date
         -------------------------------------------------------------
      */
      parameter dDate
      local bufDate, bufLen, nLen, cStr, cResult

      if type( "dDate" ) # "D" and type( "dDate" ) # "O"
         return null
      endif

      if type('GetDateFormatA') # 'FP'
         extern cint GetDateFormatA(cint,cint,cptr,cptr,cstring,cint) kernel32
      endif

      cStr=replicate(chr(0),16)  // this is the wrapper for the structure

      // now fill it with the date values :  year
      cStr.setByte(0, bitand(dDate.year, 0xFF))
      cStr.setByte(1, bitrshift(bitand(dDate.Year, 0xFF00), 8))

      // month
      cStr.setByte(2, bitand(dDate.month+1, 0xFF))
      cStr.setByte(3, bitrshift(bitand(dDate.month+1, 0xFF00), 8))

      // date
      cStr.setByte(6, bitand(dDate.date, 0xFF))
      cStr.setByte(7, bitrshift(bitand(dDate.date, 0xFF00), 8))

      // Prepare the output buffer
      bufDate=space(64)
      bufLen=len(bufDate)

      // note, we are getting the localeID from
      // the array in the DateEx class, using the Language
      // specified by the user:
      this.localeID = this.asLang[this.Language]

      // Now do the API-Call to fill the output buffer with the long date
      nLen = GetDateFormatA(this.localeID,1,cstr,0,bufDate,bufLen)
      cResult = substr(bufDate, 1, nLen - 1)

   return cResult
   // EoM: IntlLongFormat()


   Function IntlSetDate
   /*
      -------------------------------------------------------------
       Programmer..: Ken Mayer, Ivar B. Jessen
       Date........: March, 2004
       Notes.......: Sets the date format using the three primary
                     formats: 
                          MDY
                          DMY
                          YMD
                     Uses dBASE SET DATE to do it, from that point
                     on date formats (until reset) will use this
                     format. This is based on the language property
                     of the date object
       Written for.: dBASE Plus
       Rev. History: None
       Calls.......: None
       Usage.......: DateEx.IntlSetDate()
       Example.....: // Save original date and mark settings:
                     cOldDate = set("date")
                     cOldMark = set("mark")
                     // Change them:
                     dMyDate.IntlSetDate()
                     dMyDate.IntlSetMark()
                     // do whatever you need to with dates
                     // in the language format ...
                     // Reset the date and mark settings:
                     cCmd = "set date "+cOldDate
                     &cCmd.
                     set mark to cOldMark
       Returns.....: None
       Parameters..: None
      -------------------------------------------------------------
   */
      local bufAdr, bufLen, localeID, nOffset, cStr, nResult

      if type('GetLocaleInfoA') # 'FP'
         extern CINT GetLocaleInfoA( CLONG, CINT, CSTRING, CINT) kernel32
      endif

      bufAdr = space(64)
      bufLen = len(bufAdr)

      // note, we are getting the localeID from
      // the array in the DateEx class, using the Language
      // specified by the user:
      this.localeID = this.asLang[this.Language]
    
      nOffset =  0x00000021 // Short Date format

      cStr = GetLocaleInfoA(this.localeID, nOffset, bufAdr, bufLen)  
      nResult = substr(bufAdr, 1, cStr - 1) 

      // Based on nResult, we set the date format:
      if nResult == 0
         set date MDY
      elseif nResult == 1
         set date DMY
      else // nResult == 2
         set date YMD
      endif

   return
   // EoM: IntlSetDate()

   Function IntlSetMark
   /*
      -------------------------------------------------------------
       Programmer..: Ken Mayer, Ivar B. Jessen
       Date........: March, 2004
       Notes.......: Sets the character used as a marker between
                     elements of a date, based on the language
                     property of the DateEx class. This uses
                     the dBL command:
                         set mark to "x"
                     where "x" is the character(s) used as the
                     separator. This can (and should) be used
                     in conjunction with IntlShortFormat().
       Written for.: dBASE Plus
       Rev. History: None
       Calls.......: None
       Usage.......: DateEx.IntlSetMark()
       Example.....: // Save original date and mark settings:
                     cOldDate = set("date")
                     cOldMark = set("mark")
                     // Change them:
                     dMyDate.IntlShortFormat()
                     dMyDate.IntlSetMark()
                     // do whatever you need to with dates
                     // in the language format ...
                     // Reset the date and mark settings:
                     cCmd = "set date "+cOldDate
                     &cCmd.
                     set mark to cOldMark
       Returns.....: None
       Parameters..: None
      -------------------------------------------------------------
   */
      local bufAdr, bufLen, localeID, nOffset, cStr, cMark

      if type('GetLocaleInfoA') # 'FP'
         extern CINT GetLocaleInfoA( CLONG, CINT, CSTRING, CINT) kernel32
      endif

      bufAdr = space(64)
      bufLen = len(bufAdr)

      // note, we are getting the localeID from
      // the array in the DateEx class, using the Language
      // specified by the user:
      this.localeID = this.asLang[this.Language]
    
      nOffset =  0x0000001D // Short Date format

      cStr = GetLocaleInfoA(this.localeID, nOffset, bufAdr, bufLen)  
      cMark = substr(bufAdr, 1, cStr - 1) 

      // now we set the mark character(s):
      set mark to cMark

   return
   // EoM: IntlShortFormat()

   function ChangeLanguage
   /*
      -------------------------------------------------------------
       Programmer..: Ken Mayer
       Date........: March 8, 2004
       Notes.......: Brings up a dialog allowing the user
                     to see the options for the languages
                     used by the International date methods 
                     above.
       Written for.: dBASE Plus 2.2
       Rev. History: None
       Calls.......: ChangeLanguageForm (at end of this class)
       Usage.......: dMyDate.ChangeLanguage()
       Example.....: dMyDate.ChangeLanguage()
       Returns.....: None
       Parameters..: None
       Effects.....: If 'cancel' button is selected, none.
                     If 'accept' button is selected, the
                     language property of the dateEx class
                     is changed for the current instance.
      -------------------------------------------------------------
   */
      local fCL
      fCL = new ChangeLanguageForm()
      fCL.language = this.Language
      fCL.readModal()

      // once form returns:
      this.Language := fCL.Language

      // release the form from memory:
      fCL.release()

   return
   // -- EoM: ChangeLanguage()

   FUNCTION IsLeap( nYear )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 08/03/1993
       Notes.......: Is the year given a Leap Year? Year given must
                     be after 1500 C.E.
       Written for.: dBASE IV, 1.1
       Rev. History: 11/08/1991 - Jay Parsons - original function
                     01/13/1992 - Jay Parsons - updated to handle 
                                  two-digit OR four-digit year.
                     08/03/1993 - Jay Parsons - removed change to 
                                  parameter
                     Fall, 1994 - revised for dBASE 5.0 for Windows.
                                  Jay Parsons
                     11/02/1996 - Tinkered with minimally to make 
                                  part of custom class. Ken Mayer
                     01/26/1998 -- Jose (set epoch) fix
       Calls.......: None
       Usage.......: DateEx.IsLeap(<nYear>)
       Example.....: ? dMyDate.IsLeap(91)
       Returns.....: Logical (true/false) 
       Parameters..: nYear  = Numeric form of year -- last two 
                              digits
                              (i.e., 91), or all four digits 
                              (i.e., 1991)
      -------------------------------------------------------------
   */
      local nYr, lReturn, nDec, nCen
   
      // (Jose's fix) Deal with two-digit year, and/or Set Epoch ...
      nYr = nYear
      if nYr < 100
         nCen = set( "EPOCH" )
         nDec = mod( nCen, 100 )
         nCen -= nDec
         nYr := nYr + iif( nYr => nDec, nCen, nCen + 100 )
      endif
   
      lReturn = mod( iif( mod( nYr, 100 ) == 0, ;
                      nYr / 100, nYr ), 4 ) = 0
      
   RETURN lReturn
   // -- EoM: IsLeap()

   FUNCTION Jul2Dat( nJulian )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 09/25/1994
       Notes.......: Converts Julian # of days to dBASE Date
       Rev. History: 03/01/1992 -- Original Release
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/02/1996 -- Tinkered with minimally to 
                                   make part of custom class. KM
       Written for.: dBASE IV
       Rev. History: None
       Calls.......: None
       Called by...: Any
       Usage.......: DateEx.Jul2Dat(nJulian)
       Example.....: ? dMyDate.Jul2Dat(2448691)
       Returns.....: Date
       Parameters..: nJulian = Julian date to convert to dBase Date
      -------------------------------------------------------------
   */

   RETURN ( ctod( "001/001/001" ) + (nJulian) - 1721426)
   // -- EoM: Jul2Dat()

   FUNCTION LDoD( dDate, cDay )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 08/02/1993
       Notes.......: Last Date of Day. This function works to give 
                     the date of the last occurrence in the month 
                     of the given given day of the week (i.e., 
                     last Monday of the month). It returns a blank 
                     date if the day of week is invalid, but is 
                     not case sensitive.
       Written for.: dBASE IV, 1.1
       Rev. History: 02/25/1992 - Original Release
                     08/02/1993 - Jay Parsons - declared nDay 
                                  private.
                     11/02/1996 - Tinkered with minimally to make 
                                  part of custom class. Ken Mayer
       Calls.......: DateEx.LDoM()
                     DateEx.NDoW()
       Usage.......: DateEx.LDoD(<dDate>,"<cDay>")
       Example.....: ? dMyDate.LDoD(DATE(),"Tuesday")
       Returns.....: dBASE Date
       Parameters..: dDate  -- date to work from ...
                     cDay   -- Day of week to look for ...
      -------------------------------------------------------------
   */
      local dReturn, nDay
      
      nDay = CLASS::NDoW( cDay )
      dReturn = CLASS::LDoM( dDate )
         
   RETURN dReturn - mod( dow( dReturn ) + 7 - nDay, 7 )
   // -- EoM: LDoD()

   FUNCTION LDoM( dDate )
   /*
      -------------------------------------------------------------
       Programmer..: Kenneth Chan 
       Date........: 02/26/1992
       Notes.......: Last Date of Month. Returns the date of the 
                     last day of the month of the given date.
       Written for.: dBASE IV, 1.1
       Rev. History: 02/26/1992 -- Original Release
                     11/02/1996 -- Tinkered with minimally to make 
                                   part of custom class. Ken Mayer
       Calls.......: None
       Usage.......: DateEx.LDoM(<dDate>)
       Example.....: ? dMyDate.LDoM(DATE())
       Returns.....: dBASE Date
       Parameters..: dDate  -- date to work from ...
      -------------------------------------------------------------
   */
      local dNxtMonth
   
      dNxtMonth = dDate - day( dDate ) + 45 // middle of next month
   
   RETURN dNxtMonth - day( dNxtMonth )
   // -- EoM: LDoM()

   FUNCTION LDoY( dArg )
   /*
      -------------------------------------------------------------
       Programmer..: Kenneth Chan
       Date........: 01/05/1993
       Notes.......: Returns December 31 of year in date argument 
                     passed to function.
       Written for.: dBASE IV, 1.5
       Rev. History: 01/05/1993 -- Original Release
                     11/02/1996 -- Tinkered with minimally to make 
                                   part of custom class. Ken Mayer
       Calls.......: DateEx.LDoM() 
       Usage.......: DateEx.LDoY(<dArg>)
       Example.....: ? dMyDate.LDoY(Date())
       Returns.....: Last Day of Year
       Parameters..: dArg = Date 
      -------------------------------------------------------------
   */   
      local dDec
      
      dDec = dArg - day( dArg ) + 28 * ( 13 - month( dArg ))

   RETURN CLASS::LDoM( dDec )
   // -- EoM: LDoY()

   FUNCTION NDoW( cDay )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons  
       Date........: 09/25/1994
       Notes.......: Numeric Day of Week -- returns the numeric dow 
                     value of a named day of the week for use by 
                     some of the other date routines in the 
                     library.
       Written for.: dBASE IV, 1.1
       Rev. History: 02/25/1992 - original function.
                     04/22/1992 - Jay Parsons - modified example 
                                  and descriptions, added ltrim() 
                                  of argument.
                     Fall, 1994 - revised for dBASE 5.0 for Windows.
                                  Jay Parsons
                     11/02/1996 - Tinkered with minimally to make 
                                  part of custom class. Ken Mayer
       Calls.......: None
       Usage.......: NDoW(<cDay>)
       Example.....: nDay = oMyString.NDoW("Tues")
       Returns.....: Numeric dow value of day of week given
       Parameters..: cDay  -- Character memvar containing "day" of 
                              week ('MONDAY', etc ...)
      -------------------------------------------------------------
   */

   RETURN ( at( upper( left( ltrim( cDay ), 3 ) ), ;
          "   SUN MON TUE WED THU FRI SAT" ) / 4 )
   // -- EoM: NDoW

   FUNCTION Num2Dat( nYyMmDd )
   /*
      -------------------------------------------------------------
       Programmer..: Raymond S. Leventhal
       Date........: 04/22/1994
       Notes.......: Changes Date format from *numeric* yymmdd to 
                     dBASE date format. The author uses an AS/400 
                     and downloads data. The YYMMDD format is the 
                     date format used by the AS/400.
      
                     As written, this function will not work for 
                     dates in the first decade of a century, such 
                     as January 1, 2000, because the numeric 
                     representation of the date will not have six 
                     digits. This appears to be an AS/400 problem. 
                     To convert the function to make sense of a 
                     date like 101 meaning January 1, 2000, use 
                     int( nYyMmDd/10000 ) as the year and the 
                     result of int( mod( nYyMmDd, 10000 ) / 100 ) 
                     as the month.
       Rev. History: 06/04/1992 - Original version
                     04/22/1994 - Jay Parsons - made independent 
                                  of date set() setting
                     11/02/1996 - Tinkered with minimally to make 
                                  part of custom class. Ken Mayer
                     10/20/1998 - Modified to use date properties
                                  to avoid internationalization
                                  issues ...
       Calls.......: None
       Usage.......: DateEx.Num2Dat(<nYyMmDd>)
       Example.....: ? oMyDate.Num2Dat(930820)
       Returns.....: Date format value of nYyMmDd (i.e. 08/20/93)
       Parameters..: nYyMmDd = Numeric value YyMmDd to be converted 
                               to date
      -------------------------------------------------------------
   */   
      local cS, dDate
		
		cS	= str(nYyMmDd, 8, 0 , '0')
		// sequence altered  (Vasile Buza, April 2020)
		this.setTime( new date( ;
			val( substr( cS , 1, 4 ) ), ;			// year
			val( substr( cS , 5, 2 ) ) - 1, ;	// month
			val( substr( cS , 7, 2 ) ) ;			// date (day)
						).gettime() )

      dDate = CtoD( left( this.toLocaleString(), ;
                    iif( set("CENTURY") == "ON", 10, 8 ) ) )
   RETURN dDate
   // -- EoM: Num2Dat()

   FUNCTION NumDoD( dDate, nWeek, cDay )
   /*
      -------------------------------------------------------------
       Programmer..: Ken Mayer
       Date........: 02/24/1992
       Notes.......: Numbered Date of a Day. This function will 
                     return the occurrence numbered by its second 
                     argument of the day of the week given as its 
                     third argument within the month in which the 
                     date given as the first argument occurs.
                     Example: what if you need the third Monday 
                     of the month?
                     Send to this function a date (any date) of 
                     the month, the number you need (first, 
                     second...) and the day you need. The function 
                     is not case sensitive.
       Written for.: dBASE IV, 1.1
       Rev. History: 02/24/1992 -- Original Release
                     11/02/1996 -- Tinkered with minimally to 
                                   make part of custom class. KM
       Calls.......: DateEx.FDoD()
       Usage.......: DateEx.NumDoD(<dDate>,<nWeek>,<cDay>)
       Example.....: ? dMyDate.NumDoD({02/03/92},3,"Monday")
       Returns.....: Date
       Parameters..: dDate  =  Any date of the month (and year) 
                               needed 
                     nWeek  =  Number of week you need (i.e., 
                               third cDay of the month
                     cDay   =  Character name of day ("Monday", 
                               etc.)
      -------------------------------------------------------------
   */
      local dReturn
      
      dReturn = CLASS::FDoD( dDate, cDay ) + 7 * ( nWeek - 1 )
      
   RETURN dReturn
   // -- EoM: NumDoD()

   FUNCTION Quarter( dDate )
   /*
      -------------------------------------------------------------
       Programmer..: Bowen Moursund 
       Date........: 09/25/1994
       Notes.......: Returns the quarter of the year of a specific 
                     date ...
       Written for.: dBASE IV, 1.1
       Rev. History: 02/03/1992 -- Original Release
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/02/1996 -- Tinkered with minimally to make
                                   part of custom class. Ken Mayer
                     August, 2006 -- Modified by Rich Muller
                                   and Ivar B. Jessen for
                                   more international date
                                   compatibility.
       Calls.......: None
       Usage.......: DateEx.Quarter(<dDate>)
       Example.....: ? dMyDate.Quarter({05/25/1992})
       Returns.....: Numeric (integer) value from 1 to 4 (or 0 on 
                         error )
       Parameters..: dDate = date to be checked
      -------------------------------------------------------------
   */
  RETURN ( iif( type("Argvector(1)")=="D", ;
           int( ceiling(month(dDate)/3) ), 0) )
   // -- EoM: Quarter()

   FUNCTION StoD( cString )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/20/1998
       Notes.......: Convert string YYYYMMDD or YYMMDD to a date 
                     regardless of SET DATE.
       Written for.: dBASE IV, 1.1
       Rev. History: 11/10/1991 - Original Release
                     08/02/1993 - Jay Parsons - cS added to 
                                  eliminate changes to parameter
                     04/22/1994 - Jay Parsons - greatly simplified 
                                  to use Romain Strieff's technique
                                  of changing the date setting.
                     11/02/1996 - Tinkered with minimally to make 
                                  part of custom class. Ken Mayer
                     10/20/1998 - Modified to use date class
                                  to avoid internationalization
                                  issues ...
       Calls.......: None
       Usage.......: DateEx.StoD(<cString>)
       Example.....: ? dMyDate.StoD("19910101")
       Returns.....: Date
       Parameters..: <cString> = Date string you wish converted to 
                                 "normal" dBASE date. Must be in 
                                 either YYYYMMDD or YYMMDD format.
      -------------------------------------------------------------
   */   
      local cS, dReturn
		
      cS = ltrim( cString )
      if len( cS ) < 8
         cS = right( '00000000' + cS , 8 )
      endif
      
		// sequence altered (Vasile Buza, April 2020)
		this.setTime( new date( ;
					val( substr( cS , 1, 4 ) ), ;			// year
					val( substr( cS , 5, 2 ) ) - 1, ;	// month
					val( substr( cS , 7, 2 ) ) ;			// date (day)
								).gettime() )

      // deal with century setting (either two digit or
      // four digit year)
      dReturn = CtoD( left( this.toLocaleString(), ;
                      iif( set("CENTURY") == "ON", 10, 8 ) ) )
    
   RETURN dReturn
   // -- EoM: StoD()

   FUNCTION Text2Date( cDate )
   /*
      -------------------------------------------------------------
       Programmer..: Angus Scott-Fleming
       Date........: 10/20/1998
       Notes.......: Convert string from dd-mmm-yy to date
       Written for.: dBASE IV, 1.5+
       Rev. History: 04/05/1994  first pass
                     11/02/1996 -- Tinkered with minimally to 
                                   make part of custom class. KM
                     10/20/1998 - Modified to use date class
                                  to avoid problems with
                                  internationalization ... (KM)
                     May, 2004 -- Modified to deal with a problem
                                  with February, and assigning the
                                  values to the date class in 
                                  sequence. For some reason, if
                                  assigned in the wrong sequence, 
                                  the date returned in February comes
                                  back as in March. This was found
                                  by Niall Mulcahy. 

                                  Also took to heart the comment
                                  about modifying the code to
                                  deal with months such as "SEPT",
                                  so modified the handling of the
                                  month to use an array. This should
                                  handle JUN and JUNE; JUL and JULY;
                                  SEP and SEPT; APR and APRIL ... KJM
       Calls.......: None
       Usage.......: dateex.Text2Date(<string>)
       Example.....: ? dMyDate.Text2Date("02-Feb-94")
       Returns.....: date, or null date for bad input
       Parameters..: cDate = string to be converted
      -------------------------------------------------------------
   */      
      local nFirstDash, nLastDash, cDay, cMonth, cYear, nDay, ;
            nMonth, nYear, dReturn, aMonths, nElement, nRow, i
       
      // assumes date is of form 0d-Mmm-yy or 0d-Mmm-yyyy
      // check input for two hyphens
      nFirstDash = at("-",cDate)
      nLastDash  = at("-",cDate,2)
		
		if	nFirstDash == 0						// assumes date is of form 0d/Mmm/yy or 0d/Mmm/yyyy
			nFirstDash = at("/",cDate)
			nLastDash  = at("/",cDate,2)
		endif
		
		if	nFirstDash == 0						// assumes date is of form 0d.Mmm.yy or 0d.Mmm.yyyy
			nFirstDash = at(".",cDate)
			nLastDash  = at(".",cDate,2)
		endif
		
      if nFirstDash = 0 or nLastDash = 0
         // bogus input - return an empty date
         return {}
      endif
      // extract characters for day, month, and year
      cDay = left(cDate,nFirstDash - 1)
      cMonth = substr(cDate, nFirstDash + 1, ;
               nLastDash - nFirstDash - 1)
      cYear  = substr(cDate, nLastDash + 1)
      nDay   = val(cDay)

      // Code expanded to deal with options such as SEPT, instead
      // of just SEP ...
		
		aMonths = { "JAN" , "FEB"	, "MAR"	, "APR"	, "APRIL"	,	"MAY"	, "JUN"	, "JUNE"	, "JUL"	, "JULY"	, "AUG"	, "SEP"	, "SEPT"	, "OCT"	, "NOV"	, "DEC"	}
		nMonths = {	 1		,	2		,	3		,	4		,	4			,	5		,	6		,	6		,	7		,	7		,	8		,	9		,	9		,	10		,	11		,	12		}
		
      // find the month in the array:
		
      nElement = aMonths.scan( UPPER( left(cMonth,4) ) )
		if	nElement == 0
			nElement = aMonths.scan( UPPER( left(cMonth,3) ) )
		elseif nElement == 0
			return {}
		endif
		
		nMonth = nMonths[ nElement ]

      // Year
      nYear  = val(cYear)

      // validate numbers
      do case
         case nDay <= 0 or nMonth <= 0 or nYear <= 0
              // bogus input - return an empty date
              return {}
         case nDay > 31 or nMonth > 12
              // bogus input - return an empty date
              // could expand this checking to check for Feb 30 
              // and so forth
              return {}
         otherwise
			   // sequence altered (Vasile Buza, April 2020)
				this.setTime( new date(nYear, -1 + nMonth, nDay).getTime() )
      endcase

      // deal with century setting (either two digit or
      // four digit year)
      dReturn = CtoD( left( this.toLocaleString(), ;
                      iif( set("CENTURY") == "ON", 10, 8 ) ) )
      
   RETURN dReturn
   // -- EoM: Text2Date()


   FUNCTION WeekNo( dDate, nFDoW )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 05/26/1994
       Notes.......: Returns the week number of the year of a date 
                     (from beginning of the year). By default, week
                     #1 begins on the first Sunday of the year. 
                     Days in any partial week before that date are 
                     assigned to week #0. To use this function but 
                     start week # 1 on a different day, use the 
                     second parameter to specify the day of the 
                     week the first occurrence in the year of 
                     which starts week #1, 2 for Monday through 7 
                     for Saturday.
       Written for.: dBASE IV, 1.1
       Rev. History: 11/10/1991 - original function.
                     11/14/1991 - Ken Mayer - expanded for 
                                  readability 
                     05/26/1994 - Frank A. Deviney, Jr. - added 
                                  second parameter
                     11/02/1996 - Tinkered with minimally to make 
                                  part of custom class. Ken Mayer
       Calls.......: DateEx.DoY()
       Usage.......: DateEx.WeekNo(<dDate>,<nFDoW>)
       Example.....: ? dMyDate.WeekNo({01/01/91})   = 0
                     ? dMyDate.Weekno({01/01/91},3) = 1
                     ? dMyDate.Weekno({01/08/91})   = 1
                     ? dMyDate.Weekno({01/08/91},;
                                          dow(FDoY({01/08/91)))= 2
       Returns.....: Numeric value of week number
       Parameters..: dDate  = Date being tested for ...
                     nFDoW  = first day of week [1..7]
      -------------------------------------------------------------
   */
      
      local dBaseDate, nFD

      nFD = iif (pcount() < 2, 1, nFDoW )
      dBaseDate = dDate - CLASS::DoY(dDate)
      dBaseDate = dBaseDate - mod( dow( dBaseDate - nFD ), 7 )

   RETURN int( (dDate - dBaseDate) / 7 )
   // -- EoM: WeekNo()


   // ------------------------------------------------------------
   // The following methods all deal with European Week Numbers:
   // ------------------------------------------------------------
   FUNCTION EweekNo(dDate)
   /*
      -------------------------------------------------------------
       Programmer..: Ivar B. Jessen
       Date........: 01 Oct 1999
       Notes.......: Returns the number of the week for a given date 
                     based on European Norm EN 28601:1993, according 
                     to which the first day of a week is a Monday and
                     week no.1 of a year is the week containing the
                     fourth of January.
                     Week no.1 thus contains at least four days of
                     the new year and always contains the first
                     Thursday of the year.
       Written for.: Visual dBase 7.01
       Rev. History: 01 Oct 1999
       Calls.......: dateex.FDoEW() // below
       Usage ......: d.EweekNo(dDate)
       Example.....: ? d.EweekNo({03-01-99})    //  1
                     ? d.EweekNo({04-01-1999})  //  1
                     ? d.EweekNo({25/11/99})    // 47
                     ? d.EweekNo({03.01.2000})  //  1
       Returns.....: Numerical number of week
       Parameters..: dDate.  dDate is type date
      -------------------------------------------------------------
   */
      local dFDoY      // Date of first day of input year
      local dFDoNexY   // Date of first day of next year 
      local dFDoPrvY   // Date of first day of previous year 
      local nRtv       // Return value
      local dFDoFW     // Date of first day of first week in input year
      local dLDoLW     // Date of last  day of last  week in input year
      local dFDoFWPrvY // Date of first day of first week of previous year

      dFDoY      = class::FDoY( dDate )
      dFDoNexY   = class::FDoY( dFDoY + 366 )
      dFDoPrvY   = class::FDoY( dFDoY - 1 )
      dFDoFW     = class::FDoEW( dFDoY )
      dLDoLW     = class::FDoEW( dFDoNexY ) - 1 
      dFDoFWPrvY = class::FDoEW(( dFDoY) - 1 )

      do case
         case dDate < dFDoFW   
            nRtv = int(( dFDoFW - 1 - dFDoFWPrvY ) / 7 ) + 1 
         case dDate > dLDoLW   
            nRtv = 1              
         otherwise
            nRtv = int(( dDate-dFDoFW ) / 7 ) + 1 
      endcase

   RETURN nRtv
   // -- EoM: EWeekNo()

   FUNCTION FDoEW( dateY ) 
   /*
      -------------------------------------------------------------
       Programmer..: Ivar B. Jessen
       Date........: 01 Oct 1999
       Notes.......: Returns the date of the first day of 
                     the first European week.
                     ** Note that this method only really serves one
                     purpose, and that is it's use by the EWeekNo()
                     method above ...
       Written for.: Visual dBase 7.01
       Rev. History: 01 Oct 1999
       Calls.......: dateex.FDoY()
       Usage ......: d.EDoEW( dateY )
       Example.....: ? d.EDoEW( {31-12-1997} )
       Returns.....: First day of First European Week in a year ...
       Parameters..: dateY

       More notes:
       If first day of year is Wednesday (dow()=4) then first day of
       week no.1 is 2 days (-4+2=-2) earlier than first day of year 
       (i.e. 30-Dec-previous year).
       If first day of week is Friday (dow()=6) week no.1 starts 3 days
       (-6+2+7=3) later than first day of year (i.e. 3-Jan-same year).
      -------------------------------------------------------------
   */
      local dFDoY, nDays 

      dFDoY= class::FDoY( dateY )  // First day of year

      nDays = -dow( dFDoY ) + 2  
      If nDays <= -4
         nDays += 7   
      endif                  

   RETURN dFDoY + nDays    
   // -- EoM FDoEW()

   FUNCTION BeginEweekDate( dDate )
   /*
      -------------------------------------------------------------
       Programmer..: Ivar B. Jessen
       Date........: 21 Nov 2003
       Notes.......: Returns the date of the first day of the European week
       					containing the input date 
       Written for.: dBASE PLUS version 2.01
       Rev. History: 
       Calls.......: DateEx.FDoEW()
       Usage ......: Date.BeginEweekDate( dDate )
       Example.....: ? _app.Date.BeginEweekDate({03-01-99})  //  28-12-1998
                     ? _app.Date.BeginEweekDate({04-01-1999})//  04-01-1999
                     ? _app.Date.BeginEweekDate({25/11/99})  //  22-11-1999
                     ? _app.Date.BeginEweekDate({01.01.1999})//  28-12-1998

                     Notice: There is no function for the date of the end 
                     of an European week. It may however be calculated from
                     the begin of the week plus six days, for example
                     ? _app.Date.BeginEweekDate( {03.01.2003} )
                     + 6 // 05-01-2003

       Returns.....: Date of first day of European week as date
       Parameters..: dDate.  dDate is type date
      -------------------------------------------------------------
   */
      local dFDoY      // Date of first day of input year
      local dFDoNexY   // Date of first day of next year 
      local dFDoPrvY   // Date of first day of previous year 
      local dRtv       // Return value
      local dFDoFW     // Date of first day of first week in input year
      local dLDoLW     // Date of last  day of last  week in input year
      local dFDoFWPrvY // Date of first day of first week of previous year

      dFDoY      = class::FDoY( dDate )
      dFDoNexY   = class::FDoY( dFDoY + 366 )
      dFDoPrvY   = class::FDoY( dFDoY - 1 )
      dFDoFW     = class::FDoEW( dFDoY )
      dLDoLW     = class::FDoEW( dFDoNexY ) - 1 
      dFDoFWPrvY = class::FDoEW( ( dFDoY) - 1 )

      do case
   case dDate < dFDoFW   
   	dRtv = dFDoFWPrvY + ;
      		 ( int( ( dFDoFW - 1 - dFDoFWPrvY ) / 7 ) + 1 ) * 7 - 7
   case dDate > dLDoLW 
   	dRtv = dLDoLW + 1 
   otherwise
   	dRtv = dFDoFW + ( int( ( dDate-dFDoFW ) / 7 ) + 1 ) * 7 - 7
      endcase

   RETURN dRtv
   // -- EoM: BeginEweekDate()

   FUNCTION BeginEweekNoYear( nEweekNo, nYear )
   /*
      -------------------------------------------------------------
       Programmer..: Ivar B. Jessen
       Date........: 21 Nov 2003
       Notes.......: Returns the date of the first day of the week having
       					the input European week number and belonging to the
                     input year.
       Written for.: dBASE PLUS version 2.01
       Rev. History: 
       Calls.......: DateEx.FDoEW()
       					DateEx.EweeksInYear()
       Usage ......: Date.BeginEweekNoYear( nEweekNo, nYear )
       Example.....: ? _app.Date.BeginEweekNoYear( 46, 2003 ) // 10-11-2003
                     ? _app.Date.BeginEweekNoYear((1, 2004)   // 29-12-2003
       Returns.....: Date of the first day of the European week as date 
       					or null for	bad year input
       Parameters..: nEweekNo	EweekNo is type numeric
       					nYear.  nYear is type numeric
      -------------------------------------------------------------
   */
		local nEweeks, d

		nEweeks = CLASS::EweeksInYear( nYear )	

      if nEweekNo <= 0 or nEweekNo > nEweeks
		   return null
		else
     		d = new date( nYear, 0, 1 )
			return dttod( CLASS::FDoEW( d ) + ( 7 * nEweekNo ) - 7 )
      endif
   // -- EoM: BeginEweekNoYear()


   FUNCTION EweeksInYear( nYear )
   /*
      -------------------------------------------------------------
       Programmer..: Ivar B. Jessen
       Date........: 21 Nov 2003
       Notes.......: Returns the number of European weeks in the input year
       Written for.: dBASE PLUS version 2.01
       Rev. History: 
       Calls.......: DateEx.FDoEW()
       Usage ......: Date.EweeksInYear( nYear )
       Example.....: ? _app.Date.EweeksInYear( 2003 )  // 52
                     ? _app.Date.EweeksInYear( 2004 )  // 53
       Returns.....: Number of European weeks in a year as integer
       Parameters..: nYear.  nYear is type numeric
      -------------------------------------------------------------
   */
		local FDthisYear, FDnextYear

		FDthisYear = CLASS::FDoEW( new date( nYear, 0, 1 ) )
		FDnextYear = CLASS::FDoEW( new date( nYear + 1, 0, 1 ) )

   return int( ( FDnextYear - FDthisYear ) / 7 )
	// -- EoM: EweeksInYear()


   FUNCTION YearOfEweek( dDate )
   /*
      -------------------------------------------------------------
       Programmer..: Ivar B. Jessen
       Date........: 21 Nov 2003
       Notes.......: Returns the year to which an European week belongs. 
       Written for.: dBASE PLUS version 2.01
       Rev. History: 
       Calls.......: DateEx.FDoEW() 
       Usage ......: Date.YearOfEweekDate( dDate )
       Example.....: ? _app.Date.YearOfEweek( {28-12-2003} ) //  2003
                     ? _app.Date.YearOfEweek( {30-12-2003} ) //  2004
       Returns.....: Year of week as integer
       Parameters..: dDate.  dDate is type date
      -------------------------------------------------------------
   */
      local dFDoY      // Date of first day of input year
      local dFDoNexY   // Date of first day of next year 
      local dFDoPrvY   // Date of first day of previous year 
      local nRtv       // Return value
      local dFDoFW     // Date of first day of first week in input year
      local dLDoLW     // Date of last  day of last  week in input year
      local dFDoFWPrvY // Date of first day of first week of previous year

      dFDoY      = class::FDoY( dDate )
      dFDoNexY   = class::FDoY( dFDoY + 366 )
      dFDoPrvY   = class::FDoY( dFDoY - 1 )
      dFDoFW     = class::FDoEW( dFDoY )
      dLDoLW     = class::FDoEW( dFDoNexY ) - 1 
      dFDoFWPrvY = class::FDoEW( ( dFDoY) - 1 )

      do case
         case dDate < dFDoFW   
            nRetv = int(dDate.Year - 1)
         case dDate > dLDoLW 
				nRtv = int(dDate.Year + 1)
         otherwise
            nRtv = int(dDate.Year)
      endcase

   RETURN nRtv
   // -- EoM: YearOfEweek()

   function IFirstWeekOfYear
   /*
      -------------------------------------------------------------
       Programmer..: Ivar B. Jessen, Ken Mayer
       Date........: March, 2004
       Notes.......: Returns a numeric value corresponding to the 
                     Locale of the user, based on the language
                     property of this class, showing the
                     what is considered to be the first week of 
                     the year. Particularly useful to determine
                     if a language used by the user is a European
                     language (for all those EWeek routines). (See
                     'Returns' below.)
       Written for.: dBASE PLUS version 2.01
       Rev. History: None
       Calls.......: DateEx.IFirstWeekOfYear() 
       Usage ......: Date.IFirstWeekOfYear()
       Example.....: ? Date.IFirstWeekOfYear()
                     ? Date.IFirstWeekOfYear()==2 // Logical
       Returns.....: Number:
                     0  Week containing 1/1 is the first week 
                        of that year. (US/English?)
                     1  First full week following 1/1 is the first 
                        week of that year.
                     2  First week containing at least 4 days is 
                        the first week of that year. (European)
       Parameters..: None
      -------------------------------------------------------------
   */
      local bufAdr, bufLen, localeID, nOffset, cStr, nResult

      if type('GetLocaleInfoA') # 'FP'
         extern CINT GetLocaleInfoA( CLONG, CINT, CSTRING, CINT) kernel32
      endif

      bufAdr = space(64)
      bufLen = len(bufAdr)

      // note, we are getting the localeID from
      // the array in the DateEx class, using the Language
      // specified by the user:
      localeID = this.asLang[ this.Language ]
    
      nOffset =  0x0000100D // LOCALE_IFIRSTWEEKOFYEAR

      cStr = GetLocaleInfoA(localeID, nOffset, bufAdr, bufLen)  
      nResult = substr(bufAdr, 1, cStr - 1) 

   return nResult
   // EoM: IFirstWeekOfYear()

   function getDifference( date1, date2 )
   /*
     -------------------------------------------------------------
      Programmer..: Roland Wingerter
      Date........: July 8, 2005
      Notes.......: Calculates difference between two dates.
                    Returns custom object with properties
                    nDays, nMonths, nYears.

      Rev. History: March 10, 2009
                    Fixed bug reported by Keith Hooker

      Written for.: dBASE PLUS
      Rev. History: None
      Calls.......: LDoM2
      Usage.......: oResult = new DateEx().getDifference(<dDate1>,<dDate2>)
      Example.....: oResult = new DateEx().getDifference({1.1.2005},{31.12.2005})
                    ? oResult.nDays // 30
                    ? oResult.nMonths // 11
                    ? oResult.nYearss // 0
      Returns.....: custom object
      Parameters..: dDate1, dDate2 can be a primitive date like
                    {1.1.2000} or a date object.
     -------------------------------------------------------------
   */
      local dLowdate, dHidate, o

      // Assign dates to appropriate variables
      dLowdate = iif(date1 < date2, date1, date2)
      dHidate  = iif(date1 > date2, date1, date2)

      o = new object()

      o.nDays   = dHidate.date  - dLowdate.date
      o.nMonths = dHidate.month - dLowdate.month
      o.nYears  = dHidate.year  - dLowdate.year

      // Adjust negative values for nDays
      if o.nDays < 0
         // Subtract one month
         o.nMonths--

         // Recalculate days as number of days left
         // in month of earlier date ...
         o.nDays = class::LDoM2(dLowdate) - dLowdate
         // ... plus number of days in month of higher date
         o.nDays += dHidate.date
      endif

      // Adjust values for nMonths
      if o.nMonths < 0
         o.nMonths = o.nMonths+12
         // decrement year
         o.nYears--
      endif

      // Note we don't have to adjust values for years
      // because we ensure that dLowdate <= dHidate
   return o
   // EoM: getDifference()

	function getDiffInMonths( date1, date2 )
      /*
      -------------------------------------------------------------
       Programmer..: Roland Wingerter
       Date........: July 8, 2005
       Notes.......: Returns the number of months between the two dates. 
       Written for.: dBASE PLUS
       Rev. History: None                
       Calls.......: getDifference()
       Usage.......: ? new DateEx().getDiffInMonths(<dDate1>,<dDate2>)
       Example.....: ? new DateEx().getDiffInMonths({1.1.2005},{31.12.2005}) // 11
       Returns.....: integer
       Parameters..: dDate1, dDate2 can be a primitive date like 
                     {1.1.2000} or a date object. 
      -------------------------------------------------------------
   */
   	local oDiff
      oDiff = class::getDifference( date1, date2 )
   return (oDiff.nYears*12) + oDiff.nMonths 
   // EoM: getDifferenceInMonths()

	function getDiffInYears(date1, date2)
   /*
      -------------------------------------------------------------
       Programmer..: Roland Wingerter
       Date........: July 8, 2005
       Notes.......: Difference in years between two dates. 
                     Can be used to calculate a person's age.
                     See also dateEx.Age2()
       Written for.: dBASE PLUS
       Rev. History: None                
       Calls.......: getDifference()
       Usage.......: ? new DateEx().getDiffInYears(<dDate1>,<dDate2>)
       Example.....: ? new DateEx().getDiffInYears({1.1.2000},{31.12.2005}) // 5
       Returns.....: integer
       Parameters..: dDate1, dDate2 can be a primitive date like 
                     {1.1.2000} or a date object. 
      -------------------------------------------------------------
   */
   	// Can be used to calculate a person's age
      // See also dateEx.Age2()
   	local oDiff
      oDiff = class::getDifference(date1, date2)
   return oDiff.nYears
   // EoM: getDifferenceInYears()

   function LDoM2( dDate )
   /*
      -------------------------------------------------------------
       Programmer..: Ivar B. Jessen
       Date........: July 8, 2005
       Notes.......: Returns Last day of Month.
                     Original code posted by Ivar B. Jessen in dbasetalk 
                     newsgroup. 
					      This version is more straightforward than LDoM(), 
					      which was written before there was a date object 
                     in dBASE.
       Written for.: dBASE PLUS
       Rev. History: None                
       Calls.......: None
       Usage.......: ? new DateEx().LDoM2(<dDate>)
       Example.....: ? new DateEx().LDoM2({1.1.2005}) 
                     ? new DateEx().LDoM2(new date())
       Returns.....: date
       Parameters..: dDate can be a primitive date like 
                     {1.1.2000} or a date object. 
      -------------------------------------------------------------
   */
	return  dttod(new date(dDate.year, dDate.month + 1, 1) - 1)
   // EoM: LDoM2()

   FUNCTION FDoQ( dArg )
   /*
      -------------------------------------------------------------
       Programmer..: Unknown, may be Rich Muller, contributed by 
                     him.
       Date........: August, 2006
       Notes.......: Returns First Day of the Quarter
                     in which a specific date is passed.
       Written for.: dBASE PLUS
       Rev. History: None                
       Calls.......: None
       Usage.......: ? new DateEx().FDoQ(<dArg>)
       Example.....: ? new DateEx().FDoQ({1.1.2005}) 
                     ? new DateEx().FDoQ(new date())
       Returns.....: date
       Parameters..: dArg can be a primitive date like 
                     {1.1.2000} or a date object. 
      -------------------------------------------------------------
   */
   return iif(type("Argvector(1)") == "D", ;
          dttod(new date(year(dDate),(this.quarter(dDate) * 3) - 3,1)), 0)
   // EoM: FDoQ()

   FUNCTION LDoQ( dArg )
   /*
      -------------------------------------------------------------
       Programmer..: Unknown, may be Rich Muller, contributed by 
                     him.
       Date........: August, 2006
       Notes.......: Returns Last Day of the Quarter
                     in which a specific date is passed.
       Written for.: dBASE PLUS
       Rev. History: None                
       Calls.......: None
       Usage.......: ? new DateEx().FDoQ(<dArg>)
       Example.....: ? new DateEx().FDoQ({1.1.2005}) 
                     ? new DateEx().FDoQ(new date())
       Returns.....: date
       Parameters..: dArg can be a primitive date like 
                     {1.1.2000} or a date object. 
      -------------------------------------------------------------
   */
   return iif(type("Argvector(1)") == "D", ;
          dttod(new date(year(dDate),(this.quarter(dDate) * 3), 1) - 1), 0)


ENDCLASS
// -- End of Class: DateEx

/*
   ----------------------------------------------------------------
   End of Class: DATEEX.CC
   ----------------------------------------------------------------
*/

// Form used by the ChangeLanguage method of this class:
** END HEADER -- do not remove this line
//
// Generated on 03/08/2004
//
parameter bModal
local f
f = new ChangeLanguageForm()
if (bModal)
   f.mdi = false // ensure not MDI
   f.readModal()
else
   f.open()
endif

class ChangeLanguageForm of FORM
   with (this)
      onOpen = class::FORM_ONOPEN
      height = 11.5
      left = 42.0
      top = 0.0
      width = 40.0
      text = ""
      autoCenter = true
      mdi = false
      showTaskBarButton = false
      maximize = false
      minimize = false
   endwith

   this.TEXT1 = new TEXT(this)
   with (this.TEXT1)
      height = 1.3182
      left = 0.0
      top = 0.0
      width = 40.0
      anchor = 2	// Top
      alignHorizontal = 1	// Center
      fontSize = 14.0
      text = "Change Language"
      borderStyle = 1	// Raised
   endwith

   this.TEXT2 = new TEXT(this)
   with (this.TEXT2)
      height = 2.3636
      left = 0.0
      top = 1.3182
      width = 40.0
      fontSize = 9.0
      fontItalic = true
      text = "Select the language used by the International Date methods in the DateEX class of the dUFLP."
      borderStyle = 10	// Etched Out
   endwith

   this.LANGUAGESLISTCOMBO = new COMBOBOX(this)
   with (this.LANGUAGESLISTCOMBO)
      onChange = class::LANGUAGESLISTCOMBO_ONCHANGE
      height = 1.1364
      left = 0.8571
      top = 5.0909
      width = 37.8571
      style = 2	// DropDownList
   endwith

   this.TEXTLABEL1 = new TEXTLABEL(this)
   with (this.TEXTLABEL1)
      height = 0.6818
      left = 0.8571
      top = 4.3636
      width = 7.7143
      text = "Options:"
   endwith

   this.TEXTLABEL2 = new TEXTLABEL(this)
   with (this.TEXTLABEL2)
      height = 0.7727
      left = 1.1429
      top = 6.3182
      width = 24.4286
      text = "Currently Selected Language:"
   endwith

   this.CURRENTTEXT = new TEXTLABEL(this)
   with (this.CURRENTTEXT)
      height = 0.7273
      left = 1.1429
      top = 7.2273
      width = 37.1429
      text = "Current"
   endwith

   this.CURRENTMONTH = new TEXTLABEL(this)
   with (this.CURRENTMONTH)
      height = 0.7273
      left = 3.4286
      top = 8.0
      width = 35.0
      text = "CurrentMonth"
   endwith

   this.CURRENTDAY = new TEXTLABEL(this)
   with (this.CURRENTDAY)
      height = 0.7273
      left = 3.5714
      top = 8.7727
      width = 34.7143
      text = "CurrentDay"
   endwith

   this.ACCEPTBUTTON = new PUSHBUTTON(this)
   with (this.ACCEPTBUTTON)
      onClick = class::ACCEPTBUTTON_ONCLICK
      height = 1.0909
      left = 11.2857
      top = 10
      width = 7.8571
      text = "Accept"
   endwith

   this.CANCELBUTTON = new PUSHBUTTON(this)
   with (this.CANCELBUTTON)
      onClick = class::CANCELBUTTON_ONCLICK
      height = 1.0909
      left = 19.4286
      top = 10
      width = 7.8571
      text = "Cancel"
   endwith


   function ACCEPTBUTTON_onClick
      form.Language := form.oDate.Language
      form.close()      
   return

   function CANCELBUTTON_onClick
      form.close()      
   return

   function form_onOpen
      local i, cKey
      form.oDate = new DateEx()
      // passed to the form from the ChangeLanguage method
      // of the class:
      form.oDate.language := form.Language
      // if user cancels, we return to original setting:
      form.OLanguage = form.Language
      // set values on form:
      form.CurrentText.text  := form.Language
      form.CurrentMonth.text := form.oDate.IntlCMON( date() )
      form.CurrentDay.text   := form.oDate.IntlCDOW( date() )
      // Load array used for combobox dataSource:
      form.LArray = new array()
      cKey = form.oDate.asLang.firstKey
      for i = 1 to form.oDate.asLang.count()

          form.LArray.add( cKey )
          cKey = form.oDate.asLang.nextKey( cKey )

      next
      // sort it ...
      form.LArray.sort()

      form.LanguagesListCombo.dataSource := "array form.LArray"
      form.LanguagesListCombo.value      := form.Language
      
   return

   function LANGUAGESLISTCOMBO_onChange
      form.oDate.language := this.value
      // set values on form:
      form.CurrentText.text  := this.value
      form.CurrentMonth.text := form.oDate.IntlCMON( date() )
      form.CurrentDay.text   := form.oDate.IntlCDOW( date() )
   return

endclass
