/*
Class Library: FDialog.CC 
Author:  Daniel Howard (sproket@total.net)

History: 
   1.0 December 12, 1998
       Initial Release

Purpose:
   These are a set of simple containers for handling common file 
   dialogs. We have:

   FileGetter
   FilePutter
   DirGetter
   ImageGetter
   IconGetter

Notes:
   See each class for a description of what each one does.
   Version control will be applied to the entire library.

*------------------------------------------------------------*
This software is donated to the public domain. You may modify 
this code in anyway you wish however the author reserves the 
right to whatever official version is published. If you have 
further ideas/bug reports/additions then contact the author and 
your name will be added to the contributors if your idea is 
implemented.

The author accepts no responsibility for the use or consequences 
of the use of this software.
*/


/*
Class:  FileGetter
Author: Daniel Howard (sproket@total.net)

Purpose:
   Allows a user to get a file using the GetFile dialog.
         
Notes:
   I don't bother actually validating whether the file exists or not.
   The user can type anything they want. Experiment with the 
   fileGetter's CanChange event to add that functionality.
*/
class FileGetter(parentObj) of FileDialog(parentObj) custom

   with (this.Rectangle)
      text = "Get a file"
   endwith

   this.FileMask = "*.*"

   this.Setter.OnChange := Class::Setter_OnChange
   this.Setter.BeforeGetValue := Class::Setter_BeforeGetValue

   this.EntryField.OnChange := Class::EntryField_OnChange

   Function Init
      Super::Init()
      with (this.WRENCHBUTTON)
         onClick = class::WRENCHBUTTON_ONCLICK
         speedTip = "Get a file"
      endwith
   return      

   function WRENCHBUTTON_onClick
      local cFile
      cFile = GetFile(this.Parent.FileMask)
      form.Repaint()
      if not empty(cFile)
         this.Parent.Setter.Value := cFile
      endif
   return

   Function EntryField_OnChange
      this.Parent.Setter.Value := this.Value
      this.Value := this.Parent.Setter.Value
   return

   Function Setter_OnChange
      this.Owner.EntryField.Value := this.Value
   return

   // Gets the actual case of the file.
   Function Setter_BeforeGetValue
      if file(this.Value)
         local a,cFile
         a = New Array()
         a.Dir(this.Value)
         cFile = a[1,1]
         local n
         n = rat("\",this.Value)
         cFile = proper(left(this.Value,n))+cFile
         return ( cFile )
      endif
   return ( proper(this.Value) )
endclass

/*
Class:  FilePutter
Author: Daniel Howard (sproket@total.net)

Purpose:
   Allows a user to put a file using the PutFile dialog.
         
Notes:
   I don't bother actually validating whether the file exists or not.
   The user can type anything they want. Experiment with the 
   filePutter's CanChange event to add that functionality.
*/
class FilePutter(parentObj) of FileDialog(parentObj) custom

   with (this.Rectangle)
      text = "Put a file"
   endwith

   this.FileMask = "*.*"

   this.Setter.OnChange := Class::Setter_OnChange
   this.Setter.BeforeGetValue := Class::Setter_BeforeGetValue

   this.EntryField.OnChange := Class::EntryField_OnChange

   Function Init
      Super::Init()
      with (this.WRENCHBUTTON)
         onClick = class::WRENCHBUTTON_ONCLICK
         speedTip = "Put a file"
      endwith
   return

   function WRENCHBUTTON_onClick
      local cFile
      cFile = PutFile("",this.Parent.FileMask)
      form.Repaint()
      if not empty(cFile)
         this.Parent.Setter.Value := cFile
      endif
   return

   Function EntryField_OnChange
      this.Parent.Setter.Value := this.Value
      this.Value := this.Parent.Setter.Value
   return

   Function Setter_OnChange
      this.Owner.EntryField.Value := this.Value
   return

   Function Setter_BeforeGetValue
   return ( proper(this.Value) )
endclass


/*
Class:  ImageGetter
Author: Daniel Howard (sproket@total.net)

Purpose:
   Allows a user to get an image file using the GetFile dialog.
         
Notes:
   This controls adds an image object as a previewer for the selected
   image. It's petty much the same as FileGetter except I validate the 
   entry.
*/
class ImageGetter(parentObj) of FILEDIALOG(parentObj) custom
   with (this)
      left = 2.3333
      top = 1.8
      width = 50.8333
      height = 2.5
      metric = 0        // Chars
   endwith


   with (this.RECTANGLE)
      left = 0.5
      top = 0
      width = 48.8333
      height = 2.25
      text = "Get image"
   endwith


   this.IMAGE = new DIMAGE(this)
   with (this.IMAGE)
      height = 1.6
      left = 43.17
      top = 0.45
      width = 5.3333
      pageno = 0
   endwith

   this.FileMask = "*.ico;*.bmp;*.gif;*.jpg;*.jpe;*.jpeg;*.png;*.xbm;*.wmf;*.emf;*.tiff;*.tif;*.pcx;*.eps"

   this.Setter.OnChange := Class::Setter_OnChange
   this.Setter.CanChange := Class::Setter_CanChange
   this.Setter.BeforeGetValue := Class::Setter_BeforeGetValue

   this.EntryField.OnChange := Class::EntryField_OnChange

   Function Init
      Super::Init()
      with (this.WRENCHBUTTON)
         onClick = class::WRENCHBUTTON_ONCLICK
         speedTip = "Get an Image"
      endwith
   return

   Function EntryField_OnChange
      this.Parent.Setter.Value := this.Value
      this.Value := this.Parent.Setter.Value
   return

   function WRENCHBUTTON_onClick
      local cFile
      cFile = GetFile(this.Parent.FileMask)

      if not empty(cFile)
         this.Parent.Setter.Value := cFile
      endif
   return

   Function Setter_OnChange
      this.Owner.EntryField.Value := this.Value

      if not empty(this.Value)
         local cImage
         cImage = this.Value
         if at(" ",cImage) > 0
            cImage = '"' + cImage + '"'
         endif
         this.Owner.Image.DataSource = "FILENAME " + cImage
      else
         this.Owner.Image.DataSource = ""
      endif
   return

   Function Setter_CanChange
   para xValue

      local cImage
      cImage = xValue
      if at(" ",cImage) > 0
         cImage = '"' + cImage + '"'
      endif

      local f,i,lValid
      lValid = true
      f = New form()
      i = New Image(f)
      try
         i.DataSource = "FILENAME "+cImage
      catch( exception e )
         msgbox("Invalid Image file.","Image Getter")
         lValid = false
      endtry
   return ( lValid )

   Function Setter_BeforeGetValue
      if file(this.Value)
         local a,cFile
         a = New Array()
         a.Dir(this.Value)
         cFile = a[1,1]
         local n
         n = rat("\",this.Value)
         cFile = proper(left(this.Value,n))+cFile
         return ( cFile )
      endif
	return ( proper(this.Value) )
endclass

/*
Class:  DirGetter
Author: Daniel Howard (sproket@total.net)

Purpose:
   Allows a user to get a directory using the Getdirectory() dialog.
         
Notes:
   I don't bother actually validating whether the directory exists or
   not. The user can type anything they want. Experiment with the 
   DirGetter's CanChange event to add that functionality.
*/
class DirGetter(parentObj) of FileDialog(parentObj) custom

   with (this.Rectangle)
      text = "Get a Directory"
   endwith

   this.Setter.OnChange  := Class::Setter_OnChange

   this.EntryField.OnChange := Class::EntryField_OnChange

   Function Init
      Super::Init()
      this.WrenchButton.OnClick = Class::WRENCHBUTTON_onClick
      this.WrenchButton.speedTip = "Get a file"
   return

   function WRENCHBUTTON_onClick
      local cDir
      cDir = GetDirectory()
      if not empty(cDir)
         this.Parent.Setter.Value := cDir
      endif
   return

   Function EntryField_OnChange
      this.Parent.Setter.Value := this.Value
      this.Value := this.Parent.Setter.Value
   return

   Function Setter_OnChange
      this.Owner.EntryField.Value := this.Value
   return
endclass

/*
Class:  IconGetter
Author: Daniel Howard (sproket@total.net)

Purpose:
   This ugly little control is used to get an icon file and preview it.
   It's not based on FileDialog like the other classes in this file.
   I use this in my TreeView form for customizing the dTreeView 
   control.  The only thing I like about it is that it doesn't use 
   up much screen real-estate.
            
Notes:
   This is ugly. I use a pushbutton as the image preeviewer because the
   image object doesn't have a speedtip property. The pushbutton really
   warps the image.
*/
class IconGetter(parentObj) of DCONTAINER(parentObj) custom
   with (this)
      left = 3.1667
      top = 1.3
      width = 27
      height = 1.1
      metric = 0	// Chars
   endwith


   this.ICONBUTTON = new DPUSHBUTTON(this)
   with (this.ICONBUTTON)
      onClick = class::ICONBUTTON_ONCLICK
      height = 0.8
      left = 0
      top = 0
      width = 2.6667
      text = ""
      speedBar = true
      speedtip = "Click to change image"
   endwith


   this.TEXT = new DTEXT(this)
   with (this.TEXT)
      height = 0.7
      left = 3
      top = 0.05
      width = 17.8333
      text = "Icon"
   endwith

   this.Setter.OnChange = Class::Setter_OnChange
   this.Setter.CanChange = Class::Setter_CanChange
   this.Setter.BeforeGetValue := Class::Setter_BeforeGetValue

   function ICONBUTTON_onClick
      this.Parent.Setter.Value := GetFile(".ico")
   return

   Function Setter_CanChange
   para xValue

      local cImage
      cImage = xValue
      if at(" ",cImage) > 0
         cImage = '"' + cImage + '"'
      endif
      
      local lValid 
      lValid = true

      if not empty(cImage)
         local f,p
         f = New form()
         p = New PushButton(f)
         try
            p.UpBitmap = "FILENAME "+cImage
         catch( exception e )
            msgbox("Invalid Image file.","Icon Getter")
            lValid := false
         endtry
      endif
   return ( lValid )

   Function Setter_OnChange
      local cImage

      cImage = this.Value
      if not empty(cImage)
         if at(" ",cImage) > 0
            cImage = '"' + cImage + '"'
         endif
      endif

      if not empty(cImage)
         this.Owner.IconButton.UpBitmap = "FILENAME "+cImage
      else
         this.Owner.IconButton.UpBitmap = ""
      endif
   return

   Function Setter_BeforeGetValue
   return ( proper(this.Value) )
endclass

// ------------------------------------------------------------- //
/*
      The rest of these classes are abstract classes which I 
      subclass from. Not much to say about them. They simply
      have the base settings / behaviours that I inherit from.
      Except for IconGetter.
*/
// ------------------------------------------------------------- //

class FileDialog(parentObj) of DContainer(parentObj) custom
   with (this)
      left = 2.3333
      top = 1.8
      width = 44.8333
      height = 2.5
      metric = 0	// Chars
   endwith


   this.RECTANGLE = new FD_RECTANGLE(this)
   with (this.RECTANGLE)
      left = 0.5
      top = -0.1
      width = 43
      height = 2
      text = "File Dialog"
      pageno = 0
   endwith


   this.ENTRYFIELD = new FD_ENTRYFIELD(this)
   with (this.ENTRYFIELD)
      height = 1
      left = 1
      top = 0.65
      width = 42
      value = ""
   endwith


   this.Setter.Value := space(254)
      
   // Create the WrenchButton at the right edge of the entryfield.
   Function Init
      Super::Init()
      this.WRENCHBUTTON = new FD_PUSHBUTTON(this)
      with (this.WRENCHBUTTON)
         height = this.EntryField.Height - .15
         left = this.EntryField.Width - 3.33
         top = 0.75
         width = 4
         speedTip = "do something to a file"
         group = true
         value = false
      endwith
   return
endclass

// I protect onClick since you'll not need it when using these 
// controls. Ovverride the field property's OnChange insead.
CLASS FD_PUSHBUTTON(FormObj) OF dPUSHBUTTON(FormObj) Custom
protect onClick
   this.FontBold = .F.
   this.Text = ""
   this.Height = 1.4131
   this.Width = 5
   this.Group = .T.
   this.UpBitmap = "RESOURCE #144"
   this.Left = 66
   this.Top = 1.8809
   this.SpeedBar = true
ENDCLASS



Class FD_Rectangle(parentObj) of dRectangle(parentObj)
endclass

// I protect onChange since you'll not need it when using these 
// controls. Override the Setter property's OnChange insead.
Class FD_EntryField(parentObj) of dEntryField(parentObj)
protect onChange
endclass

