// Metric: 6 - Pixels only
/* 
   ----------------------------------------------------------------
   FormCntl.CC
   Authors.: Ken Mayer
             Gary White
             Many others over the years, see
             individual classes.
   Original Date December 19, 1999

   NOTE: This version is for dBASE Plus 8 or later,
	which is why some code has been cleaned out (there
	was code going back to 2.x).

   ***************************************************************
   ***************************************************************

   This file contains custom control definitions for each
   control I use in an application. The idea is to derive all 
   controls from these, so that a simple change here will 
   update everything in the system. (see the Knowledgebase
   either in the product or at the dBASE, Inc. website
   and read the HOW TO document on form controls -- this
   discusses most of these controls, and the document on
   using the grid covers the grid in some detail)

   There is at least one of each of the following controls
   contained in this file:
      Text
      Label  // only available in Visual dBASE 7.5 or later!!
      EntryField
      SpinBox
      Editor
      ListBox
      Combobox
      Checkbox
      Radiobutton
      Rectangle
      Line (subclassed from Rectangle, actually)
      Container
      Notebook
      Tabbox
      Image
      Grid
      Slider
      Progress 
      TreeView

   For pushbuttons, see CUSTBUTT.CC and MOREBTNS.CC

   There are no ACTIVEX, SHAPE, BROWSE (outdated -- see Grid),
         SCROLLBAR, OLE, or REPORTVIEWER (see PREVIEW.WFM) 
         controls in this file.

   You should check the headers for each control -- in some
   cases, there are very explicit instructions (the custom 
   date entryfields/spinboxes, and custom grids in particular).

   IMPORTANT NOTE: the kmCustDateEntryField and 
   kmCustDateEntryField2 (and kmCustDateSpinbox/
   kmCustDateSpinbox2) classes use HOLIDAY.CC, which in 
   turn uses HEBREW.CC and DATEEX.CC, as well as the setproc.prg 
   file, and MAKEHOLS.DBF. If any of these are not deployed 
   and you use this entryfield (which is rather handy, if 
   I may say so myself <G>) then it won't work ... see 
   HOLIDAYS.TXT for details on what needs to be deployed -- 
   there's quite a lot.

   Special thanks to Gary White for the work on the date entryfield
   kmCustGrid and holiday classes!

   The controls here are:
   ---------------------------------------------------------------
   TEXT CONTROLS:
   kmFormTitle            -- text with a border, larger font ...
   kmVerticalFormTitle1   -- vertical title subclassed from above
                             with text rotated 90 degrees ...
   kmVerticalFormTitle2   -- same, but text rotated 270 degrees
   kmCustWrapText         -- a "real" text control that wraps
                             and has an etched out border.

   TEXTLABEL CONTROLS:
   kmCustLabel            -- Used for the left of UI controls
                             this is a label that is right aligned.
   kmOtherCustLabel       -- A textLabel control that is left aligned,

   DATA CONTROLS (ones that can be used to link to data):
   kmCustEntryField       -- entryfield with various properties set
                             February 12, 1999 -- This version should
                             handle the international issues
                             without any effort ...
   kmCustDateEntryField   -- Special date entryfield
                             used for "business" purposes
                             to move a date off a holiday or
                             weekend!
   kmCustDateEntryField2  -- Similar to above, allows USER to
                             specify which way to go with date,
                             or to leave it "as is".
   kmDisableEntryField    -- as it says -- a disabled entryfield
   kmCustSpinbox          -- as it says
   kmCustDateSpinbox      -- Uses the onLostFocus event of
                             kmCustDateEntryfield (see notes
                             for that control)
   kmCustDateSpinbox2     -- Uses the onLostFocus event of
                             kmCustDateEntryfield2 (see notes
                             for that control)
   kmCustEditor           -- editor -- includes code to deal
                             with the fact that the user
                             can actually edit a datalinked
                             editor's field with the rowset's
                             autoEdit property set to false.
   kmCustCombobox         -- style 2 combobox
                             February 2, 1999 -- updated
                             to ensure that if the rowset's
                             state is "1" (Browse) that the
                             user cannot change the value in the
                             combobox ...
   kmCustComboboxStyle1   -- style 1 subclass of above
   kmCustComboboxStyle0   -- style 0 subclass of kmCustCombox
   kmCustListbox          -- listbox
   kmCustCheckBox         -- checkbox
   kmCustRadio            -- radiobutton
   kmCustImage            -- some basic settings
   kmCustGrid             -- a bunch of the options get set
                             AND you can set "global" fonts/colors
                             for your grid ... 
   kmCustEditGrid         -- grid that allows editing
   
   GADGETS:
   kmCustRectangle        -- etched-out, no text
   kmCustRectangle2       -- Uses Windows Theme/Manifest (rounded
                             corners, etc.)
   kmCustLine             -- basic line control ... subclassed
                             from the rectangle so we can have
                             a three-d line ... make sure that
                             the height stays at .1 or the
                             appearance will change ...
   kmCustContainer        -- transparent, same border as rectangle
   kmCustContainer2       -- uses kmCustRectangle2 as defined above,
                             but with no text (rounded corners, etc.)
                             for the border of the container object.
   kmCustNotebook         -- notebook ...
   kmMultiRowsetNotebook  -- notebook that doesn't allow changing
                             pages if in edit/append modes.
   kmCustTabBox           -- custom tabbox control
   kmCustSlider           -- "super class" for next two subclasses
   kmCustHorizontalSlider -- Horizontal, tics on bottom
   kmCustVerticalSlider   -- vertical, tics on both sides
   kmCustProgress         -- really basic
   kmTreeView             -- basically this is Keith Chuvala's
                             treeView from ICon 98
   kmTreeItem             -- used with above (not a custom control)
   kmTreeContainer        -- has a kmTreeView, with two pushbuttons
                             to expand/collapse the treeview

   MISC:
   Repaint()              -- Jim Sare's function with API calls
                             to repaint a form
   VerifyDateForm         -- Form used by kmCustDateEntryField2
                             and kmCustDateSpinbox2 above. 

   ---------------------------------------------------------------
   NOTE: You can change font, font sizes, colors and borderStyles
   in one place for most of these controls -- that
   is what all the #defines are below. 

   If you leave this in the dUFLP folder/Source Alias, the one
   problem with modifying the colors as noted in the paragraph
   above, is that with a new version of the dUFLP, your modifications
   will be overwritten.

   Instead, you may want to create your own code library, and subclass
   individual controls from here:

   Create a new file and add code along these lines:

   MyFormControls.cc
   // don't leave off the 'custom' keyword:
   class MyGrid( oParent ) of kmCustGrid( oParent ) from ;
                                           :dUFLP:FormCntl.cc custom
      with( this )
         bgColor = "Blue"
         // etc.
      endwith
   endclass
   // repeat for each class you wish to sub-class ...
   ----------------------------------------------------------------
*/

// ----------------------------------------------------
// Font info:
#DEFINE CONTROLFONTNAME "WindowText"    // default
#DEFINE S_CONTROLFONTSIZE 10            // For use with 'small fonts'
#DEFINE L_CONTROLFONTSIZE  8            // For use with 'large fonts'
#DEFINE S_LABELFONTSIZE   10            // For use with 'small fonts'
#DEFINE L_LABELFONTSIZE    8            // For use with 'large fonts'


// ----------------------------------------------------
// Colors:
/*
    Color Notes:
       To use Windows Schemes, so that your application matches
       the scheme your user might be using, you need to use
       information from the online help. This can be a bit
       deceptive, but ...(also see note below about Manifest
       files for Windows XP)

       By default forms use:             BtnFace (color of buttons)
       The standard Text color is:       BtnText
       Standard Highlight color is:      Highlight
       Standard Highlight text color is: HighlightText
       Data Entry is normally:           WindowText/Window
           (Color of text in windows, background is standard
            window color ...)
       The following is from the online help:

                             Corresponding Display Properties 
       Color                 Appearance Color
       ========================================================
       ActiveBorder          Active window border
       ActiveCaption         Active title bar
       AppWorkspace          Application background
       Background            Desktop
       BtnFace               3D objects
       BtnHighlight          A shade lighter than 3D objects
       BtnShadow             A shade darker than 3D objects
       BtnText               3D objects font
       CaptionText           Active title bar font
       GrayText              preset gray -- not available
       Highlight             Selected items
       HighlightText         Selected items font
       InactiveBorder        Inactive window border
       InactiveCaption       Inactive title bar
       InactiveCaptionText   Inactive title bar font
       InfoText              ToolTip
       InfoBk                ToolTip font
       Menu                  Menu
       MenuText              Menu font
       Scrollbar             A shade lighter than 3D objects
       Window                Window
       WindowFrame           preset drop shadow -- not available
       WindowText            Window font

       Some basic testing of the settings given below with 
       the code onDesignOpen for many of the controls, 
       appears to set the colors properly for when a user 
       wants to use Windows Schemes other than basic 
       default ones, except as noted (highlight is different, 
       used for things like colorHighlight for controls, 
       rowSelectColor for the grid ...). The onDesignOpen 
       code used for many of the controls will only change the 
       colors *once* -- when the control is placed onto
       the form. However, if the colors used are the standard
       Windows colors, if the Windows scheme is changed, your
       forms will automatically update with the scheme.

       *** NOTE: If you want to use a "Manifest" file, which
           will attempt to force your application to use
           the Windows XP Theme/Visual Style the user has set,
           see the program in the dUFLP: Create_dBASE_Manifest.prg.
*/

#DEFINE HIGHLIGHT       "White/Blue"
#DEFINE NORMAL          "WindowText/Window"
#DEFINE NORMALCHECK     "BtnText/BtnFace" // Checkbox and RadioButton Normal
#DEFINE NORMALNOBACK    "BtnText/BtnFace"
#DEFINE DISABLED        "White/Maroon"
#DEFINE TITLECOLOR      "BLUE"
// Grid specific colors:
#DEFINE GRIDBACK        "WHITE"        // bgColor property
#DEFINE GRIDHEADER      "BtnText/BTNFACE"
// TabBox colors:
#DEFINE TABBOXBACK      "BtnFace"      // background
#DEFINE TABBOXNORMAL    "BtnText/BtnFace" // current tab (used for Notebook also)
// TreeView colors:
#DEFINE TREEVIEWCOLORS  "WINDOWTEXT/WINDOW" // text/background
// Slider colors:
#DEFINE SLIDERCOLORS    "BtnText/BtnFace"

// ----------------------------------------------------
// Standard border for entry objects:
#DEFINE ENTRYBORDERSTYLE  7 // used in kmCustEntryField, kmCustEditor,
                            // kmCustImage, kmCustGrid, kmCustCombobox
                            // kmCustTreeView
#DEFINE CONTAINERBORDER  10 // kmCustRectangle, kmCustContainer
#DEFINE CHECKBORDER       3 // kmCustCheckBox, kmCustRadio
#DEFINE GADGETBORDER      0 // kmCustSlider, kmCustProgress
      /* border options (so you don't have to go look 'em up):
             0 = default
             1 = raised
             2 = lowered
             3 = none
             4 = single (single-line)
             5 = double (double-line)
             6 = drop shadow
             7 = client
             8 = modal
             9 = etched in
             10 = etched out
       */
// ----------------------------------------------------
// end of definitions of constants ...
// ----------------------------------------------------

/*
   ----------------------------------------------------------------
   kmSeeker -- this is a subclass of the original Seeker in the
   :FormControls: source code alias for dBASE Plus.

   The ONLY reason for putting this here is to ensure that
   the seeker control uses the same color scheme as other
   controls on the form (at least the ones from here).

   NOTE: If you use "kmSeeker" you will need to ensure that
   you include in any deployment the Seeker.cc file:
   :FormControls:seeker.cc ... this means in the Project
   Manager, etc.
   ----------------------------------------------------------------
*/
class kmSeeker( oParent ) of seeker( oParent ) from :FormControls:seeker.cc custom
   with( this )
        borderStyle    := ENTRYBORDERSTYLE
        colorNormal    := NORMAL
        colorHighlight := HIGHLIGHT
        fontName       := CONTROLFONTNAME
        if smallFonts()
           fontSize    := S_CONTROLFONTSIZE
        else
           fontSize    := L_CONTROLFONTSIZE
        endif
   endwith
endclass

/*
   ----------------------------------------------------------------
   kmNumSeeker -- this is a subclass of the original Seeker in the
   :FormControls: source code alias for dBASE Plus.

   This subclass of the seeker control works for numeric indexes,
   and also will get the color and font scheme as with the 
   kmSeeker shown above.

   NOTE: If you use "kmNumSeeker" you will need to ensure that
   you include in any deployment the Seeker.cc file:
   :FormControls:seeker.cc ... this means in the Project
   Manager, etc.
   ----------------------------------------------------------------
*/
class kmNumSeeker( fArg ) of Seeker( fArg ) from :FormControls:seeker.cc custom
   with( this )
        borderStyle    := ENTRYBORDERSTYLE
        colorNormal    := NORMAL
        colorHighlight := HIGHLIGHT
        fontName       := CONTROLFONTNAME
        if smallFonts()
           fontSize    := S_CONTROLFONTSIZE
        else
           fontSize    := L_CONTROLFONTSIZE
        endif
   endwith
   function NormalizedValue
   return val( this.Value )
endclass

/* 
   ----------------------------------------------------------------
    kmFormTitle -- Most of my forms use a title at the top.
    It's rather handy to subclass it this way ...  
   ----------------------------------------------------------------
*/
class kmFormTitle( oParent ) of Text( oParent ) custom
   with (this)
        alignHorizontal := 1	// Center
        alignVertical   := 1	// Middle
        anchor          := 2  // Top
        border          := true
        borderStyle     := 1  // Raised
        colorNormal     := TITLECOLOR
        fontName        := CONTROLFONTNAME
        if smallFonts()
           fontSize    := 18
        else
           fontSize    := 15
        endif
        metric          := 6   // Pixels
        prefixEnable    := false
        text            := "kmFormTitle"
        transparent     := true
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
    kmVerticalFormTitle1 -- Take the formTitle control above, and 
    rotate it 90 degrees, and set the height/width ... and:
   ----------------------------------------------------------------
*/
class kmVerticalFormTitle1(parentObj) of kmFormTitle(parentObj) custom
   with (this)
      anchor   := 4 // right

      if SmallFonts()
         fontSize := 22
      else
         fontSize := 18
      endif
      rotate   := 1	// 90 degrees
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
    kmVerticalFormTitle2 -- Take the VerticalFormTitle1 control 
    above, and rotate it 270 degrees: 
   ----------------------------------------------------------------
*/
class kmVerticalFormTitle2(parentObj) of kmVerticalFormTitle1(parentObj) custom
   with (this)
      anchor := 3 // left
      rotate := 3	// 270 degrees
   endwith
endclass

/*
   ----------------------------------------------------------------
    kmCustWrapText -- Ok, we sometimes need a text control 
    that is really a text control ... I use this kind of 
    control a lot for brief explanations on forms ...
   ----------------------------------------------------------------
*/
class kmCustWrapText( oParent ) of text( oParent ) custom
   with (this)
        border          := true
        borderStyle     := CONTAINERBORDER // etched out
        colorNormal     := NORMALNOBACK
        fontItalic      := true
        fontName        := CONTROLFONTNAME
        if smallFonts()
           fontSize     := S_LABELFONTSIZE
        else
           fontSize     := L_LABELFONTSIZE
        endif
        metric          := 6   // Pixels
        prefixEnable    := false
        text            := "kmCustWrapText"
        transparent     := true
    endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
    kmCustText -- standard text control
   ----------------------------------------------------------------
*/
class kmCustText( oParent ) of TEXT( oParent ) custom
   with( this )
        fontName        := CONTROLFONTNAME
        if smallFonts()
           fontSize    := S_LABELFONTSIZE
        else
           fontSize    := L_LABELFONTSIZE
        endif
        metric          := 6 //Pixels
        text            := "kmCustText"
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return
	
endclass

/*
   ----------------------------------------------------------------
    kmCustLabel -- this is only available if you are using
    Visual dBASE 7.5 or later. Note, you may want to consider replacing
    your kmCustText controls with these, but you will need
    to replace the 'value' property with 'text' on your forms.

    The label control in 7.5 is a text control without
    HTML, wrap, and justify options. This means (particularly
    due to the lack of HTML) that they use a lot less GDI
    than the TEXT control ...
   ----------------------------------------------------------------
*/
class kmCustLabel( oParent ) of TEXTLABEL( oParent ) custom
   with( this )
        alignHorizontal := 2 // Right
        alignVertical   := 1 // Middle
        fontName        := CONTROLFONTNAME
        if smallFonts()
           fontSize    := S_LABELFONTSIZE
        else
           fontSize    := L_LABELFONTSIZE
        endif
        metric          := 6 // Pixels
        text            := "kmCustLabel:"
        width           := 15
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
    kmCustOtherLabel - Other "label" controls ... this has the 
    text on the left (which is the only difference), so we 
    subclass it from above:
   ----------------------------------------------------------------
*/
class kmOtherCustLabel( oParent ) of kmCustLabel( oParent ) custom
   with (this)
        alignHorizontal := 0 // Left
        text            := "kmOtherCustLabel"
   endwith
endclass

/* 
   ----------------------------------------------------------------
   kmCustEntryField -- Standard entryfield. The onGotFocus event 
   can be over-ridden easily enough ... it's just useful to 
   be sure that the cursor isn't somewhere you don't want it. 
   I generally prefer selectAll to be off so that I (or a user) 
   don't over-write the contents of the entryfield by accident.

   NOTE: if you override the onGotFocus event, and want the
   popup to work properly, make sure you add:
      kmCustEntryfield::onGotFocus()
   in the overridden event ...
   ----------------------------------------------------------------
*/
class kmCustEntryField( oParent ) of Entryfield( oParent ) custom
   with (this)
        borderStyle    := ENTRYBORDERSTYLE
        colorNormal    := NORMAL
        colorHighlight := HIGHLIGHT
        fontName       := CONTROLFONTNAME
        if smallFonts()
           fontSize    := S_CONTROLFONTSIZE
        else
           fontSize    := L_CONTROLFONTSIZE
        endif
        metric         := 6  // Pixels
        selectAll      := false
        value          := "kmCustEntryField"
   endwith

   // -- Code added to allow the use of popups ...
   // instantiate the popup object:
   if type( "form.PopNum" ) # "N"
      form.PopNum = 1
   else
      form.PopNum++
   endif
   form.pop1 = new DefaultPopup( oParent, "Popup"+form.PopNum )
   // necessary to set the column and row positions for
   // the mouse:
   this.onRightMouseDown = class::SavePos
   // using the onRightMouseUp event to call the
   // popup itself:
   this.onRightMouseUp = {; do ShowPopup with "", this.col, this.row, this}

   function SavePos( flags, col, row )
      // this is necessary for the positioning to work
      // properly in the onRightMouseUp code:
      this.col = col
      this.row = row
   return

   function onGotFocus
       this.keyboard( "{Home}" )
       // necessary for the popup to work ...
       form.obj = this
   return

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
   kmCustDateEntryField
   This specialized entryfield uses the holidays.cc class 
   to create a holiday table, if necessary, and if not 
   necessary, check for it, and use it. It checks a date 
   to make sure it is not a holiday -- if it is, it should
   move the date OFF the holiday, either forward or
   backward (default is forward) to the next business
   day.

   The custom class HOLIDAY.CC is necessary for this
   to work properly, as is MAKEHOLS.DBF. (HOLIDAY.CC
   uses dUFLP.H, HEBREW.CC, and DATEEX.CC, which 
   also uses TIME.CC) -- This is important when you
   deploy an application -- make sure all the parts
   are there ... (see HOLIDAYS.TXT for details)

   Custom property: transfer == either "F" or any other value.
   If this value is set to "F" (default) it will move a date
   *forward* off the date, if set to anything else, it will
   move it backward (i.e., if the date is a Saturday, it 
   will move it backward to a Friday). This can be set
   in the form's onOpen, or in the control's onOpen, or if you
   wish a permament change in behavior, change it below.
   ----------------------------------------------------------------
*/
class kmCustDateEntryField( oParent ) of kmCustEntryField( oParent ) custom
   with (this)
      value = {}   // default to a date
   endwith

   this.transfer = "F" // default

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

   function onLostFocus

      // if the date field is empty, don't do anything!
      if empty( this.value)
         return
      endif

      // Make sure this is open and available:
      set procedure to :dUFLP:SetProc.prg additive
      if not setproc( ":dUFLP:HOLIDAY.CC" )
         set procedure to :dUFLP:holiday.cc additive
         this.setHol = true
      else
         this.setHol = false
      endif
      this.holiday = new Holiday()

      // here's the work -- run routine to 
      // check the date and move it if needed:
      this.value = this.Holiday.isHoliday( this.value, ;
                                           this.Transfer )

   return

endclass

/*
   ----------------------------------------------------------------
   kmCustDateEntryField2
   This is an even MORE specialized date entryfield,
   it works similar to kmCustDateEntryField, but
   instead, it uses isHol to check the date, and if
   the date falls on a holiday, it brings up a small
   form allowing the user to select whether to leave 
   the date alone, or move it forward/backward to the
   next working date available.

   The custom class HOLIDAY.CC is necessary for this
   to work properly, as is MAKEHOLS.DBF. (HOLIDAY.CC
   uses dUFLP.H, HEBREW.CC, and DATEEX.CC, which 
   also uses TIME.CC) -- This is important when you
   deploy an application -- make sure all the parts
   are there ... (see HOLIDAYS.TXT for details)
   ----------------------------------------------------------------
*/
class kmCustDateEntryField2( oParent ) of kmCustEntryField( oParent ) custom
   with (this)
      value = {}   // default to a date
   endwith

   this.transfer = "F" // default

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 0 and form.metric # 6
         msgbox( "Please use either form metric of 0 (Chars) or "+;
                 "6 (Pixels) for this control." )
         return
      endif
      if form.Metric == 6 // Pixels
         this.height := Chars2Pixels(this.height,"Y" )
         this.width  := Chars2Pixels(this.width, "X" )
      endif
   return

   function onLostFocus

      // if the date field is empty, don't do anything!
      if empty( this.value)
         return
      endif

      // Make sure this is open and available:
      set procedure to :dUFLP:SetProc.prg additive
      if not setproc( ":dUFLP:HOLIDAY.CC" )
         set procedure to :dUFLP:holiday.cc additive
         this.setHol = true
      else
         this.setHol = false
      endif
      this.holiday = new Holiday()

      // Check to see if we're on a non-working day:
      if this.Holiday.isHol( this.value )
         /* 
            We know it's a non-working date. Now
            we'll bring up a form, and ask the user
            what to do. If they want to move off the
            Non-Working-Day, we check the button clicked
            and loop through until we find a work day.
         */

         // Gary White added this (and the form at the end of
         // this file), Lawrence Bain added (in the
         // dBASE Newsgroups, April 27, 1999) the
         // "this." part, so that multiple instances
         // of this control don't mess each other up:
         this.f=new verifydateform()
         this.f.parent=this
         this.f.readmodal()
         this.f.release()

         do while this.Holiday.isHol( this.value )
            if this.transfer $ "FB"
               // transfer is either forward or backward:
               if this.transfer = "F"
                  this.value++
               else
                  this.value--
               endif // this.transfer = "F"
            else // (this.transfer = "" )
               exit // out of loop because we're leaving date
                    // "as is".
            endif // this.transfer $ "FB"
         enddo // while this.Holiday.isHol()
      endif // this.Holiday.isHol()

   return

endclass

/*
   ----------------------------------------------------------------
   kmDisableEntryField
   Disabled entryfield ... 
   I particularly like the use of the mousePointer ...
   ----------------------------------------------------------------
*/
class kmDisableEntryField( oParent ) of kmCustEntryField( oParent ) custom
   with (this)
        colorNormal      := DISABLED
        colorHighlight   := DISABLED
        when             := {; return false }
        mousePointer     := 12 // No
        // disable the popup we added for kmCustEntryfield:
        onRightMouseDown := null
        onMouseRightUp   := null
        text             := "kmDisableEntryfield"
   endwith
endclass

/*
   ----------------------------------------------------------------
   kmCustSpinbox
   I'm not really doing a lot with this one because I don't
   use spinboxes in most of my apps ...

   For dates, you can use this spinbox or you could use
   kmCustDateSpinbox or kmCustDateSpinbox2, which use the
   onLostFocus events of the code shown above for 
   kmCustDateEntryfield and kmCustDateEntryfield2.
   ----------------------------------------------------------------
*/
class kmCustSpinbox( oParent ) of SPINBOX( oParent ) custom
   with (this)
        borderStyle    := ENTRYBORDERSTYLE
        colorNormal    := NORMAL
        colorHighlight := HIGHLIGHT
        fontName       := CONTROLFONTNAME
        if smallFonts()
           fontSize    := S_CONTROLFONTSIZE
        else
           fontSize    := L_CONTROLFONTSIZE
        endif
		  metric         := 6 // Pixels
        rangeRequired  := true
        selectAll      := false
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

   // -- Code added to allow the use of popups ...
   // set form property in the onGotFocus:
   this.onGotFocus = {;form.obj = this}
   // instantiate the popup object:
   if type( "form.PopNum" ) # "N"
      form.PopNum = 1
   else
      form.PopNum++
   endif
   form.pop1 = new DefaultPopup( oParent, "Popup"+form.PopNum )
   // necessary to set the column and row positions for
   // the mouse:
   this.onRightMouseDown = class::SavePos
   // using the onRightMouseUp event to call the
   // popup itself:
   this.onRightMouseUp = {; do ShowPopup with "", this.col, this.row, this}

   function SavePos( flags, col, row )
      // this is necessary for the positioning to work
      // properly in the onRightMouseUp code:
      this.col = col
      this.row = row
   return

endclass

/*
   ----------------------------------------------------------------
    kmCustDateSpinbox
    This is a spinbox that calls the kmCustDateEntryfield's
    onLostFocus event -- obviously it's designed to
    work with dates ...
   ----------------------------------------------------------------
*/
class kmCustDateSpinbox( oParent ) of kmCustSpinbox( oParent ) custom

   with( this )
      value = {}   // default to a date
      onLostFocus := kmCustDateEntryfield::onLostFocus
   endwith
   this.transfer = "F" // default

endclass

/*
   ----------------------------------------------------------------
   kmCustDateSpinbox2
   As above, but this one calls the kmCustDateEntryfield2's
   onLostFocus event.
   ----------------------------------------------------------------
*/
class kmCustDateSpinbox2( oParent ) of kmCustSpinbox( oParent ) custom

   with( this )
      value = {}   // default to a date
      onLostFocus := kmCustDateEntryfield2::onLostFocus
   endwith
   this.transfer = "F" // default

endclass

/*
   ----------------------------------------------------------------
   kmCustEditor
   One of the frustrating things with editor controls (before
   dBASE Plus vers. 2.02) is that there's no colorHighlight 
   property. So, let's make it act as if there were by changing
   the colorNormal ... 

   The popup can be handy, because it
   allows editing IN the control for simple formatting.
   However, for that to be useful, the evalTags property
   needs to be set to true also. This control defaults
   to them being off ... (also note that the format toolbar/
   palette does not work with modal forms (ones opened
   with readModal() )

   Note the "Key" event below by Gary White. This avoids
   a problem with the fact that the editor control
   ignores the rowset's autoEdit property and will
   allow the user to add data to the editor even if
   autoEdit is false and the rowset is not in append
   or edit modes.
   ----------------------------------------------------------------
*/
class kmCustEditor( oParent ) of Editor( oParent ) custom
   with (this)
        borderStyle    := ENTRYBORDERSTYLE
        colorNormal    := NORMAL
        colorHighlight := HIGHLIGHT
        evalTags       := false
        fontName       := CONTROLFONTNAME
        if smallFonts()
           fontSize    := S_CONTROLFONTSIZE
        else
           fontSize    := L_CONTROLFONTSIZE
        endif
        metric         := 6 // Pixels
        popupEnable    := false
        value          := "kmCustEditor"
        // needed for popup ...:
        onGotFocus     := {; form.obj = this }
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

   // only enable this if the popupEnable property is
   // false -- if it's true, we end up with two popups
   // which is wrong and bad.
   if not this.popupEnable
      // Code added to allow right-click popup
      // instantiate the popup object:
      if type( "form.PopNum" ) # "N"
         form.PopNum = 1
      else
         form.PopNum++
      endif
      form.pop1 = new DefaultPopup( oParent, "Popup"+form.PopNum )
      // necessary to set the column and row positions for
      // the mouse:
      this.onRightMouseDown = class::SavePos
      // using the onRightMouseUp event to call the
      // popup itself:
      this.onRightMouseUp = {; do ShowPopup with "", this.col, this.row, this}
   endif

   function SavePos( flags, col, row )
      // this is necessary for the positioning to work
      // properly in the onRightMouseUp code:
      this.col = col
      this.row = row
   return

   function key
      /*
          This code by Gary White is provided to
          get around a problem with rowsets that
          have the autoEdit property set to false,
          and editors. The editor seems to be immune
          to this property once you make a change
          in it -- if you then save or abandon, you
          can actually edit the contents of the editor
          object ... 
      */
      // this     = editor
      // dataLink = field
      // parent   = fieldArray
      // parent   = rowset
      if type( "this.datalink.parent.parent" ) # "U"
         r = this.datalink.parent.parent
         if r.autoEdit == false  and ;
            ( r.state # 2 and r.state # 3 )
            return 0
         endif
      endif
endclass

/*
   ----------------------------------------------------------------
   kmCustListbox
   Note: ShowSelected() was removed as not functioning/not necessary
   in dB2K. This method isn't necesary because changing the curSel
   property automatically moves the currently selected item to the
   top of the viewable area of the listbox; and not functioning
   because the API calls may not be working properly with the 32-bit
   listbox. -- Discovered by Lane Coddington
   ----------------------------------------------------------------
*/
class kmCustListbox( oParent ) of Listbox( oParent ) custom
   with (this )
      borderStyle     := ENTRYBORDERSTYLE
      colorHighLight  := HIGHLIGHT
      colorNormal     := NORMAL
      curSel          := 1
      dataSource      := "ARRAY {'kmCustListbox'}"
      fontName        := CONTROLFONTNAME
        if smallFonts()
           fontSize    := S_CONTROLFONTSIZE
        else
           fontSize    := L_CONTROLFONTSIZE
        endif
		  metric         := 6 // Pixels
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
   kmCustCombobox
   I always use style 2 comboboxes ...  same problem with 
   colorHighlight as editors -- the property doesn't exist.
   (The style 2 combobox requires the user to click on the
   button to see the dropdown, and does not allow user input)
   NOTE: the addition of the isDropped method ...
   NOTE: the addition of the onGotFocus and onChange
         events ... there is a problem in that a combobox
         acts like it is active when rowset.autoEdit 
         is false. What happens is the user looks at something
         in the combobox and leaves it -- and it appears that
         the value has been changed. This will make sure that 
         when the user leaves the combobox, if they are 
         in the rowset's "browse" state, the value that 
         was shown previously is returned (meaning it doesn't 
         look like the user changed anything ...).
   March 13, 1999 -- added Jim Sare's code for handling
         autoDrop being set to true, and the pull down
         button (when the CB doesn't have focus, clicking
         on the pulldown causes the drop-down list to 
         drop and then immediately retract the first time
         the combobox is clicked -- after that it has focus 
         and the problem doesn't exist).
   July, 2016 -- modified based on code posted by M. Dumont
         in the dBASE Plus newsgroups. A weirdness in the 
         way Windows and Style 2 comboboxes works, the control
         defaults to a gray color. Solutions: change the systemTheme
         to false, but that loses appearance issues in other ways,
         or what is done here: change the style to 1 (which
         uses the correct colors), and then force the dropdown
         to read-only. If dBASE fixes the control, we can remove
         this code as unnecessary again ... See comments
         in the onOpen event handler, and "Struct_Base" at the
         end of this file ...
   ----------------------------------------------------------------
*/
class kmCustCombobox( oParent ) of Combobox( oParent ) custom
   with (this)
        borderStyle    := ENTRYBORDERSTYLE
        colorNormal    := NORMAL
        colorHighlight := HIGHLIGHT
        fontName       := CONTROLFONTNAME
        if smallFonts()
           fontSize    := S_CONTROLFONTSIZE
        else
           fontSize    := L_CONTROLFONTSIZE
        endif
        metric         := 6 // Pixels
        //style          := 2 // DropDownList
        style          := 1 // DropDown -- for now
        value          := "kmCustCombobox"
        dataSource     := "array {'kmCustCombobox'}"
   endwith

   // define API calls for drop-down fix:
   if type( "PostMessage" ) # "FP"
      extern CLONG PostMessage( CHANDLE, CINT, CLONG, CLONG ) user32;
             from "PostMessageA"
   endif

   if type( "SendMessage" ) # "FP"
      extern CLONG SendMessage( cHANDLE, CINT, CLONG, CLONG ) user32;
             from "SendMessageA"
   endif

   #ifndef WM_USER
      #define WM_USER HToI( "400" )
   #endif
   #ifndef CB_SHOWDROPDOWN
      #define CB_SHOWDROPDOWN WM_USER + 15
   #endif
   #ifndef WM_CANCELMODE
      #define WM_CANCELMODE HToI( "1F" )
   #endif
   
   function onOpen
      // due to a weirdness in Windows 10 controls, a style 2 combobox
      // is grey. What we're doing is setting the control to style 1,
      // then using this code (and "Struct_Base" at the end of this
      // file) to force this to a read-only combobox. The appearance
      // should match standard Windows (default -- white) instead
      // of grey.
      // This code, along with the "Struct_Base" code was
      // created and posted in the newsgroups by M. Dumont.
      
      #define EM_SETREADONLY 0x00CF
      extern CLONG GetWindowLongA(CHANDLE, CINT) user32
      extern CLONG SetWindowLongA(CHANDLE, CINT, CLONG) user32
      extern CLONG CallWindowProcA(CPTR, CHANDLE, CUINT, CUINT, CUINT) user32
      extern clogical GetComboBoxInfo(chandle,cptr) user32
      extern cuint SendMessage(chandle,cuint,clong,clong) user32 from "SendMessageA"
                
      local comboinfo
      comboinfo = new Struct_base(26)
      comboinfo.setlong(0,52)
      GetComboBoxInfo(this.hwnd,comboinfo)
      this.edithandle = comboinfo.getlong(44)
      sendmessage(this.edithandle,EM_SETREADONLY,1,0) 
   return // onopen

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

   function onGotFocus
      // this is necessary for the onChange
      // event
      this.savedValue = this.value

      // if this is an autodrop combobox ...:
      if this.autoDrop
         SendMessage( this.hWnd, WM_CANCELMODE, 0, 0 )
         this.drop()
      endif

   function onChange
      // if rowset was in "Browse" mode (non-editable),
      // return the value to what it was when 
      // the combobox got focus.
      // this = combobox
      // datalink = field
      // parent = field array
      // parent = rowset
      if type( "this.datalink.parent.parent" ) # "U"
         r = this.datalink.parent.parent
         if r.autoEdit == false  and ;
            ( r.state # 2 and r.state # 3 )
            this.value = this.savedValue
            // these two statements by Lawrence Bain
            // are because he says the combobox can
            // change the rowset's state like the editor:
            r.modified = false // force to state of 1
            r.abandon()        // just in case ...
         endif
      endif
   return

   function Drop
      // cause the drop-down to drop:
      PostMessage( this.hWnd, CB_SHOWDROPDOWN, 1, 0 )
   return

   function isDropped
      /* 
         This method created by Romain Strieff
         (using the Windows API) returns true/false depending 
         on the state of the drop-down ... if down it 
         returns true.

         Usage: if form.kmCustCombobox1.isDropped()
                   // do something
                endif
      */
      #define CB_GETDROPPEDSTATE (WM_USER+23)
      if type( "SendMessage" ) # "FP"
         extern cLogical SendMessage( cHandle, cInt, cWord, cLong );
                user32 from "SendMessageA"
      endif
   return SendMessage( this.hwnd, CB_GETDROPPEDSTATE, 0, 0 )
   // EoM: isDropped

endclass

/*
   ----------------------------------------------------------------
   kmCustComboboxStyle1
   For those who want a style 1 combobox, here's a subclass
   of the above (the dropdown list requires the user to click
   the button and the entryfield accepts input):
   ----------------------------------------------------------------
*/
class kmCustComboboxStyle1( oParent ) of kmCustCombobox( oParent ) custom
   with (this)
        // remove this once Combobox style 2 works properly
        // (kmCustCombobox):
        open           := null
        style          := 1 // DropDown
        value          := "kmCustComboboxStyle1"
        dataSource     := "array {'kmCustComboboxStyle1'}"
   endwith
endclass

/*
   ----------------------------------------------------------------
   kmCustComboboxStyle0
   For those who want a style 0 combobox, here's a subclass
   of the above (the dropdown list is displayed automatically 
   when the combobox gets focus, entryfield accepts input):
   ----------------------------------------------------------------
*/
class kmCustComboboxStyle0( oParent ) of kmCustCombobox( oParent ) custom
   with (this)
        // remove this once Combobox style 2 works properly
        // (kmCustCombobox):
        open           := null
        style          := 0 // Simple
        value          := "kmCustComboboxStyle0"
        dataSource     := "array {'kmCustComboboxStyle0'}"
   endwith
endclass

/*
   ----------------------------------------------------------------
   kmCustCheckbox
   Not a lot to say about this ....
   ----------------------------------------------------------------
*/
class kmCustCheckbox( oParent ) of CheckBox( oParent ) custom
   with (this)
        borderStyle    := CHECKBORDER
        colorNormal    := NORMALCHECK
        fontName       := CONTROLFONTNAME
        if smallFonts()
           fontSize    := S_CONTROLFONTSIZE
        else
           fontSize    := L_CONTROLFONTSIZE
        endif
        metric         := 6 // Pixels
        text           := "kmCustCheckbox"
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
   kmCustRadio
   Not much to say about this, either ...
   When you place a group of radiobuttons on a form, 
   make sure that the one you want to be the default
   is set to a value of true, and the others to a value
   of false. By default, the first one placed on a form 
   will have the group property set to true. Also make sure
   no other controls get between these in the "z-order" ...
   ----------------------------------------------------------------
*/
class kmCustRadio( oParent ) of RadioButton( oParent ) custom
   with (this)
        borderStyle     := CHECKBORDER
        colorNormal     := NORMALCHECK
        fontName        := CONTROLFONTNAME
        if smallFonts()
           fontSize    := S_CONTROLFONTSIZE
        else
           fontSize    := L_CONTROLFONTSIZE
        endif
        metric          := 6 // Pixels
        text            := "kmCustRadio"
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
   kmCustPushbutton
   A simple pushbutton with the font settings to match
   everything else in the app ...
   ----------------------------------------------------------------
*/
class kmCustPushbutton( oParent ) of Pushbutton( oParent ) custom
   with (this)
        if SmallFonts()
           fontSize := S_CONTROLFONTSIZE
        else
           fontSize := L_CONTROLFONTSIZE
        endif
        fontName    := CONTROLFONTNAME
        text        := "kmCustPushbutton"
        metric      := 6 // Pixels
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
   kmCustRectangle
   A simple rectangle, but I like the etched-out appearance:
   ----------------------------------------------------------------
*/
class kmCustRectangle( oParent ) of Rectangle( oParent ) custom
   with (this)
        border          := true
        borderStyle     := CONTAINERBORDER
        if smallFonts()
           fontSize    := S_LABELFONTSIZE
        else
           fontSize    := L_LABELFONTSIZE
        endif
        fontName        := CONTROLFONTNAME
        metric          := 6 // Pixels
        text            := "" // no text -- which with
                              // this style you wouldn't get
                              // to appear anyway!
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
   kmCustRectangle2
   Windows Rectangle -- this uses the XP Theme if a manifest
   file is available and the user is working with the XP theme.
   ----------------------------------------------------------------
*/
class kmCustRectangle2( oParent ) of Rectangle( oParent ) custom
   with (this)
        if smallFonts()
           fontSize    := S_LABELFONTSIZE
        else
           fontSize    := L_LABELFONTSIZE
        endif
        fontName        := CONTROLFONTNAME
        text            := "kmCustRectangle2" 
		  metric          := 6 // Pixels
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
   kmCustLine
   A basic line control -- subclassed from the rectangle! 
   (Be nice if the line control itself could create three-d lines
   like we have rectangles and other controls ...)

   The most important part of this is the height. If you have
   a vertical line, the width would need to be set properly.
   ----------------------------------------------------------------
*/
class kmCustLine( oParent ) of Rectangle( oParent ) custom
   with (this)
      borderStyle       := 9 // etched in
		metric            := 6 // Pixels
      height            := 2 
      width             := 83.00
      text              := "" // no text
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
   kmCustContainer
   A basic container class ... this uses the "containerborder"
   definition, which defaults to the etched out look. I prefer
   my containers to be 'transparent', so that I can put something
   under them (you can put a background image on a form, and may 
   not want the container to be opaque on top of it ...), or 
   use the form's color setting ...
   ----------------------------------------------------------------
*/
class kmCustContainer( oParent ) of Container( oParent ) custom
   with (this)
      borderStyle     := CONTAINERBORDER
      transparent     := true
		metric          := 6 // Pixels
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
   kmCustContainer2
   This container has a rectangle
   object placed on it, and instead of the borderStyle defined
   elsewhere, we set the Border of the Container to none, and use
   the Windows Rectangle style (rounded corners, etc.).

   There are two reasons for wanting to do this:

      1) A border on a container cannot have text over the 
         border itself, because setting the top of the text
         to a value that would work truncates the top
         of the text (you would have to set the top to
         negative numbers and then the text starts to
         disappear).

      2) If using Windows XP, the rectangle object can
         have those spiffy rounded corners, and take on
         the default color of the XP Theme. The Container
         object's borderStyles do not allow for that.

   Hence, the extra work done here. If you want text in the
   rectangle, just click on the rectangle, and set the text
   property in the Inspector. 

   Problems found:

      * Resizing the container object does not resize the
        rectangle automatically in the designer. However,
        the code given here will ensure when the form is
        run that the rectangle will size correctly.

   This should be flexible and work on non-manifest/XP systems
   as well. 

   Special thanks to J.P. Martel for assistance in getting the
   size to work properly for both the container and the
   rectangle ...
   ----------------------------------------------------------------
*/
class kmCustContainer2( oParent ) of Container( oParent ) custom
  with (this)
     onOpen = class::OnOpen
     borderStyle     := 3 // None
     metric          := 6
  endwith

  this.MyRectangle = new kmCustRectangle2( this )
  with( this.MyRectangle )
     left   = 0
     top    = 0
     height = this.height
     width  = this.width
     pageno = 0
     text = "" // No text by default
  endwith

  function OnOpen
     // set Rectangle height/width:
     this.MyRectangle.top    := 0
     this.MyRectangle.height := this.height
     this.MyRectangle.left   := 0
     this.MyRectangle.width  := this.width
  return

  function onDesignOpen( lFromPalette )
     // set Rectangle height/width:
     this.MyRectangle.top    := 0
     this.MyRectangle.height := this.height
     this.MyRectangle.left   := 0
     this.MyRectangle.width  := this.width

     if not lFromPalette
        return
     endif
     if form.metric # 6
        msgbox( "Please use either form metric of "+;
                "6 (Pixels) for this control." )
        return
     endif
   return
endclass

/*
   ----------------------------------------------------------------
   kmCustNotebook
   Custom notebook control ...
   Nothing real fancy here. Note that you cannot change
   the color of the text on a notebook -- it is black. Period.
   Very odd ...
   ----------------------------------------------------------------
*/
class kmCustNotebook( oParent ) of Notebook( oParent ) custom
   with (this)
      colorNormal     := TABBOXNORMAL
      dataSource      := 'ARRAY {"Tab 1", "Tab 2"}'
      fontBold        := true
      fontName        := CONTROLFONTNAME
      if smallFonts()
         fontSize    := S_CONTROLFONTSIZE
      else
         fontSize    := L_CONTROLFONTSIZE
      endif
   endwith

endclass

/*
   ----------------------------------------------------------------
   kmMultiRowsetNotebook
   This is subclassed from above, with the additional use
   of the new in 7.5 canSelChange() event, which allows you
   to disable the ability to change pages. 

   In this case, we are checking the form's rowset to see
   if the state property is append or edit mode. In those
   cases, the canSelChange() event is going to be set to
   return 'false', otherwise 'true.

   The idea here is for forms where you have one rowset per
   notebook page ... when you change rowsets, you need
   to ensure that you set the form's rowset to the new one,
   i.e.,
       form.rowset = form.query2.rowset 
          or, if using a datamodule:
       form.rowset = form.datamodref1.ref.query2.rowset
   This would be done in the onSelChange() event of the notebook.

   NOTE: special property:
      warnOnPageChange
         This is logical, defaults to true
         If true, it will warn the user that they are
            editing or appending data, and ask if they
            want to save changes before moving ... they
            can save (yes), abandon (no) or cancel (which
            will return them to the notebook without changing
            pages)
      You would probably want to set this property
      in the notebook's onOpen event, along these lines:

         function kmMultiRowsetNotebook1.onOpen
            this.warnOnPageChange := false
            return
   ----------------------------------------------------------------
*/
class kmMultiRowsetNotebook( oParent ) of kmCustNotebook( oParent ) custom

   // custom property -- if it is set to true 
   // and the user attempts to change pages
   // while the current rowset is being edited/appended,
   // then a warning will come up, and they can save, abandon
   // or cancel ...
   this.warnOnPageChange = true

   function canSelChange
      if form.rowset.state == 2 or form.rowset.state == 3 
         if this.warnOnPageChange
            nWarn = msgbox( iif( form.rowset.state==2,"Editing","Appending" )+;
                         " current row, save changes?", "Change Rowset?", 35 )
            if nWarn == 6      // Yes -- save
               form.rowset.save()
            elseif nWarn == 7 // No -- abandon
               form.rowset.abandon()
            else
               return false   // don't allow page change, but don't
                              // save/abandon row ...
            endif // nWarn == 6
         else
            return false
         endif // this.warnOnPageChange
      else
         return true
      endif // form.rowset.state ...
	return
endclass

/*
   ----------------------------------------------------------------
   kmCustTabbox
    Custom Tabbox -- pretty basic, we want the tabbox to
    change the form's pageNo property to whatever the
    tabbox's current selection (curSel) is. When it opens, 
    we want to force the page number to 1 (useful since if we
    are working in the designer, and save a form,
    it saves the "current" page number out to the
    form definition ...). The onSelChange event may
    be something you will need to override ... s'okay.
    I do it all the time ...
      If you wish to call back to this onSelChange event in
      an overriden onSelChange in your form:
         kmCustTabbox::OnSelChange()   will do the trick.
   ----------------------------------------------------------------
*/
class kmCustTabbox(parentObj) of TABBOX(parentObj) custom
   with (this)
      colorNormal    := TABBOXBACK
      colorHighLight := TABBOXNORMAL
      dataSource     := 'ARRAY {"Individual Record","Find Record"}'
      fontBold       := true
      fontName       := CONTROLFONTNAME
        if smallFonts()
           fontSize    := S_CONTROLFONTSIZE
        else
           fontSize    := L_CONTROLFONTSIZE
        endif
   endwith

   function onOpen
      /*
         If you need your form to start on a different
         page, this code will override that setting,
         so you may want to add to the constructor code
         of any form that uses this control (inside the
         with/endwith block) something like:
            onOpen = {; this.curSel := 2; form.pageNo := 2 }
      */
      this.curSel := 1
      form.pageNo := 1

   function onSelChange
      form.pageNo := this.curSel
      repaint(form) // repaint this form -- API calls 
                    // by Jim Sare -- function toward end
                    // of this file
   return
endclass

/*
   ----------------------------------------------------------------
    kmCustImage -- the important aspect of this one
    is the alignment (I hate having images distorted!) ... 
   ----------------------------------------------------------------
*/
class kmCustImage(parentObj) of IMAGE(parentObj) custom
   with (this)
      alignment    := 3	// Keep Aspect Stretch
      borderStyle  := ENTRYBORDERSTYLE
      metric       := 6	// Pixels
   endwith

endclass

/*
   ----------------------------------------------------------------
   kmCustGrid
   Custom grid -- sets a bunch of the options -- I never
   allow editing in a grid, so it's useful to just deal with
   all that here. (However, there's an "editing" grid
   after this one that inherits the same functionality of this
   one but allows editing ...)

   A whole bunch of code was stripped out -- the grid,
   starting with the first release of dBASE Plus, has
   default font and color properties ... There is now
   no need for all that overhead, as it's built into
   the grid!
   ----------------------------------------------------------------
*/
class kmCustGrid(parentObj) of GRID(parentObj) custom
   with (this)
      allowAddRows      := false
      allowColumnMoving := false
      allowColumnSizing := false
      allowEditing      := false
      allowRowSizing    := false
      bgColor           := GRIDBACK
      borderStyle       := ENTRYBORDERSTYLE
      // default for memo fields:
      #if __version__ >= 2.21
         defaultMemoEditor := 0 // 0 for editor, 1 for entryfield
      #endif
      fontName          := CONTROLFONTNAME
      if smallFonts()
         fontSize    := S_CONTROLFONTSIZE
      else
         fontSize    := L_CONTROLFONTSIZE
      endif
      colorNormal       := NORMAL    
      colorHighlight    := HIGHLIGHT 
      colorRowSelect    := HIGHLIGHT
      headingColorNormal := GRIDHEADER
      headingFontName    := CONTROLFONTNAME
      if smallFonts()
         headingFontSize := S_CONTROLFONTSIZE
      else
         headingFontSize := L_CONTROLFONTSIZE
      endif
      hScrollBar        := 0	// Off
      integralHeight    := true
      rowSelect         := true
      vScrollBar        := 1 // on
		metric            := 6 // Pixels
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
   kmCustEditGrid
   Custom editing grid -- a few properties have been
   modified, but otherwise everything else is the
   same as kmCustGrid ...
   ----------------------------------------------------------------
*/
class kmCustEditGrid(parentObj) of kmCustGrid(parentObj) custom
   with (this)
      allowAddRows      := true
      allowEditing      := true
      rowSelect         := false
   endwith
endclass

/*
   ----------------------------------------------------------------
   kmCustSlider
   Sliders ... kinda fun little toys, but I don't have a lot of 
   uses for 'em. Provided here to give something basic to work 
   from. (This is the super class the other two are subclassed from)
   ----------------------------------------------------------------
*/
class kmCustSlider(parentObj) of SLIDER(parentObj) custom
   with (this)
      borderStyle     := GADGETBORDER
      colorNormal     := SLIDERCOLORS
		metric          := 6 // Pixels
   endwith

endclass

/*
   ----------------------------------------------------------------
   kmCustHorizontalSlider
   subclassed from kmSlider:
   ----------------------------------------------------------------
*/
class kmCustHorizontalSlider(parentObj) of kmCustSlider(parentObj) custom
   with (this)
      height          := 21
      width           := 206
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
   kmCustVerticalSlider
   subclassed from kmSlider:
   ----------------------------------------------------------------
*/
class kmCustVerticalSlider(parentObj) of kmCustSlider(parentObj) custom
   with (this)
	   metric          := 6 // Pixels
      height          := 155
      width           := 31
      ticsPos         := 0	// Both
      vertical        := true
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
   kmCustProgress
    Progress bar -- really useful. Be nice if we could
    change the color, but it's a Windows object that's
    been surfaced, and it uses the Windows colors ...

    NOTE: you may want to use jpProgress.cc, also
    in the dUFLP library, rather than this one ...

    When using one, drop it onto a form, and set
    the rangeMin (1 is good, but may not be what
    you want); rangeMax (if doing something with a
    table use the rowset's count() ...) and 
    set the value (start at 1, and increment as you
    do whatever you are doing ... )
   ----------------------------------------------------------------
*/
class kmCustProgress(parentObj) of PROGRESS(parentObj) custom
   with (this)
      borderStyle  := GADGETBORDER
      metric       := 6	// Pixels
      mousePointer := 12	// No (can't interact with it ...)
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

endclass

/*
   ----------------------------------------------------------------
   kmTreeView
    This treeview object is one that Keith Chuvala threw together
    and demoed at ICON 98 -- this version is only slightly 
    altered by me for inclusion here, largely for readability 
    (and to place nearly all the properties here, with some
    descriptions, as this is probably the hardest control
    to work with ...) ...
    It has four methods attached to it:
       Collapse:       Collapses all expanded treeItems
       Expand:         Expands all treeItems
       FillFromArray:  Like it says -- you must define
                       the array, first ...
          Usage: form.kmTreeView1.FillFromArray( arrayreference )
       TreeFill:       Called from FillFromArray, fills treeView
                       with treeItems

       Jan. 13, 1999 added two more methods: onGotFocus
       and onLostFocus -- these are actually event
       handlers that are being overwritten -- they
       deal with the fact that the allowEditTree 
       option when set to false still allows the user
       to delete treeItems ...

    Suggestion: put on a container with two pushbuttons
    to call the Collapse and Expand methods ... (see
    kmTreeContainer below)

    For another example, see SLIDESHOW.ZIP in the dUFLP
    library -- this one interacts heavily with a table.
   ----------------------------------------------------------------
*/
class kmTreeView(parentObj) of treeview(parentObj) CUSTOM
   with (this)
      allowEditLabels  := false // allow user to edit
      allowEditTree    := false // allow user to add/delete items
      borderStyle      := ENTRYBORDERSTYLE
      colorNormal      := TREEVIEWCOLORS
      disablePopup     := true  // whether tree's popup is disabled
      fontName         := CONTROLFONTNAME
        if smallFonts()
           fontSize    := S_CONTROLFONTSIZE
        else
           fontSize    := L_CONTROLFONTSIZE
        endif
		metric           := 6 // Pixels
      height           := 174 // to start with
      width            := 104 // to start with
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 0 and form.metric # 6
         msgbox( "Please use either form metric of 0 (Chars) or "+;
                 "6 (Pixels) for this control." )
         return
      endif
      if form.Metric == 6 // Pixels
         this.height := Chars2Pixels(this.height,"Y" )
         this.width  := Chars2Pixels(this.width, "X" )
      endif
   return

   procedure onGotFocus
      // disable delete key if necessary:
      if NOT this.allowEditTree
         on key label del * // this means to execute a comment ...
      endif

   procedure onLostFocus
      // reset the delete key
      on key label del

   procedure expand
      // Loop through tree and expand all expandable items
      local tItem
      tItem = this.firstchild
      do while tItem # NULL
         tItem.expanded = true
         tItem = tItem.nextsibling
      enddo

   procedure collapse
      // Loop through tree and collapse all collapseable items
      local tItem
      tItem = this.firstchild
      do while tItem # NULL
         tItem.expanded = false
         tItem = tItem.nextsibling
      enddo

   procedure FillFromArray
      // Loads tree from an array -- uses method treeFill to do it
      parameter aArray
      if type("aArray") # "A"
         msgbox("Array required!","Can't fill TreeView")
         return
      endif
      this.treefill(this,aArray)  // Assume it's out there!
   return

   /* 
      TreeFill: A recursive function for filling a TreeView Control
      Pass an array of strings and/or arrays of strings to define
      TreeItems, e.g. {"Item1",{"SubItem1_1","SubItem1_2"},"Item2"}
   */
   Procedure TreeFill
      Parameters tContainer, ; // Object reference to treeview or treeitem
                 aItems        // Array used to construct tree items

      local i, t, tname
      private uItem

      if type("tContainer") # "O"
         msgbox("Invalid container reference","Oops!")
         return
      endif
      if type("aItems") # "A"
         msgbox("Expecting Array reference","Oops!")
         return
      endif    

      t = tContainer

      for i = 1 to aItems.Size   
         uItem = aItems[i]   // private required to get type of aItems[i]
         if type("uItem") == "C"
            tName = "treeitem" + i
            t = new kmTreeItem(tContainer,tName)
            t.text = aItems[i]
         elseif type("uItem") == "A"         
            //t.TreeFill(t,aItems[i])
            // bug fix by Roland Wingerter,
            // replace above with :
            this.TreeFill(t,aItems[i])
         endif
      next i
   return

endclass

/*
   ----------------------------------------------------------------
   kmTreeItem
   TreeItem -- used with TreeView:
   Note that this is not a custom control ...
   ----------------------------------------------------------------
*/
class kmTreeItem(parentObj,name) of TreeItem(parentObj,name) 
   with(this)
//      bold          := false // boldface?
//      checked       := false // is it "checked"?
//      expanded      := false // if sub-items ...
//      image         := ""    // image when not selected
//      selectedImage := ""    // image when selected
   endwith

endclass

/*
   --------------------------------------------------------------
   kmTreeContainer
   Tree Container -- Container that has a kmTreeView control
   with a pushbutton for the two treeView events to expand
   and collapse the treeview's items ...

   NOTE: To use the fill routine to fill the treeView with
   treeItems, you need to use:
      form.kmTreeContainer1.FillIt( aArray )
         (where aArray is the array reference)

   NOTE2: There's some code below that you can copy to a simple
   program to test the tree container ... works jest fine ...
   --------------------------------------------------------------
*/

CLASS kmTreeContainer(parentObj) of KMCUSTCONTAINER(parentObj) custom
   with (this)
	   metric := 6 // Pixels
      width  := 129.00
      height := 223.00
   endwith

   this.KMTREEVIEW1 = new KMTREEVIEW(this)
   with (this.KMTREEVIEW1)
      height := 175
      left   := 10
      top    := 33
      width  := 105
   endwith

   this.EXPANDPUSHBUTTON = new PUSHBUTTON(this)
   with (this.EXPANDPUSHBUTTON)
      height   := 19
      left     := 32
      top      := 10
      width    := 27
      text     := ""
      speedBar := true
      upBitmap := "RESOURCE TS_APPEND"
      speedTip := "Expand tree items"
      onClick  := {|| this.parent.kmTreeView1.expand() }
   endwith

   this.COLLAPSEPUSHBUTTON = new PUSHBUTTON(this)
   with (this.COLLAPSEPUSHBUTTON)
      height   := 19
      left     := 61
      top      := 10
      width    := 27
      text     := ""
      speedBar := true
      upBitmap := "RESOURCE TS_DELETE"
      speedTip := "Collapse tree items"
      onClick  := {||this.parent.kmTreeView1.collapse() }
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use either form metric of "+;
                 "6 (Pixels) for this control." )
         return
      endif
   return

   function FillIt // we're really just calling the super class's
                   // TreeFill method ...
      parameter aArray 
      this.kmTreeView1.TreeFill( this.kmTreeView1, aArray )

endclass

/*
   ----------------------------------------------------------------
   The following code can be copied and pasted to test
   the treeview container on a form ...:
   This code appears to be broken somehow, and it won't
   work if you set the form's metrics to pixel (f.metric := 6)
   I am not sure if the broken code is here or in 
   the recursive code in the fillIt method ...

   // copy from here -----------------------
   set proc to :dUFLP:formcntl.cc additive
   // create array:
   a = {"item1",{"item11","item12"},;
        "item2",;
        "item3",;
        "item4",{"item41",;
                 "item42",{"item42_1",;
                           "item42_2",;
                           "item42_3"}},;
        "item5"}

   // instantiate form:
   f = new form()
   f.text = "Test kmTreeContainer"
   f.autoCenter = true
   // instantiate container:
   f.t = new kmTreeContainer( f )
   // load treeview object with items in array 'a'
   f.t.fillit( a)
   // open form
   f.open()
   // to here ------------------------------

   // ----------------------------------------------------
   if someone comes up with a way to read data out
   of a table directly into the treeView, that'd
   be really useful ... it would need to be generic --
   i.e., no specific field names (although using
   the OODML, it could use the fields array of the
   rowset to loop ...

   The code for a table would have to be recursive,
   and the code would have to do some sort of checking
   to see when a group changed ... 
   // ----------------------------------------------------

   ----------------------------------------------------------------
*/
// end of sample code for kmtreeContainer object



// -----------------------------------------------------------------
// POPUP CODE FOR ENTRYFIELDS, SPINBOXES, EDITOR
// -----------------------------------------------------------------
/*
   Code used by kmCustEntryfield, kmCustSpinbox, kmCustEditor
   to display a popup when using the right click of the mouse.
   The popup menu below this function is what is used.
   Code by: 
            Heinz Kesting (Original code)
            Marc van der Bergen (API code to determine
                                 if anything is highlighted)
            Ken Mayer     (Fleshed it out a bit ...)
            David Stone -- code from dBulletin article
                           incorporated below to set
                           top/left based on containers
                           when control is on a container.
*/
function ShowPopUp(flags, col, row, oRef)
   // Form reference:
   oForm = oRef.form
   // see if anything is in the clipboard, this is used
   // to determine if the Paste option is enabled (if anything
   // useful is in the clipboard -- no images, binary, etc. 
   // please):
   local cbv1, cbv2, cbv3
   private oRef, oFirstObj 
   extern clogical IsClipboardFormatAvailable(cuint) user32
   cbv1 = IsClipboardFormatAvailable(1) // CF-Text
   cbv2 = IsClipboardFormatAvailable(7) // CF_OEMTEXT
   cbv3 = IsClipboardFormatAvailable(13) // CF_UNICODETEXT
   if cbv1 = true or cbv2 = true or cbv3 = true // there is text in the clipboard, so paste is enabled
      oForm.pop1.m_paste.enabled = true
   else
      oForm.pop1.m_paste.enabled = false
   endif

   // if the entryfield is empty we disable the cut/copy options:
   if empty(oRef.value)
      oForm.pop1.m_cut.enabled  := false
      oForm.pop1.m_copy.enabled := false
   else
      oForm.pop1.m_cut.enabled  := true
      oForm.pop1.m_copy.enabled := true
   endif

   // Code by Marc van der Bergen to determine
   // if anything has been selected -- doesn't work
   // for editor control:
   if oRef.baseClassName # "EDITOR"
      local nResult,nStart,nEnd
      extern culong GetLen(chandle,culong,clong,clong) user32 from "SendMessageA"
      extern culong GetSel(chandle,culong,cptr culong,cptr culong) user32 from "SendMessageA"
      nend=0 ; nstart=0
      nResult=GetSel(oRef.hwnd,0xC1,-1,0)  // C1=EM_LINELENGTH
      GetSel(oRef.hwnd,0xB0,nStart,nEnd)   // B0=EM_GETSEL
      // if nEnd has any value other than zero, some text
      // has been selected by the user ...
      // If so, we want to disable the Select All option,
      // and enable the Delete option.
      oForm.pop1.m_selectall.enabled = (nEnd == 0)
      oForm.pop1.m_delete.enabled    = (nEnd  > 0 )
   endif

   // disallow SelectAll if object is an editor ...
   if oRef.baseClassName == "EDITOR"
      if type( "oForm.pop1.m_Line2" ) # "U"
         oForm.pop1.m_SelectAll.enabled := false
      endif
   endif

   // Code from David's article, modified for use here:
   oFirstObj = oRef//currentObj 
   oForm.pop1.top = Row 
   oForm.pop1.left = Col 

   do while oRef # oForm 
      oForm.pop1.top += oRef.top 
      oForm.pop1.left += oRef.left 
      oRef := oRef.parent 
   enddo 
   oRef = oFirstObj 

   // Ensure it loads left-aligned to the starting point. 
   if oForm.pop1.alignment <> 1 
      oForm.pop1.alignment := 1 
   endif 

   oForm.pop1.trackRight := false 
   oForm.pop1.open()
   // end of said code

   oRef.setfocus()
return

/*
   Popup Menu used for kmCustEntryfields, kmCustSpinboxes, kmCustEditor
   Allows standard Windows copy/cut/paste ... 
*/
class DEFAULTPOPUP(formObj, popupName) of POPUP(formObj, popupName)

   this.M_UNDO = new MENU(this)
   with (this.M_UNDO)
      onClick = {;form.obj.undo()}
      text = "Undo"
   endwith

   this.M_LINE1 = new MENU(this)
   with (this.M_LINE1)
      separator = true
   endwith

   this.M_CUT = new MENU(this)
   with (this.M_CUT)
      onClick = {;form.obj.cut()}
      enabled = false
      text = "Cut"
   endwith

   this.M_COPY = new MENU(this)
   with (this.M_COPY)
      onClick = {;form.obj.copy()}
      enabled = false
      text = "Copy"
   endwith

   this.M_PASTE = new MENU(this)
   with (this.M_PASTE)
      onClick = {;form.obj.paste()}
      enabled = false
      text = "Paste"
   endwith

   this.M_DELETE = new MENU(this)
   with (this.M_DELETE)
      onClick = {;form.obj.keyboard("{del}")}
      text = "Delete"
   endwith

   this.M_LINE2 = new MENU(this)
   with (this.M_LINE2)
      separator = true
   endwith

   this.M_SELECTALL = new MENU(this)
   with (this.M_SELECTALL)
      onClick= class::SelectAll
      enabled = false
      text = "Select All"
   endwith

   function SelectAll
      // deal with user deciding to "Select All":
      form.obj.keyboard( "{home}" )
      form.obj.keyboard( "{shift+end}" )
   return

endclass


function Repaint
/*
   ----------------------------------------------------------------
   This is a routine by Jim Sare with minor
   modifications by Ken Mayer. It can be used
   to ensure that if any processing has occurred
   that the form gets repainted (which sometimes
   doesn't happen) ...
   ----------------------------------------------------------------
*/
   parameter oForm
   if type( "oForm" ) == "U" or empty( oForm )
      oForm = form
   endif
   if type( "UpdateWindow" ) # "FP"
      extern CLOGICAL UpdateWindow( CHANDLE ) USER32
   endif
RETURN UpdateWindow( oForm.hWnd )
// end of function: Repaint()


/*
   ----------------------------------------------------------------
   This form is by Gary White and is used with
   the kmCustDateEntryField2 above -- it is a
   simple modal (child) form used to ask if the 
   user wants to move the date forward/backward
   off the holiday/weekend found.
   I set the onOpen event for the form so we can force
   the "nextWorkDay" button to be the one with focus 
   (pressing "Enter" will cause this button to be the
   one to fire, unless user changes buttons)
   ----------------------------------------------------------------
*/
class VerifyDateForm of FORM
   with (this)
      onOpen        := {; this.nextWorkDay.setFocus() }
   	mdi           := false
      scaleFontBold := false
      height        := 6.1364
      left          := 39.1429
      top           := 6.4545
      width         := 43.2857
      text          := "Verify Date"
      smallTitle    := true
      autoCenter    := true
      moveable      := true
   endwith

   this.TEXT1 = new TEXT(this)
   with (this.TEXT1)
      height = 3.8
      left = 0
      top = 0
      width = 43.2857
      border = true
      anchor = 2	// Top
      marginHorizontal = 1
      marginVertical = 0.2
      text = "The date you entered is either a weekend, or a holiday.  You may move this date to the first working day prior to that date, the first working day after that date, or leave it as is."
      borderStyle = CONTAINERBORDER // defined at top of this file
   endwith

   this.PRIORWORKDAY = new PUSHBUTTON(this)
   with (this.PRIORWORKDAY)
      height = 1.0909
      left = 1.2857
      top = 4.5
      width = 13.4286
      text = "Prior Workday"
      value = false
      onclick = {;form.parent.transfer="B";form.close()}
   endwith

   this.NEXTWORKDAY = new PUSHBUTTON(this)
   with (this.NEXTWORKDAY)
      height = 1.0909
      left = 14.7143
      top = 4.5
      width = 13.4286
      text = "Next Workday"
      value = false
      onclick = {;form.parent.transfer="F";form.close()}
   endwith

   this.LEAVEIT = new PUSHBUTTON(this)
   with (this.LEAVEIT)
      height = 1.0909
      left = 28.1429
      top = 4.5
      width = 13.4286
      text = "Leave As Is"
      value = false
      onclick = {;form.parent.transfer="X";form.close()}
   endwith
endclass
// End of Form: VerifyDateForm

// parse a color string and return the "background":
function bgColor( c )
return iif( "/" $ c, right( c, ( len(c) - at( "/", c ) ) ), c )

Function Smallfonts // from MiscAPI.prg
/*
   Programmer   : Romain Strieff [dBVIPS]
   Last Modified: 04/20/1999
   Rev. History.: 04/20/1999 -- Original
   Notes........: Returns True if Smallfonts are used
   .............: in Windows menus
   Written for..: Visual dBASE (32-bit versions)
   Usage .......: ? SmallFonts()
*/
   local dc,result
   extern CINT        GetDeviceCaps(cHandle, CINT) gdi32
   extern CINT        ReleaseDC(cHandle, cHandle) user32
   extern cHandle     GetDC(cHandle) user32
   #define LOGPIXELSX    88
   dc=getdc(0)
   result=GetDeviceCaps(dc,LOGPIXELSX)
   ReleaseDc(0,dc)
return result == 96
// eof: SmallFonts()

// by M. Dumont posted in the dBASE Newsgroups, used by
// combobox style 2 to change the appearance to what it
// should be. In Windows 10 the style 2 combobox
// is grey, it should be white. By changing it to 
// a style 1 and then setting it to read-only (see
// code associated with that class above), we trick
// dBASE and the world into thinking it's a style 2 
// combobox.
Class Struct_Base(nSize) Of String
    this.String = Replicate(Chr(0),nSize ) // Initialize with 0

   function getLong(nIndex)            // get the long(dword) value
     local n
     n = this.getByte(nIndex) + ;
         bitlshift(this.getByte(nIndex + 1), 8) + ;
         bitlshift(this.getByte(nIndex + 2), 16) + ;
         bitlshift(this.getByte(nIndex + 3), 24)
   return n   

   function setLong(nIndex, nValue)        // set the long value at the desired
     this.setByte(nIndex, bitand(nValue, 0xFF))    // position in the structure
     this.setByte(nIndex + 1, bitrshift(bitand(nValue, 0xFF00), 8))
     this.setByte(nIndex + 2, bitrshift(bitand(nValue, 0xFF0000), 16))
     this.setByte(nIndex + 3, bitrshift(bitand(nValue, 0xFF000000), 24))
   return
Endclass

/*
   ----------------------------------------------------------------
   End of file: FORMCNTL.CC
   ----------------------------------------------------------------
*/




