/*
   File.....: POINT.CC
   Author...: Marc van den Berghen
   Revision.: Extracted and combined code from two of 
              Marc's examples and re-packaged as a 
              .cc by Mervyn Bick
   Date.....: May 7, 2019

   When using functions from the Windows API one often
   needs a POINT structure.  A POINT structure contains
   two LONG values representing the x and y coordinates
   of a point.
   
   A typical use of a POINT structure would be to get
   the coordinates of the cursor. 

   GetCursorPos() from the API will give the corsor
   position on the screen.  To convert the cursor
   position to coordinates on the form use ScreenToClient() 

   if type('GetCursorPos') # 'FP'
      extern clogical GetCursorPos(cptr) user32
   endif
   if type('ScreenToClient') # 'FP'
      extern clogical ScreenToClient(chandle, cptr) user32
   endif       
   
   set procedure to :duflp:point.cc
   p = new POINT()
   GetCursorPos(p)
   ScreenToClient(form.hwnd,p) 
   
   nXcurpos = p.getx()
   nYcurpos = p.gety()
   
   To place the cursor at a given point use SetCursorPos() 
   in conjunction with ClientToScreen()
   
   if type("SetCursorPos") # "FP"
      extern cLogical SetCursorPos(cInt,cInt) user32
   endif
   if type('ClientToScreen') # 'FP'
      extern clogical ClientToScreen(chandle, cptr) user32
   endif
   
   set procedure to :duflp:point.cc
   p = new POINT()
   nNewCurX = 20
   nNewCurY = 40
   p.sety(nNewCurX)
   p.setx(nNewCurY)
   clientToScreen(form.hwnd,p)
   setCursorPos(p.getx(),p.gety())
   
   
   
*/
class POINT of string
   this.string := replicate(chr(0),4)

   function getX
      return int(this.getLong(0))

   function getY
      return int(this.getLong(4))
      
   function setX(nVal)
      this.setlong(0,nVal) 
      return int(this.getLong(0))

   function setY(nVal)
      this.setlong(4,nVal)
      return int(this.getLong(4))

   function getLong(nIndex)
      local n
      n = this.getByte(nIndex) + ;
      bitlshift(this.getByte(nIndex + 1), 8) + ;
      bitlshift(this.getByte(nIndex + 2), 16) + ;
      bitlshift(this.getByte(nIndex + 3), 24)
      if bitset(n, 31)
         n := - bitnot(n)
      endif
      return n
      
   function setLong(nIndex, nValue)		// set the long value at the desired 
      this.setByte(nIndex, bitand(nValue, 0xFF))    // position in the structure
      this.setByte(nIndex + 1, bitrshift(bitand(nValue, 0xFF00), 8))
      this.setByte(nIndex + 2, bitrshift(bitand(nValue, 0xFF0000), 16))
      this.setByte(nIndex + 3, bitrshift(bitand(nValue, 0xFF000000), 24))
   return      

   function release
      this.string = ''
      release object this
      return
endclass
