/*
   --------------------------------------------------------------------
    Class.....: TableDefine.cc
    Author....: Ken Mayer
    Date......: June 30, 2003
    Updates...: July 2, 2003 -- modified the form to use
                                a notebook, and a bit more ...
                                Fixed a problem with the code
                                for filling the index
                                table ...
                September 9, 2003 --
                                Modified to use Michael Nuwer's
                                BDEIndex.cc in the dUFLP
                                as of this release. Added
                                a couple of columns for the 
                                indexes, including Primary and
                                Distinct (for DBFs).
    Notes.....: This class is meant to be usable by a developer in 
                to display data about a table. The table does
                not need to be open, and can be in a SQL database,
                a BDE Alias, or a local table ...

    Usage.....: set procedure to :dUFLP:TabelDefine.cc additive
                oTable = new TableDefine( <cTableName>[, <cUserid>] )
                oTable.ShowTable()
                  // a dialog form (Defined below) is called that 
                  // displays the information needed -- field 
                  // and field descriptions (type, width, 
                  // decimals ...)
                  // Index tag names, expressions, etc.
    Parameters: cTableName -- can be a plain tablename (test.dbf) or
                              it can include a bde alias 
                              (:myalias:Test.dbf), but it should 
                              include the extension ...
                cUserid    -- This is the userid with
                              the password, in the format
                              of USERID/Password -- the slash
                              is vital. This is optional, but
                              can be used for databases
                              that have logins ...
    Returns...: None

       ******* dUFLP NOTE: ******* 
    This uses the following code from the dUFLP, and 
    needs to be deployed with your application if that's
    what you need:
             :dUFLP:BDEAlias.cc
             :dUFLP:BDEIndex.cc
               // which uses:
               :dUFLP:vStruct.h
               :dUFLP:vStruct.prg
    ******* End of dUFLP NOTE ******* 
   --------------------------------------------------------------------
*/

class TableDefine( cTable, cUserid )
   protect SetTable, Error

   // just to make sure we don't display multiple
   // errors ...
   this.Error = false
   this.UserID = ""
   this.Alias  = ""

   if argCount() == 0
      msgbox( "There is no tablename passed to this object --"+;
              "this is required.", "TableDefine Error", 16 )
      return
   endif

   // when instantiated we need to do ensure we have a table
   // and if it's got a BDE alias, we need to work with that ...
   if class::SetTable(cTable, cUserid)
 
   else
      if not this.error
         // error:
         msgbox( "The tablename is not valid.", "TableDefine Error", 16 )
      endif
      return
   endif

   function SetTable
      parameters cTable, cUserid
      local bSuccess
      bSuccess = false
      this.Table = null
      this.Alias = null
      if argCount() => 1 and type("cTable") == "C" 
			if file(cTable)      	
            this.Table = cTable
            this.TableName = this.Table
            bSuccess := true
         elseif cTable.indexOf(":") = 0 and cTable.lastIndexOf(":") > 1
        		// Maybe a database alias -- code borrowed from
            // DBF7File.cc ...:
        		set procedure to :dUFLP:BDEAlias.CC additive
            local b, nAt, cAliasName, cTableName, cDatabasePath

				// Extract alias and tableName 
         	nAt 		  = cTable.lastIndexOf(":")
            cAliasName = cTable.substring(1,nAt)
            cTableName = cTable.Right(cTable.Length-nAt-1)
            this.TableName = cTableName

            // Get database path
		      b = new BDEAlias()
            // Check to see if Alias exists
            if not b.databaseFound( cAliasName )
               msgbox( "BDE Alias does not exist.", "TableDefine Error", 16 )
               this.Error = true
               return bSuccess
            endif

            // Check here for non SQL server databases?
            try
               oData = new Database()
               if ArgCount() == 2 // we have a userid/password string
                  if type( "cUserID" ) # "C"
                     cUserID := ""
                  endif
                  oData.loginString := cUserID
                  this.UserId = cUserId
               endif
               oData.databaseName := cAliasName
               oData.active       := true
               oTable = new TableDef()
               oTable.database  := oData
               oTable.TableName := cTableName
               oTable.load()
            catch( Exception E )
               msgbox( "Unexpected error: "+;
                       e.code+" -- "+e.message+" -- Line: "+e.LineNo,;
                       "TableDefine Error", 16 )
               this.Error = true
               return bSuccess
            endtry
            cDatabasePath = b.databaseDir( cAliasname )

            // if here, we've loaded the table ...
            if oTable.tableType == "DBASE" or;
               oTable.tableType == "PARADOX"
               if file( cDatabasePath + "\" + cTableName )
                  this.Table = cDatabasePath + "\" + cTableName
                  this.Alias = cAliasName
                  bSuccess = true
               endif
            else // if we have no error:
               this.Alias = cAliasName
               this.Table = cTableName
               bSuccess = true
            endif
            release object oTable
            close procedure :dUFLP:BDEAlias.CC
         endif
      endif
   return bSuccess
      
   function ShowTable
      private oBDE, cPath, i, cTag, oTable, oData

      // check for problems:
      if this.error
         msgbox( "Can't do this, due to previous error.",;
                 "TableDefine Error", 16 )
         return
      endif

      // set up tabledef:
      oTable = new TableDef()
      if not empty( this.Alias )
         oData = new database()
         oData.databaseName := this.Alias
         if ArgCount() == 2 // we have a userid/password string
            oData.loginString := cUserID
         endif
         oData.active       := true
         oTable.database := oData
      endif
      oTable.tableName := this.Table
      oTable.load()

      // Create a temporary BDE Alias:
      set procedure to :dUFLP:BDEAlias.cc additive
      oBDE = new BDEAlias()
      cPath = GetEnv("TEMP")
      if empty( cPath )
         cPath = set("dire") 
      endif
      try
         oBDE.createAlias( "dUFLP_TableDefine", cPath )
      catch( Exception E )
         msgbox( "Unexpected error: "+e.code+" -- "+;
                 e.message, "TableDefine Error", 16 )
         return
      endtry

      // Now we need to create the table:
      try
         open Database dUFLP_TableDefine
      catch( Exception E )
         if e.code == 232 // Database already open
            // do nothing
         else
            msgbox( "Unexpected error: "+;
                    e.code+" -- "+e.message+" -- Line: "+e.LineNo,;
                    "TableDefine Error", 16 )
            return
         endif
      endtry
      // Fields table:
      if file( cPath+"\TempFieldsTable.dbf" ) or;
         file( cPath+"\TempFieldsTable.db" )
         drop table :dUFLP_TableDefine:TempFieldsTable
      endif
      if oTable.tableType == "DBASE"
         create table :dUFLP_TableDefine:TempFieldsTable (;
                      FieldName     char( 32 ),;
                      FieldType     char( 40 ),;
                      FieldLength   numeric( 3 ),;
                      DecimalLength numeric( 3 ) )
      elseif oTable.tableType == "PARADOX"
         create table :dUFLP_TableDefine:TempFieldsTable (;
                      FieldName     char( 32 ),;
                      FieldType     char( 40 ),;
                      FieldLength   numeric( 3 ) )
      else
         create table :dUFLP_TableDefine:TempFieldsTable (;
                      FieldName                   char( 32 ),;
                      FieldType                   char( 40 ),;
                      FieldLength                 numeric( 3 ),;
                      TempFieldsTable."Precision" numeric( 3 ),;
                      TempFieldsTable."Scale"     numeric( 3 ) )
      endif
      // Index table:
      if file( cPath+"\TempIndexTagsTable.dbf" ) or;
         file( cPath+"\TempIndexTagsTable.db" )
         drop table :dUFLP_TableDefine:TempIndexTagsTable
      endif
      if oTable.tableType == "DBASE"
         create table :dUFLP_TableDefine:TempIndexTagsTable (;
                      IndexName                       char( 40 ),;
                      IndexExpr                       char( 220 ),;
                      ForExpr                         char( 220 ),;
                      TempIndexTagsTable."Descending" boolean,;
                      TempIndexTagsTable."Unique"     boolean,;
                      TempIndexTagsTable."Primary"    boolean,;
                      TempIndexTagsTable."Distinct"   boolean )
      else // other than dBASE index (Paradox, SQL table ...)
         create table :dUFLP_TableDefine:TempIndexTagsTable (;
                      IndexName                          char( 40 ),;
                      Fields                             char( 220 ),;
                      TempIndexTagsTable."CaseSensitive" boolean,;
                      TempIndexTagsTable."Descending"    boolean,;
                      TempIndexTagsTable."Unique"        boolean,;
                      TempIndexTagsTable."Primary"       boolean )
      endif

      // Check here for non SQL server databases?
      try
         if not empty( this.Alias )
            oData = new Database()
            if not empty( this.UserID)
               oData.loginString := this.UserID
            endif
            oData.databaseName := this.Alias
            oData.active       := true
         endif
         oTable = new TableDef()
         if not empty( this.Alias)
            oTable.database  := oData
         endif
         oTable.TableName := this.TableName
         oTable.load()
      catch( Exception E )
         msgbox( "Unexpected error: "+;
                 e.code+" -- "+e.message+" -- Line: "+e.LineNo,;
                 "TableDefine Error", 16 )
         this.Error = true
      endtry

      // Create a database object:
      oData2 = new Database( "dUFLP_TableDefine" )

      // Set up the query for the Fields list:
      oQuery = new Query()
      oQuery.database := oData2
      oQuery.sql      := "select * from TempFieldsTable"
      oQuery.active   := true

      // Loop through fields:
      for i = 1 to oTable.fields.size

          // start adding a new row:
          oQuery.rowset.beginAppend()

          // FieldName
          oQuery.rowset.fields["FieldName"].value := ;
                              oTable.fields[i].fieldName

          // If it's not a lockfield:
          if oTable.fields[i].fieldName <> "_DBASELOCK"
             // Field Type
             oQuery.rowset.fields["FieldType"].value :=;
                              oTable.fields[i].type
             // Field Length:
             oQuery.rowset.fields["FieldLength"].value :=;
                              oTable.fields[i].length
          
             // dBFField Specific:
             if oTable.TableType == "DBASE"
                oQuery.rowset.fields["DecimalLength"].value := ;
                              oTable.fields[i].decimalLength

             // PDXField Specific:
             elseif oTable.TableType == "PARADOX"

                // ??

             // SQLField specific:
             else
                oQuery.rowset.fields["Precision"].value := ;
                              oTable.fields[i].Precision
                oQuery.rowset.fields["Scale"].value := ;
                              oTable.fields[i].Scale
             endif
          else
             // dBASE Lock field:
             oQuery.rowset.fields["FieldType"].value := "(Lock Field)"
          endif
          oQuery.rowset.save()

      next // end of fields

      if oTable.indexes.size > 0 // we actually have something
                                 // to do here:

         // Index tag table:
         oQuery = new Query()
         oQuery.database := oData2
         oQuery.sql      := "select * from TempIndexTagsTable"
         oQuery.active   := true

         // need a query for the table we're viewing
         oQuery2 = new Query()
         if not empty( this.alias )
            oQuery2.database := oData
         endif
         oQuery2.sql      := "select * from '"+this.Table+"'"
         oQuery2.active   := true

         local oIDX, aTagNames, nTags, nIndex, aTagProps
         // use the new :dUFLP:BDEIndex.cc class by Michael Nuwer
         set proc to :dUFLP:BDEIndex.cc additive
         oIDX = new BDEIndex()

         // loop through indexes:
         aTagNames = oIDX.TagList( oQuery2 ) 
         nIndexes = aTagNames.size

         // loop through indexes array
         for nIndex = 1 to nIndexes
             // add a row:
             oQuery.rowset.beginAppend()
             // get properties for this specific index tag:
             aTagProps = oIDX.TagProperties( oQuery2, nIndex )                
             if oQuery2.rowset.tableDriver == "DBASE"
                oQuery.rowset.fields["IndexName"].value  := aTagProps[ "IndexName" ]
                oQuery.rowset.fields["IndexExpr"].value  := aTagProps[ "Expression" ]
                oQuery.rowset.fields["ForExpr"].value    := aTagProps[ "ForExpression" ]
                oQuery.rowset.fields["Descending"].value := aTagProps[ "Descending" ]
                oQuery.rowset.fields["Unique"].value     := aTagProps[ "Unique" ]
                oQuery.rowset.fields["Distinct"].value   := aTagProps[ "Distinct" ]
                oQuery.rowset.fields["Primary"].value    := aTagProps[ "Primary" ]
             else // Non-dBF Index:
                oQuery.rowset.fields["IndexName"].value  := aTagProps[ "IndexName" ]
                if empty( aTagProps[ "IndexName" ] ) and aTagProps[ "Primary" ]
                   oQuery.rowset.fields["IndexName"].value := "(Primary Key)"
                endif
                // loop through number of fields:
                cFields = ""
                for i = 1 to aTagProps[ "NumFieldsInKey"]
                    cFieldNum = "KeyField"+str(i,2,0,"0")
                    cField = aTagProps[ cFieldNum ]
                    cFields += left( cField, at( "(", cField ) - 2 )
                    if i < aTagProps[ "NumFieldsInKey" ]
                       cFields += ";"
                    endif
                next
                oQuery.rowset.fields["Fields"].value        := cFields
                oQuery.rowset.fields["CaseSensitive"].value := aTagProps[ "CaseSensitive" ]
                oQuery.rowset.fields["Descending"].value    := aTagProps[ "Descending" ]
                oQuery.rowset.fields["Unique"].value        := aTagProps[ "Unique" ]
                oQuery.rowset.fields["Primary"].value       := aTagProps[ "Primary" ]
             endif // dBASE
            // save it
            oQuery.rowset.save()
         next // index tag
      endif // number of index tags

      // Now to use the form (this is shown after this class,
      // if you wish to modify it, feel free)
      fTableForm = new TableDefineForm()

      // The table name:
      if not empty( this.Alias )
         cOutString = ":"+this.Alias+":"+this.TableName
      else
         cOutString = this.TableName
      endif
      cOutString += "  -- Table Type: "+oTable.tableType
      if oTable.version > 0
         cOutString += " (Version/Level: "+oTable.version+")"
      endif

      // put values in text controls on the form:
      fTableForm.TableNameText.text := "Table Name: "+cOutString
      fTableForm.NumberRecordsText.text := ;
             "There "+;
             iif( oTable.RecordCount > 1 or oTable.RecordCount == 0,;
                  "are ", "is ")+;
             oTable.RecordCount+;
             iif( oTable.RecordCount > 1 or oTable.RecordCount == 0,;
                  " records ", " record ")+;
             "in this table currently."
      fTableForm.TableNotebook.NumFieldsText.text :=;
             "There "+;
             iif( oTable.fields.size > 1, "are ", "is ")+;
             oTable.fields.size+;
             iif( oTable.fields.size > 1, " Fields ", " Field ")+;
             " in this table."
      fTableForm.TableNotebook.NumIndexesText.text :=;
             "There "+;
             iif( oTable.Indexes.size > 1 or oTable.indexes.size == 0,;
                  "are ", "is ")+;
             oTable.indexes.size+;
             iif( oTable.Indexes.size > 1 or oTable.indexes.size == 0,;
                  " Indexes/Tags ", " Index/Tag ")+;
             " in this table."
      fTableForm.PRIMARYKEYTEXT.text := oTable.primaryKey

      // We have to set up the grids here, based on
      // the type of table:
      if oTable.TableType == "DBASE"

         // Fields Grid:
         fTableForm.TableNotebook.FIELDGRID.columns["Column1"] = new GRIDCOLUMN(fTableForm.TableNotebook.FieldGrid)
         fTableForm.TableNotebook.FIELDGRID.columns["Column1"].dataLink = fTableForm.tempfieldstable1.rowset.fields["fieldname"]
         fTableForm.TableNotebook.FIELDGRID.columns["COLUMN1"].editorType = 1	// EntryField
         fTableForm.TableNotebook.FIELDGRID.columns["Column1"].width = 23.7143
         fTableForm.TableNotebook.FIELDGRID.columns["Column2"] = new GRIDCOLUMN(fTableForm.TableNotebook.FieldGrid)
         fTableForm.TableNotebook.FIELDGRID.columns["Column2"].dataLink = fTableForm.tempfieldstable1.rowset.fields["fieldtype"]
         fTableForm.TableNotebook.FIELDGRID.columns["COLUMN2"].editorType = 1	// EntryField
         fTableForm.TableNotebook.FIELDGRID.columns["Column2"].width = 14.1429
         fTableForm.TableNotebook.FIELDGRID.columns["Column3"] = new GRIDCOLUMN(fTableForm.TableNotebook.FieldGrid)
         fTableForm.TableNotebook.FIELDGRID.columns["Column3"].dataLink = fTableForm.tempfieldstable1.rowset.fields["fieldlength"]
         fTableForm.TableNotebook.FIELDGRID.columns["COLUMN3"].editorType = 1	// EntryField
         fTableForm.TableNotebook.FIELDGRID.columns["Column3"].width = 6.7143
         fTableForm.TableNotebook.FIELDGRID.columns["Column4"] = new GRIDCOLUMN(fTableForm.TableNotebook.FieldGrid)
         fTableForm.TableNotebook.FIELDGRID.columns["Column4"].dataLink = fTableForm.tempfieldstable1.rowset.fields["decimallength"]
         fTableForm.TableNotebook.FIELDGRID.columns["COLUMN4"].editorType = 1	// EntryField
         fTableForm.TableNotebook.FIELDGRID.columns["Column4"].width = 8.5714
         fTableForm.TableNotebook.FIELDGRID.columns["Column1"].headingControl.value = "Field Name"
         fTableForm.TableNotebook.FIELDGRID.columns["Column2"].headingControl.value = "Type"
         fTableForm.TableNotebook.FIELDGRID.columns["Column3"].headingControl.value = "Length"
         fTableForm.TableNotebook.FIELDGRID.columns["Column4"].headingControl.value = "Decimals"

         // Indexes Grid
         fTableForm.TableNotebook.INDEXGRID.columns["Column1"] = new GRIDCOLUMN(fTableForm.TableNotebook.IndexGrid)
         fTableForm.TableNotebook.INDEXGRID.columns["Column1"].dataLink = fTableForm.tempindextagstable1.rowset.fields["indexname"]
         fTableForm.TableNotebook.INDEXGRID.columns["Column1"].editorType = 1	// EntryField
         fTableForm.TableNotebook.INDEXGRID.columns["Column1"].width = 21.1429
         fTableForm.TableNotebook.INDEXGRID.columns["Column2"] = new GRIDCOLUMN(fTableForm.TableNotebook.IndexGrid)
         fTableForm.TableNotebook.INDEXGRID.columns["Column2"].dataLink = fTableForm.tempindextagstable1.rowset.fields["indexexpr"]
         fTableForm.TableNotebook.INDEXGRID.columns["Column2"].editorType = 1	// EntryField
         fTableForm.TableNotebook.INDEXGRID.columns["Column2"].width = 30.0
         fTableForm.TableNotebook.INDEXGRID.columns["Column3"] = new GRIDCOLUMN(fTableForm.TableNotebook.IndexGrid)
         fTableForm.TableNotebook.INDEXGRID.columns["Column3"].dataLink = fTableForm.tempindextagstable1.rowset.fields["forexpr"]
         fTableForm.TableNotebook.INDEXGRID.columns["Column3"].editorType = 1	// EntryField
         fTableForm.TableNotebook.INDEXGRID.columns["Column3"].width = 30.0
         fTableForm.TableNotebook.INDEXGRID.columns["Column4"] = new GRIDCOLUMN(fTableForm.TableNotebook.IndexGrid)
         fTableForm.TableNotebook.INDEXGRID.columns["Column4"].dataLink = fTableForm.tempindextagstable1.rowset.fields["descending"]
         fTableForm.TableNotebook.INDEXGRID.columns["Column4"].editorType = 2	// CheckBox
         fTableForm.TableNotebook.INDEXGRID.columns["Column4"].width = 10.2857
         fTableForm.TableNotebook.INDEXGRID.columns["Column5"] = new GRIDCOLUMN(fTableForm.TableNotebook.IndexGrid)
         fTableForm.TableNotebook.INDEXGRID.columns["Column5"].dataLink = fTableForm.tempindextagstable1.rowset.fields["unique"]
         fTableForm.TableNotebook.INDEXGRID.columns["Column5"].editorType = 2	// CheckBox
         fTableForm.TableNotebook.INDEXGRID.columns["Column5"].width = 6.5714
         fTableForm.TableNotebook.INDEXGRID.columns["Column6"] = new GRIDCOLUMN(fTableForm.TableNotebook.IndexGrid)
         fTableForm.TableNotebook.INDEXGRID.columns["Column6"].dataLink = fTableForm.tempindextagstable1.rowset.fields["primary"]
         fTableForm.TableNotebook.INDEXGRID.columns["Column6"].editorType = 2	// CheckBox
         fTableForm.TableNotebook.INDEXGRID.columns["Column6"].width = 6.5714
         fTableForm.TableNotebook.INDEXGRID.columns["Column7"] = new GRIDCOLUMN(fTableForm.TableNotebook.IndexGrid)
         fTableForm.TableNotebook.INDEXGRID.columns["Column7"].dataLink = fTableForm.tempindextagstable1.rowset.fields["distinct"]
         fTableForm.TableNotebook.INDEXGRID.columns["Column7"].editorType = 2	// CheckBox
         fTableForm.TableNotebook.INDEXGRID.columns["Column7"].width = 6.5714
         // column headings
         fTableForm.TableNotebook.INDEXGRID.columns["Column1"].headingControl.value = "Index Name"
         fTableForm.TableNotebook.INDEXGRID.columns["Column2"].headingControl.value = "Expression"
         fTableForm.TableNotebook.INDEXGRID.columns["Column3"].headingControl.value = "For Expression"
         fTableForm.TableNotebook.INDEXGRID.columns["Column4"].headingControl.value = "Descending"
         fTableForm.TableNotebook.INDEXGRID.columns["Column5"].headingControl.value = "Unique"
         fTableForm.TableNotebook.INDEXGRID.columns["Column6"].headingControl.value = "Primary"
         fTableForm.TableNotebook.INDEXGRID.columns["Column7"].headingControl.value = "Distinct"
      else
         // Fields Grid:
         // Paradox or SQL:
         fTableForm.TableNotebook.FIELDGRID.columns["Column1"] = new GRIDCOLUMN(fTableForm.TableNotebook.FieldGrid)
         fTableForm.TableNotebook.FIELDGRID.columns["Column1"].dataLink = fTableForm.tempfieldstable1.rowset.fields["fieldname"]
         fTableForm.TableNotebook.FIELDGRID.columns["COLUMN1"].editorType = 1	// EntryField
         fTableForm.TableNotebook.FIELDGRID.columns["Column1"].width = 23.7143
         fTableForm.TableNotebook.FIELDGRID.columns["Column2"] = new GRIDCOLUMN(fTableForm.TableNotebook.FieldGrid)
         fTableForm.TableNotebook.FIELDGRID.columns["Column2"].dataLink = fTableForm.tempfieldstable1.rowset.fields["fieldtype"]
         fTableForm.TableNotebook.FIELDGRID.columns["COLUMN2"].editorType = 1	// EntryField
         fTableForm.TableNotebook.FIELDGRID.columns["Column2"].width = 14.1429
         fTableForm.TableNotebook.FIELDGRID.columns["Column3"] = new GRIDCOLUMN(fTableForm.TableNotebook.FieldGrid)
         fTableForm.TableNotebook.FIELDGRID.columns["Column3"].dataLink = fTableForm.tempfieldstable1.rowset.fields["fieldlength"]
         fTableForm.TableNotebook.FIELDGRID.columns["COLUMN3"].editorType = 1	// EntryField
         fTableForm.TableNotebook.FIELDGRID.columns["Column3"].width = 6.7143
         fTableForm.TableNotebook.FIELDGRID.columns["Column1"].headingControl.value = "Field Name"
         fTableForm.TableNotebook.FIELDGRID.columns["Column2"].headingControl.value = "Type"
         fTableForm.TableNotebook.FIELDGRID.columns["Column3"].headingControl.value = "Length"

         // SQL only:
         if oTable.tableType # "DBASE" and oTable.tableType # "PARADOX"
            fTableForm.TableNotebook.FIELDGRID.columns["Column4"] = new GRIDCOLUMN(fTableForm.TableNotebook.FieldGrid)
            fTableForm.TableNotebook.FIELDGRID.columns["Column4"].dataLink = fTableForm.tempfieldstable1.rowset.fields["precision"]
            fTableForm.TableNotebook.FIELDGRID.columns["COLUMN4"].editorType = 1	// EntryField
            fTableForm.TableNotebook.FIELDGRID.columns["Column4"].width = 8.5714
            fTableForm.TableNotebook.FIELDGRID.columns["Column4"].headingControl.value = "Precision"
            fTableForm.TableNotebook.FIELDGRID.columns["Column5"] = new GRIDCOLUMN(fTableForm.TableNotebook.FieldGrid)
            fTableForm.TableNotebook.FIELDGRID.columns["Column5"].dataLink = fTableForm.tempfieldstable1.rowset.fields["scale"]
            fTableForm.TableNotebook.FIELDGRID.columns["COLUMN5"].editorType = 1	// EntryField
            fTableForm.TableNotebook.FIELDGRID.columns["Column5"].width = 8.5714            
            fTableForm.TableNotebook.FIELDGRID.columns["Column5"].headingControl.value = "Scale"
         endif

         // Indexes Grid
         fTableForm.TableNotebook.INDEXGRID.columns["Column1"] = new GRIDCOLUMN(fTableForm.TableNotebook.IndexGrid)
         fTableForm.TableNotebook.INDEXGRID.columns["Column1"].dataLink = fTableForm.tempindextagstable1.rowset.fields["indexname"]
         fTableForm.TableNotebook.INDEXGRID.columns["Column1"].editorType = 1	// EntryField
         fTableForm.TableNotebook.INDEXGRID.columns["Column1"].width = 21.1429
         fTableForm.TableNotebook.INDEXGRID.columns["Column2"] = new GRIDCOLUMN(fTableForm.TableNotebook.IndexGrid)
         fTableForm.TableNotebook.INDEXGRID.columns["Column2"].dataLink = fTableForm.tempindextagstable1.rowset.fields["fields"]
         fTableForm.TableNotebook.INDEXGRID.columns["Column2"].editorType = 1	// EntryField
         fTableForm.TableNotebook.INDEXGRID.columns["Column2"].width = 30.0
         fTableForm.TableNotebook.INDEXGRID.columns["Column3"] = new GRIDCOLUMN(fTableForm.TableNotebook.IndexGrid)
         fTableForm.TableNotebook.INDEXGRID.columns["Column3"].dataLink = fTableForm.tempindextagstable1.rowset.fields["CaseSensitive"]
         fTableForm.TableNotebook.INDEXGRID.columns["Column3"].editorType = 2	// Checkbox
         fTableForm.TableNotebook.INDEXGRID.columns["Column3"].width = 13.29
         fTableForm.TableNotebook.INDEXGRID.columns["Column4"] = new GRIDCOLUMN(fTableForm.TableNotebook.IndexGrid)
         fTableForm.TableNotebook.INDEXGRID.columns["Column4"].dataLink = fTableForm.tempindextagstable1.rowset.fields["descending"]
         fTableForm.TableNotebook.INDEXGRID.columns["Column4"].editorType = 2	// CheckBox
         fTableForm.TableNotebook.INDEXGRID.columns["Column4"].width = 10.2857
         fTableForm.TableNotebook.INDEXGRID.columns["Column5"] = new GRIDCOLUMN(fTableForm.TableNotebook.IndexGrid)
         fTableForm.TableNotebook.INDEXGRID.columns["Column5"].dataLink = fTableForm.tempindextagstable1.rowset.fields["unique"]
         fTableForm.TableNotebook.INDEXGRID.columns["Column5"].editorType = 2	// CheckBox
         fTableForm.TableNotebook.INDEXGRID.columns["Column5"].width = 6.5714
         fTableForm.TableNotebook.INDEXGRID.columns["Column6"] = new GRIDCOLUMN(fTableForm.TableNotebook.IndexGrid)
         fTableForm.TableNotebook.INDEXGRID.columns["Column6"].dataLink = fTableForm.tempindextagstable1.rowset.fields["primary"]
         fTableForm.TableNotebook.INDEXGRID.columns["Column6"].editorType = 2	// CheckBox
         fTableForm.TableNotebook.INDEXGRID.columns["Column6"].width = 6.5714

         fTableForm.TableNotebook.INDEXGRID.columns["Column1"].headingControl.value = "Index Name"
         fTableForm.TableNotebook.INDEXGRID.columns["Column2"].headingControl.value = "Fields"
         fTableForm.TableNotebook.INDEXGRID.columns["Column3"].headingControl.value = "Case Sensitive"
         fTableForm.TableNotebook.INDEXGRID.columns["Column4"].headingControl.value = "Descending"
         fTableForm.TableNotebook.INDEXGRID.columns["Column5"].headingControl.value = "Unique"
         fTableForm.TableNotebook.INDEXGRID.columns["Column6"].headingControl.value = "Primary"
      endif

      // Open the Form:
      fTableForm.readModal()

      // when the user closes the Form:
      fTableForm.release()

      // now we're done, a bit of cleanup:
      // drop the table:
      oQuery.active := false
      oQuery2.active := false
      oData2.dropTable( "TempIndexTagsTable" )
      oData2.dropTable( "TempFieldsTable" )
      // release objects:
      release object oQuery
      release object oQuery2
      release object oTable
      oData2.active := false
      release object oData2

      // delete the temporary BDE Alias
      close database dUFLP_TableDefine
      oBDE.deleteAlias( "dUFLP_TableDefine" )
      release object oBDE
      close procedure :dUFLP:BDEAlias.cc

endclass


// ==============================================================
// TableDefine form -- this is used above to display the data
// about the table ...  it is dynamic which means that the
// grid definitions have to be modified based on what
// type of table we're viewing ...
// ==============================================================
** END HEADER -- do not remove this line
//
// Generated on 07/02/2003
//
parameter bModal
local f
f = new TableDefineForm()
if (bModal)
   f.mdi = false // ensure not MDI
   f.readModal()
else
   f.open()
endif

class TableDefineForm of FORM
   with (this)
      height = 17.3636
      left = 26.8571
      top = 1.1818
      width = 89.4286
      text = "Table Definition"
      autoCenter = true
      mdi = false
   endwith

   this.DUFLP_TABLEDEFINE1 = new DATABASE()
   this.DUFLP_TABLEDEFINE1.parent = this
   with (this.DUFLP_TABLEDEFINE1)
      left = 54.8571
      top = 1.6364
      databaseName = "dUFLP_TableDefine"
      active = true
   endwith

   this.TEMPINDEXTAGSTABLE1 = new QUERY()
   this.TEMPINDEXTAGSTABLE1.parent = this
   with (this.TEMPINDEXTAGSTABLE1)
      left = 77.2857
      top = 1.5455
      database = form.duflp_tabledefine1
      sql = "select * from tempindextagstable.dbf"
      active = true
   endwith

   this.TEMPFIELDSTABLE1 = new QUERY()
   this.TEMPFIELDSTABLE1.parent = this
   with (this.TEMPFIELDSTABLE1)
      left = 67.2857
      top = 1.6364
      database = form.duflp_tabledefine1
      sql = "select * from tempfieldstable.dbf"
      active = true
   endwith

   this.TITLETEXT = new TEXT(this)
   with (this.TITLETEXT)
      height = 1.2727
      left = 0.0
      top = 0.0
      width = 89.4286
      anchor = 2	// Top
      colorNormal = "Navy/BtnFace"
      alignHorizontal = 1	// Center
      fontSize = 14.0
      text = "Table Definition"
      borderStyle = 1	// Raised
   endwith

   this.TABLENAMETEXT = new TEXT(this)
   with (this.TABLENAMETEXT)
      height = 1.0
      left = 0.4286
      top = 1.3636
      width = 87.1429
      fontBold = true
      text = "Table Name Here"
   endwith

   this.NUMBERRECORDSTEXT = new TEXT(this)
   with (this.NUMBERRECORDSTEXT)
      height = 1.0
      left = 0.4286
      top = 2.5
      width = 87.1429
      text = "Number of Records In Table:"
   endwith

   this.TEXTLABEL1 = new TEXTLABEL(this)
   with (this.TEXTLABEL1)
      height = 0.7727
      left = 0.5714
      top = 3.3636
      width = 10.8571
      text = "Primary Key:"
   endwith

   this.PRIMARYKEYTEXT = new TEXTLABEL(this)
   with (this.PRIMARYKEYTEXT)
      height = 0.7727
      left = 12
      top = 3.3636
      width = 75.2857
      text = ""
   endwith

   this.TABLENOTEBOOK = new NOTEBOOK(this)
   with (this.TABLENOTEBOOK)
      height = 12.7727
      left = 0
      top = 4.4091
      width = 89.1429
      fontBold = true
      dataSource = 'ARRAY {"Fields", "Indexes"}'
   endwith

   this.TABLENOTEBOOK.FIELDGRID = new GRID(this.TABLENOTEBOOK)
   with (this.TABLENOTEBOOK.FIELDGRID)
      fontSize = 9.0
      headingFontSize = 9.0
      dataLink = form.tempfieldstable1.rowset
      cellHeight = 0.75
      integralHeight = true
      rowSelect = true
      allowEditing = false
      allowAddRows = false
      hScrollBar = 0	// Off
      height = 10.1818
      left = 1.2857
      top = 2.3636
      width = 73.4286
   endwith

   this.TABLENOTEBOOK.INDEXGRID = new GRID(this.TABLENOTEBOOK)
   with (this.TABLENOTEBOOK.INDEXGRID)
      fontSize = 8.0
      headingFontSize = 8.0
      dataLink = form.tempindextagstable1.rowset
      cellHeight = 0.7
      integralHeight = true
      rowSelect = true
      allowEditing = false
      allowAddRows = false
      pageno = 2
      height = 10.2273
      left = 1
      top = 2.3636
      width = 87.4286
   endwith

   this.TABLENOTEBOOK.NUMFIELDSTEXT = new TEXT(this.TABLENOTEBOOK)
   with (this.TABLENOTEBOOK.NUMFIELDSTEXT)
      height = 0.7273
      left = 1.4286
      top = 1.3636
      width = 35.1429
      text = "There are n Fields in this table"
   endwith

   this.TABLENOTEBOOK.NUMINDEXESTEXT = new TEXT(this.TABLENOTEBOOK)
   with (this.TABLENOTEBOOK.NUMINDEXESTEXT)
      height = 0.7273
      left = 1.4286
      top = 1.3636
      width = 45
      text = "There are n Indexes/Tags in this table"
      pageno = 2
   endwith

   this.rowset = this.tempfieldstable1.rowset
endclass

